/*
 ***** BEGIN LICENSE BLOCK *****
 * Version: EPL 2.0/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Eclipse Public
 * License Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.eclipse.org/legal/epl-v20.html
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * Copyright (C) 2013-2017 The JRuby Team (jruby@jruby.org)
 * 
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the EPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the EPL, the GPL or the LGPL.
 ***** END LICENSE BLOCK *****/
package org.jruby.ext.ripper;

public class YyTables {
   private static short[] combine(short[] t1, short[] t2, 
                                  short[] t3, short[] t4) {
      short[] t = new short[t1.length + t2.length + t3.length + t4.length];
      int index = 0;
      System.arraycopy(t1, 0, t, index, t1.length);
      index += t1.length;
      System.arraycopy(t2, 0, t, index, t2.length);
      index += t2.length;
      System.arraycopy(t3, 0, t, index, t3.length);
      index += t3.length;
      System.arraycopy(t4, 0, t, index, t4.length);
      return t;
   }

   public static final short[] yyTable() {
      return combine(yyTable1(), yyTable2(), yyTable3(), yyTable4());
   }

   public static final short[] yyCheck() {
      return combine(yyCheck1(), yyCheck2(), yyCheck3(), yyCheck4());
   }
   private static final short[] yyTable1() {
      return new short[] {

          309,  378,  312,  419,   97,  356,  422,  311,  652,  429, 
          108,  258,  261,   88,   88,  135,  135,  467,  490,  694, 
          254,  254,  254,  328,  358,  280,  254,  254,  219,  676, 
          254,  264,  442,  628,   83,   83,  218,  218,  134,  134, 
          292,  292,  306,  474,  292,  232,  660,  638,  134,  333, 
          337,  234,  764,  605,  640,  712,   88,  218,  639,  433, 
          354,  674,  254,  296,  296,  232,  711,  296,  861,  711, 
          308,  234,  529,  651,  364,  881,  531,  690,  651,  651, 
          366,  897,  218,  134,  587,  295,  295,  605,  736,  295, 
          284,  823,  702,  437,  712,  335,  339,  721,  107,  286, 
          833,  624,  605,  632,   78,    8,  750,  638,  810,  962, 
          651,  354,  579,  879,  305,    8,  376,  334,  587,  252, 
          252,  252,  651,  586,  830,  264,  597,  651,  435,  546, 
          548,  675,  107,  587,  134,  651,  842,   77,   90,   90, 
          254,  643,   88,  260,  571,  624,  579,  632,  662,  107, 
          517,  733,  260,  322, 1048,   76,   76,  586,  398,  874, 
          597,  579,  675,  640,    8,  610,  404,  854,  260,  463, 
          571,  308,  586,  651,  285,  597,  526,  835,  571,  445, 
         1032,  391,  526,  651,  651,  646,  440,  355,   75,  318, 
          441,  376,  259,  571,  571,  478,  675,  384,  377,  979, 
          434,  259,  651,  646,  301,  804,  305,  398,  556,  260, 
          525,  771,  887,  557,  610,  404,  528,  259,  757,  759, 
          490,  675,  389,  742,  503,  813,  304,  815,  743,  875, 
          434,  512,  651,  392,  387,  639,  355,  639,  595,  439, 
          571,  308,  639,  571, 1048,  772,  562,  563,  556,   88, 
          809,  651,  304,   87,  777,  654,  769,  390,  259,  910, 
          814,  770,  533,  385,  388,  917,  386,  962,  386,  537, 
          651,  583,  595,  377,  438,  651,  308,  419,  520, 1032, 
          314,  264,  712,  822,   79,  469,  780,  595,  881,  585, 
          564,  323,  504,  505,  419,  711,  711,  556,  401,  818, 
          328,  319,  403,  404,   87,  369,  496,  384,  385,   85, 
         1063,  740,  254,  850,  254,  254,  483,  640,  254,  651, 
          254,  370,  879,  585,  651,   88,  292,  373,  292,  292, 
          374,   88,   53,   88,  640,   79,  413,  879,  585,  426, 
          652,  424,   88,  652,  264,  425,  740,  673,  605,  296, 
          431,  296,  515,  501,  640,  929,  976, 1020,  997,  307, 
           85,  905,  443,  523,  354,  983,  430,  107,  445,  454, 
          987,  295,  457,  295,   76,  521,  522,  323,  451,  464, 
          589,  590,  591,  592,  656,  605,  452,  134,  931,  465, 
          254,  254,  254,  254,   88,  254,  254,  640,  479,  640, 
          458,  307,  605,  658,  640,  328,  610,  539,  540,  458, 
          638,  651,  469,  924,  542,  593,  445,  500,  605,  502, 
          496,  639,  284,  445,  445,  284,  651,  879,  746,  254, 
          107,  286,  846,  587,  286,  642,   78,  642,   88,  354, 
          664,  254,  642,  292,   88,  651,  640,  533,  496,  633, 
          445,  384,  385,  633,  836,  610,  651,  254,  583,   88, 
          254,  579,  134,  651,  496,   83,  852,  651,    8,   77, 
           88,  292,  586,  254,  698,  597,  571,  633,   88,  496, 
          838,  624,  843,  632,  254,  322,  719,  292,  254,  252, 
          252,  252,  252,  633,  594,  595,  653,  466,  906,  908, 
          653, 1022,  292,  862,  913,  915,  285,  398,  633,  285, 
          571,   97,  634,  391,  610,  404, 1076,  966,  254, 1098, 
           88,  712,   87,  970,  972,   83,  974,  651,  975,  384, 
          862,   88,  651,  595,  710,  711,  218,  713,  496,  695, 
          634,   83,  633,  556,  232,  717,  718,  254,  579,  354, 
          234,  539,  540,   79,  389,   80,  634,  799,  542,  677, 
          510,  292,  471,  726,  533,  392,  387,  633,  323,  846, 
          595,  634,  134,  729,  475,  519,   83,  583,   85,  583, 
          556, 1052,  579,  252,  585,   87,   88,  595,  775,  390, 
          775,  778,  639,  639,  432,  385,  388,  579,  639,  639, 
          386,  485,  786,  595,  651,  634,   80,  825,  822,  519, 
          486,  639,   81,  556,  746,  866,   79,  720,  107,  107, 
          574,  585,  775,  651,  519,  107,  107,  107,  996,  489, 
          634,  107,  107,  509,  792,  107,  386,   88,  513,  853, 
          792,   85,  415,   86,  514,  784,  457,  457,  786,  416, 
          421,  518,  457,  790,  585,  583,  519,  583,  524,  793, 
          530,  107,  817,   81,  786,  532,  354,  107,  538,  808, 
         1085,  901,  651,  323,  562,  563,  786,  792,  657,  254, 
           88, 1065, 1067, 1068, 1069,  821,  824,  807,  824,  134, 
           88,  651,  651,   84,   86,  824,  111,   82, 1017,  665, 
          803,  610,  507,  816, 1019,  686,  959,  786,   88,  651, 
           88,  218,  218,  693,  533,  678,  683,  783,  840,  336, 
          232,  919,  639,  812,  870,  789,  234,   97,  685,  687, 
          651,  789,  651,  696,  714,  746,   88,  746,  610, 1099, 
          675,  354,  715,  325,   84,  107,  870,  107,   82,  716, 
          468,  722,  725,  653,  640,  640,   17,   83,  651,  651, 
          640,  640,  730,  847,  134,  783,  848,  751,  789,  900, 
          458,  610,  752,  640,  851,  644,  753,  756,  252,  651, 
          758,  651,  663,  760,  763,   88,  865,  418,   88,  762, 
          766,   88,  642,  642,   83,  868,  776,  779,  642,  642, 
          496, 1026,  391,  392,  297,   17,   88,  574,  651,  254, 
          781,  642,  782,  519,  787,  644,  354,  663,   88,  458, 
          788,  795,  940,  292,   80,  944,  945,  920,  932,  633, 
          746, 1028,  800,   88,  802,  936,  254,  407,  408,  134, 
          291,  291,  805,  579,  291,  622,  811,   88,  822,  539, 
          540,  544, 1027,  834,  107,   88,  542,   83,  446,  934, 
          325,  828,  849,  829,  870,  724,  315,  317, 1061,  831, 
          583,  863,  291,  291,  706,  357,  359,  651,  832,  901, 
          579,   81,  651,  901,  640,   88,   88,   80,  867,  888, 
          462,  904,  634,  519,  325,  746,  911,  746,  786,  912, 
          519,  519,  930, 1091,  293,  299,  622,  933,  300,  325, 
          958,  991,   86,  579,  477,   16,  938,  107,   88,  107, 
          107,  935,  642,  107,  519,  107,  574,  622,  574, 1056, 
          107,  130,  746,  942,  134,  252,  107,  946,  107,  539, 
          540,  325,  947,  837,   81,  841,  542,  107,  948, 1014, 
          951,  547,  824,   88,  953,  955,  480,  481,  482,  956, 
          960,  484,   84,  651,   16,   88,   82,  963,  651,  964, 
          651,  965,  134,  901,  900,   86,  969,  900,  971,  900, 
          325,  900,  651,  651,  651,  973,  651,  325,  325,  651, 
          651,  651,  977,  626,  980,  107,  107,  107,  107,  107, 
          107,  107,  982, 1051, 1064,  984,  986,  992,   17,   17, 
           17,   88,  651,  651,   17,   17,  993,   17,  999, 1025, 
          902, 1033,  462,  903,  415,   84,  134,  458,   88,   82, 
           88,  416,  417, 1041,  107,   88, 1049,   95,  536, 1034, 
          107,  130,  107,  107, 1053, 1059,  107,  107,  551,  107, 
          552,  553,  554,  555, 1062, 1066, 1070,  496, 1072,  786, 
          824, 1074,  107, 1077,  107,  107,  254,  900,  941,  900, 
         1082,  900, 1090, 1101,  900,  107,  638,  651,  107,  640, 
          292,  639,  950,  107,  519,  640,  633,  644,   17,  107, 
          130,  640,  877,  107,  880,  415,  651,  652,  900,  458, 
          107,  661,  416,  445,   61,   62,   63,   64,  889,   17, 
          291,  291,  291,  359,  551,  291,  552,  553,  554,  555, 
          967,  968,  325,  107,  644,  107,  291,  415,  291,  291, 
          646,  640,  907,  909,  416,  461,  107,  508,  914,  916, 
          652,   17,  651,  375,  798,  768,  138,  445,  727,  634, 
         1089,  415,  107,  221,  794,  549,  707,  709,  416,  476, 
          336,  878, 1000,   93,   93,  136,  136,   16,   16,   16, 
          892,  883, 1044,   16,   16,  281,   16, 1043,  325,  907, 
          909,  283,  914,  916,  918,  325,  325,  446, 1018, 1055, 
          491,  626,  494,  114,  806,  884,  709, 1088,  884,  336, 
         1023,  539,  540,  541,  549,  420,   93,  896,  542,  325, 
          355,  596,  597,  598,  599,  600, 1092,  864,  601,  602, 
          603,  604,  605,  606,  607,  608,  609,  610,  611,  612, 
          613,  614,    0,    0,  615,  616,  617,  618,    0,  539, 
          540,  544,  114,  291,    0,  635,  542,   16,    0,  641, 
            0,  648,    0,  765,    0,  291,  641,  856,  857,  415, 
          858,  355,    0, 1057,    0, 1058,  416,  511,   16,    0, 
         1060,  291,  114,  622,  519,  415,    0,    0,    0,  539, 
          540,  545,  416,  535,  107,  107,  542,  291,    0,  641, 
          641,  641,   93,  522,  291,  107,    0,    0,  130,    0, 
           16, 1029,  291,  552,  553,  554,  555, 1036,    0,   45, 
          291,  291,    0,  107,  291,  107, 1042,    0,  801,   45, 
         1046,  519,    0,    0,    0,   59,   60,  522,  519,  519, 
            0,    0,   96,   96,  622,  126,  445,    0,  723,  641, 
            0,  107,  522,    0,  282,    0,    0,  101,    0,    0, 
          291,    0,    0,  291,  819,  622,  539,  540,  739,  325, 
            0,  291,    0,  542,    0,    0,  325,  325,   45,  539, 
          540,  767,  626,    0,  522,   96,  542,  844,    0,  845, 
            0,  415,  100,    0,  445,    0,  709,  336,  416,  797, 
          107,  445,  445,  107,  525,    0,  107,  107,  107,   93, 
            0,    0,    0,  107,  107,    0,    0,  291,    0,    0, 
         1095,  107,  896,    0,  107,  896,  107,  884,  445,  896, 
          489,  489,  489,  107,  415,   64,  415,  489,    0,  524, 
            0,  416, 1050,  416, 1054,   64,    0,  886,  107,    0, 
          551,  107,  552,  553,  554,  555,  103,  103,    0,    0, 
          114,  549,  107,  549,  549,  549,  549,  549,    0,    0, 
          871,   96,  549,  622,  551,  882,  552,  553,  554,  555, 
            0,    0,  291,  928,  744,   93,  629,    0,    0,  891, 
            0,   93,  745,   93,   64,  549,    0,    0,    0,  103, 
          107,  107,   93,  549,  549,  549,  549,    0,  744,  379, 
          380,  381,  382,  383,    0,  896,  949,  896,  291,  896, 
          629,  519,  896,    0,  355,    0,  101,    0,  519,  519, 
            0,    0,    0,    0,  622,  629,    0,  107,    0,    0, 
            0,  291,    0,  291,    0,    0,  896,    0,    0,  549, 
          291,  291,  277,    0,   93,  622,    0,    0,    0,    0, 
          101,    0,  277,    0,    0,  308,    0,  629,  107,    0, 
            0,    0,    0,  623,    0,    0,    0,  101,   96,    0, 
          107,   45,   45,   45,    0,  103,    0,   45,   45,    0, 
           45,    0,  120,    0,    0,  645,  277,  645,   93,  355, 
            0,  291,  645,    0,   93,    0,    0,  525,  998,    0, 
          277,  277,   45,   45,   45,   45,  641,  641,    0,   93, 
            0,  522,  641,  641,    0,    0,  107,  119,  522,  522, 
           93,  921,    0,  291,  623,  641,  336,  291,   93,    0, 
            0,  525,    0,  107,    0,  107,    0,    0,  525,  525, 
          107,    0,  522,    0,   96,  623,    0,    0,    0,   95, 
           96,   45,   96,  641,  641,    0,  641,  641,    0,  117, 
          291,   96,    0,  102,    0,    0,  524,  103,    0,    0, 
           93,  107,   45,  524,  524,    0, 1035,    0, 1037,    0, 
         1038,   93,  103,   95, 1039,    0,    0,   64,   64,   64, 
          885, 1047,   64,   64,   64,    0,   64,    0,    0,  355, 
           95,    0,  899,    0,   45,    0,   64,   64,  117,    0, 
          526,  386,    0,   96,  527,   64,   64,    0,   64,   64, 
           64,   64,   64,    0,  364,    0,    0,  399,  400,  994, 
          519,    0,    0,    0,  364,    0,  995,    0,    0,    0, 
          522,    0,    0,  401,    0,  402,  629,  403,  404,  405, 
          406,    0,  291,  409,  646,  410,  646,   96,  103,    0, 
            0,  646,    0,   96,  103,    0,  103,   64,  364,    0, 
           64,    0,  995, 1093,    0,  103,    0,    0,   96, 1094, 
          291, 1096,    0,  364, 1097,  101,    0,  364,   64,   96, 
            0,    0,    0,    0,  629,    0,    0,   96,    0,    0, 
            0,  629,  629,    0,  277,  277,  277,  625, 1105,  277, 
          277,  277,    0,  277,    0,  364,  355,    0,    0,    0, 
           64,    0,    0,  277,  277,  629,    0,  103,  629,    0, 
           93,    0,  277,  277,  525,  277,  277,  277,  277,   96, 
           93,  525,  525,    0,    0,    0,    0,  277,  101,    0, 
           96,  551,    0,  552,  553,  554,  555,  556,   93,    0, 
           93,    0,  557,    0,  277,    0,    0,    0,  103,    0, 
          103,  103,    0,    0,  277,  103,    0,  103,    0,    0, 
          291,  277,    0,    0,  277,  558,   93,  277,    0,  277, 
            0,  355,  103,  560,  561,  562,  563,    0,  121,  860, 
            0,    0,  122,  103,    0,  277,    0,  899,  272,    0, 
          899,  103, 1045,    0,  899,    0,  117,  277,   95,  622, 
          652,    0,  128,  277,    0,    0,    0,    0,    0,  623, 
          652,    0,    0,    0,    0,   93,    0,  277,   93,  564, 
            0,   93,  645,  645,    0,    0,    0,  526,  645,  645, 
            0,  527,    0,  103,  526,  526,   93,    0,  527,  527, 
            0,  645,    0,    0,  103,    0,  355,  519,   93,    0, 
            0,  128,    0,    0,  519,  519,    0,  522,    0,  652, 
          622,   95,    0,   93,  522,  522,  364,  364,  364,    0, 
          623,  364,  364,  364,    0,  364,    0,   93,    0,   96, 
          899,  622,  899,  629,  899,  364,  364,  899,    0,   96, 
            0,  623,    0,    0,  364,  364,    0,  364,  364,  364, 
          364,  364,    0,    0,    0,    0,    0,   96,    0,   96, 
          234,  899,    0,    0,    0,   93,   93,    0,    0,    0, 
          234,  364,  364,  364,  364,  364,  364,  364,  364,  364, 
          364,  364,  364,  364,  364,   96,    0,  364,  364,  364, 
          364,  136,    0,  364,    0,    0,  364,    0,    0,  364, 
            0,  364,  645,  364,  234,  364,    0,  364,  364,  364, 
          364,  364,  364,  364,    0,  364,    0,  364,  234,  234, 
            0,    0,    0,  234,    0,   98,    0,   98,    0, 1015, 
            0,    0,    0,   93,   96,  111,    0,   96,    0,    0, 
           96,  646,  646,  103,  109,   93,    0,  646,  646,  364, 
            0,    0,    0,  103,    0,   96,  277,  277,  277,   98, 
          646,    0,  277,  277,  981,  277,    0,   96,  985,  111, 
            0,  103,    0,  103,  522,    0,   98,    0,  109,    0, 
            0,    0,   96,  136,  277,  277,  111,  277,  277,  277, 
          277,   93,    0,    0,    0,  109,   96,    0,    0,  103, 
            0,    0,    0,  105,    0,    0,  522,    0,   93,  128, 
           93,    0,  652,  652,  652,   93,  325,  652,  652,  652, 
            0,  652,  629,    0,  551,  629,  552,  553,  554,  555, 
            0,  652,  652,    0,   96,   96,  277,    0,    0,    0, 
          652,  652,  519,  652,  652,  652,  652,  652,  103,    0, 
          529,  103,  519,    0,  103,  103,  103,  277,  744,    0, 
            0,  103,  103,    0,    0,    0,  961,    0,    0,  103, 
          629,  646,    0,    0,  103,    0,    0,  629,  629,    0, 
            0,  103,    0,  625,    0,    0,  519,    0,    0,  277, 
            0,    0,  652,    0,    0,  652,  103,  652, 1016,    0, 
          519,  519,   96,  114,  629,  519,    0,    0,    0,    0, 
          103,    0, 1071,  652,   96, 1073, 1075,    0,    0, 1078, 
         1079,    0,  234,  234,  234,    0,    0,  234,  234,  234, 
            0,  234,    0,  519,    0,    0,    0,    0,    0,    0, 
            0,  234,  234,    0,    0,  652,    0,    0,  103,  103, 
          234,  234,    0,  234,  234,  234,  234,  234,    0,    0, 
           96,    0,  117,  623,    0,  234,    0, 1100, 1102, 1103, 
         1104,    0,    0,    0,    0,    0,    0,   96, 1106,   96, 
            0,  522,    0,    0,   96,  103,    0,    0,    0,    0, 
            0,  522,    0,    0,   98,  623,    0,    0,    0,  234, 
            0,    0,  234,    0,  111,  234,    0,  234,    0,    0, 
            0,  522,    0,  109,  629,    0,  103,    0,  522,  522, 
            0,    0,    0,  234,  623,  522,    0,    0,  103,    0, 
            0,    0,    0,    0,    0,  234,    0,    0,  124,  522, 
          522,    0,  117,  522,  522,  623,    0,  386,    0,    0, 
          522,  522,    0,  325,    0,  234,  623,   98,    0,    0, 
          325,  325,  629,  399,  400,    0,  626,  111,    0,  629, 
          629,    0,  522,    0,  103,  625,  109,  623,    0,  401, 
            0,  402,    0,  403,  404,  405,  406,  529,    0,    0, 
            0,  103,    0,  103,  529,  529,  629,    0,  103,  622, 
          622,  622,    0,  622,  519,  519,  519,  622,  622,  519, 
          519,  519,  622,  519,  622,  622,  622,  622,  622,  622, 
          622,    0,  622,  519,  519,  622,  622,  622,  622,  622, 
          622,  622,  519,  519,  622,  519,  519,  519,  519,  519, 
            0,  622,    0,    0,  622,  622,  622,  519,  622,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  519, 
          519,  519,  519,  519,  519,  519,  519,  519,  519,  519, 
          519,  519,  519,  109,    0,  519,  519,  519,  519,  622, 
          531,  519,  622,  622,  519,  622,  622,  519,  622,  519, 
          622,  519,  622,  519,  622,  519,  519,  519,  519,  519, 
          519,  519,  622,  519,  622,  519,    0,  622,  622,  622, 
          622,  622,  622,  108,    0,    0,  622,  519,  622,  622, 
          629,  622,  622,  622,  622,    0,    0,    0,    0,    0, 
            0,    0,  120,    0,  622,  622,  622,  519,  623,  623, 
          623,    0,  623,  522,  522,  522,  623,  623,  522,  522, 
          522,  623,  522,  623,  623,  623,  623,  623,  623,  623, 
          625,  623,  522,  522,  623,  623,  623,  623,  623,  623, 
          623,  522,  522,  623,  522,  522,  522,  522,  522,    0, 
          623,  120,    0,  623,  623,  623,  522,  623,  623,  623, 
          623,  623,  623,  623,  623,  623,  623,  623,  522,  522, 
          522,  522,  522,  522,  522,  522,  522,  522,  522,  522, 
          522,  522,  104,  525,  522,  522,  522,  522,  623,  534, 
          522,  623,  623,  522,  623,  623,  522,  623,  522,  623, 
          522,  623,  522,  623,  522,  522,  522,  522,  522,  522, 
          522,  623,  522,  623,  522,    0,  623,  623,  623,  623, 
          623,  623,    0,    0,  629,  623,  522,  623,  623,  528, 
          623,  623,  623,  623,  629,    0,    0,    0,    0,    0, 
            0,    0,    0,  623,  623,  623,  522,    0,  638,  638, 
          638,    0,    0,    0,  638,  638,  551,  638,  552,  553, 
          554,  555,  556,  639,  639,  639,    0,  557,  629,  639, 
          639,    0,  639,    0,    0,    0,    0,    0,  128,  629, 
            0,    0,  629,  629,  386,  128,    0,  629,    0,    0, 
          558,  559,    0,    0,  386,    0,    0,    0,  560,  561, 
          562,  563,    0,  551,    0,  552,  553,  554,  555,  556, 
          399,  400,    0,    0,  557,  629,  401,    0,  127,  625, 
          403,  404,  405,  406,    0,    0,  401,  629,  638,    0, 
          403,  404,  405,  406,  629,  629,    0,  558,    0,    0, 
          625,    0,    0,  639,  564,  560,  561,  562,  563,  638, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  120, 
            0,  629,    0,  625,  639,    0,    0,  625,    0,    0, 
            0,    0,  386,  625,  625,  625,    0,  391,  392,    0, 
          625,  638,    0,    0,    0,    0,    0,    0,  399,  400, 
            0,  564,    0,    0,    0,    0,  639,    0,    0,    0, 
            0,  625,    0,    0,  401,    0,  402,  625,  403,  404, 
          405,  406,  407,  408,  409,    0,  410,  123,    0,    0, 
          525,  625,  625,    0,  127,    0,  625,  525,  525,    0, 
            0,    0,    0,    0,  551,    0,  552,  553,  554,  555, 
          556,    0,    0,    0,    0,  557,    0,    0,    0,    0, 
            0,    0,    0,    0,  625,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  528,    0,  558,    0, 
            0,    0,    0,  528,  528,  209,    0,    0,  562,  563, 
            0,  625,  625,  625,    0,  625,  629,  629,  629,  625, 
          625,  629,  629,  629,  625,  629,  625,  625,  625,  625, 
          625,  625,  625,    0,  629,  629,  629,  625,  625,  625, 
          625,  625,  625,  625,  629,  629,  625,  629,  629,  629, 
          629,  629,  564,  625,  209,    0,  625,  625,  625,  629, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  629,  629,  629,  629,  629,  629,  629,  629,  629, 
          629,  629,  629,  629,  629,  110,    0,  629,  629,  629, 
          629,  625,  202,  629,  625,  625,  629,  625,  625,  629, 
          625,  629,  625,  629,  625,  629,  625,  629,  629,  629, 
          629,  629,  629,  629,  625,  629,  629,  629,    0,  625, 
          625,  625,  625,  625,  625,  106,    0,    0,  625,  629, 
          625,  625,  626,  625,  625,  625,  625,    0,    0,    0, 
            0,  202,    0,    0,  119,    0,  625,  625,  625,  629, 
          625,  625,  625,    0,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  530,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,    0,  625,  119,    0,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,   97,  524,  625,  625,  625,  625, 
          625,    0,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,    0,  625,  625, 
          625,  625,  625,  625,    0,    0,  626,  625,  625,  625, 
          625,  521,  625,  625,  625,  625,  626,    0,    0,    0, 
            0,    0,    0,    0,    0,  625,  625,  625,  625,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  209,    0,  209,  209,  209,  209,  209,    0,    0, 
          626,    0,  209,    0,    0,    0,    0,    0,    0,    0, 
          129,  626,    0,  525,  626,  626,    0,  129,    0,  626, 
            0,  525,    0,    0,    0,  209,  209,    0,    0,    0, 
            0,    0,    0,  209,  209,  209,  209,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  626,    0,    0, 
          125,    0,    0,    0,    0,    0,    0,    0,    0,  626, 
            0,    0,    0,    0,    0,    0,  626,  626,    0,    0, 
            0,    0,  626,    0,    0,    0,    0,    0,  202,  209, 
          202,  202,  202,  202,  202,    0,    0,    0,    0,  202, 
            0,  119,    0,  626,    0,  316,    0,    0,    0,  530, 
          524,    0,    0,    0,    0,  316,  530,  530,  524,    0, 
            0,    0,  202,  202,    0,    0,    0,    0,    0,    0, 
          202,  202,  202,  202,    0,    0,    0,    0,    0,    0, 
            0,  386,  387,  388,  389,  390,  391,  392,  393,  316, 
          395,  396,    0,    0,    0,    0,    0,  399,  400,  116, 
            0,    0,  524,  316,  316,    0,    0,    0,  316,  524, 
          524,    0,    0,  401,    0,  402,  202,  403,  404,  405, 
          406,  407,  408,  409,    0,  410,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  316,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  521,    0, 
            0,    0,    0,    0,    0,  521,  521,  212,    0,    0, 
            0,    0,    0,  626,  626,  626,    0,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,    0,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,    0,  626,  212,    0,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,   96,    0,  626, 
          626,  626,  626,  626,  197,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
            0,  626,  626,  626,  626,  626,  626,   99,    0,    0, 
          626,  626,  626,  626,  520,  626,  626,  626,  626,    0, 
            0,    0,    0,  197,    0,    0,  121,    0,  626,  626, 
          626,  626,  624,  624,  624,    0,  624,  316,  316,  316, 
          624,  624,  316,  316,  316,  624,  316,  624,  624,  624, 
          624,  624,  624,  624,  523,  624,  316,  316,  624,  624, 
          624,  624,  624,  624,  624,  316,  316,  624,  316,  316, 
          316,  316,  316,    0,  624,  121,    0,  624,  624,  624, 
          316,  624,  624,  624,  624,  624,  624,  624,  624,  624, 
          624,  624,  316,  316,  316,  316,  316,  316,  316,  316, 
          316,  316,  316,  316,  316,  316,  113,  526,  316,  316, 
          316,  316,  624,    0,  316,  624,  624,  316,  624,  624, 
          316,  624,  316,  624,  316,  624,  316,  624,  316,  316, 
          316,  316,  316,  316,  316,  624,  316,  624,  316,    0, 
          624,  624,  624,  624,  624,  624,    0,    0,  630,  624, 
          316,  624,  624,  315,  624,  624,  624,  624,  630,    0, 
            0,    0,    0,    0,    0,    0,    0,  624,  624,  624, 
          316,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  212,    0,  212,  212,  212,  212,  212, 
            0,    0,  630,    0,  212,    0,    0,    0,    0,    0, 
            0,    0,  115,    0,    0,  526,  630,  630,    0,    0, 
            0,  630,    0,  526,    0,    0,    0,  212,  212,    0, 
            0,    0,    0,    0,    0,  212,  212,  212,  212,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  630, 
            0,    0,  118,    0,    0,    0,    0,    0,    0,    0, 
            0,  520,    0,    0,    0,    0,    0,    0,  520,  520, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          197,  212,  197,  197,  197,  197,  197,    0,    0,    0, 
            0,  197,    0,  121,    0,    0,    0,  631,    0,    0, 
            0,  523,  527,    0,    0,    0,    0,  631,  523,  523, 
          527,    0,    0,    0,  197,  197,    0,    0,    0,    0, 
            0,    0,  197,  197,  197,  197,    0,    0,    0,    0, 
            0,    0,    0,  386,  387,  388,  389,  390,  391,  392, 
            0,  631,  395,  396,    0,    0,    0,    0,    0,  399, 
          400,  132,    0,    0,  526,  631,  631,    0,    0,    0, 
          631,  526,  526,    0,    0,  401,    0,  402,  197,  403, 
          404,  405,  406,  407,  408,  409,    0,  410,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  631,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          315,    0,    0,    0,    0,    0,    0,  315,  315,  179, 
            0,    0,    0,    0,    0,  627,  627,  627,    0,  627, 
          630,  630,  630,  627,  627,  630,  630,  630,  627,  630, 
          627,  627,  627,  627,  627,  627,  627,    0,  630,  630, 
          630,  627,  627,  627,  627,  627,  627,  627,  630,  630, 
          627,  630,  630,  630,  630,  630,    0,  627,  179,    0, 
          627,  627,  627,  630,  627,  627,  627,  627,  627,  627, 
          627,  627,  627,  627,  627,  630,  630,  630,  630,  630, 
          630,  630,  630,  630,  630,  630,  630,  630,  630,  112, 
            0,  630,  630,  630,  630,  627,  180,  630,  627,  627, 
          630,  627,  627,  630,  627,  630,  627,  630,  627,  630, 
          627,  630,  630,  630,  630,  630,  630,  630,  627,  630, 
          630,  630,    0,  627,  627,  627,  627,  627,  627,    0, 
            0,    0,  627,  630,  627,  627,  326,  627,  627,  627, 
          627,    0,    0,    0,    0,  180,    0,    0,  122,    0, 
          627,  627,  627,  630,  628,  628,  628,    0,  628,  631, 
          631,  631,  628,  628,  631,  631,  631,  628,  631,  628, 
          628,  628,  628,  628,  628,  628,    0,  631,  631,  631, 
          628,  628,  628,  628,  628,  628,  628,  631,  631,  628, 
          631,  631,  631,  631,  631,    0,  628,  122,    0,  628, 
          628,  628,  631,  628,  628,  628,  628,  628,  628,  628, 
          628,  628,  628,  628,  631,  631,  631,  631,  631,  631, 
          631,  631,  631,  631,  631,  631,  631,  631,    0,  527, 
          631,  631,  631,  631,  628,    0,  631,  628,  628,  631, 
          628,  628,  631,  628,  631,  628,  631,  628,  631,  628, 
          631,  631,  631,  631,  631,  631,  631,  628,  631,  631, 
          631,    0,  628,  628,  628,  628,  628,  628,    0,    0, 
          325,  628,  631,  628,  628,    0,  628,  628,  628,  628, 
          325,    0,    0,    0,    0,    0,    0,    0,    0,  628, 
          628,  628,  631,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  179,    0,  179,  179,  179, 
          179,  179,    0,    0,  325,    0,  179,    0,    0,    0, 
            0,    0,    0,    0,  131,    0,    0,  529,  325,  325, 
            0,  130,    0,  325,    0,  529,    0,    0,    0,  179, 
          179,    0,    0,    0,    0,    0,    0,  179,  179,  179, 
          179,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  325,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  326,    0,    0,    0,    0,    0,    0, 
          326,  326,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  180,  179,  180,  180,  180,  180,  180,    0, 
            0,    0,    0,  180,    0,  122,    0,    0,    0,  444, 
            0,    0,    0,    0,  528,    0,    0,    0,    0,  444, 
            0,    0,  528,    0,    0,    0,  180,  180,    0,    0, 
            0,    0,    0,    0,  180,  180,  180,  180,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  444,  386,  387,  388,  389,  390,  391, 
          392,  393,  394,  395,  396,    0,  527,    0,  444,    0, 
          399,  400,  444,  527,  527,    0,    0,    0,    0,    0, 
          180,    0,    0,    0,    0,    0,  401,    0,  402,    0, 
          403,  404,  405,  406,  407,  408,  409,    0,  410,    0, 
          444,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  181,    0,    0,    0,    0,    0,  626,  626,  626, 
            0,  626,  325,  325,  325,  626,  626,  325,  325,  325, 
          626,  325,  626,  626,  626,  626,  626,  626,  626,    0, 
            0,  325,  325,  626,  626,  626,  626,  626,  626,  626, 
          325,  325,  626,  325,  325,  325,  325,  325,    0,  626, 
          181,    0,  626,  626,  626,  325,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  325,  325,  325, 
          325,  325,  325,  325,  325,  325,  325,  325,  325,  325, 
          325,    0,    0,  325,  325,  325,  325,  626,  138,  325, 
          626,  626,  325,  626,  626,  325,  626,  325,  626,  325, 
          626,  325,  626,  325,  325,  325,  325,  325,  325,  325, 
          626,  325,    0,  325,    0,  626,  626,  626,  626,  626, 
          626,    0,    0,    0,  626,  325,  626,  626,    0,  626, 
          626,  626,  626,    0,    0,    0,    0,  138,    0,    0, 
          124,    0,  626,  626,  626,  325,  292,  292,  292,    0, 
          292,  444,  444,  444,  292,  292,  444,  444,  444,  292, 
          444,  292,  292,  292,  292,  292,  292,  292,    0,  444, 
          444,  444,  292,  292,  292,  292,  292,  292,  292,  444, 
          444,  292,  444,  444,  444,  444,  444,    0,  292,  124, 
            0,  292,  292,  292,    0,  292,  292,  292,  292,  292, 
          292,  292,  292,  292,  292,  292,  444,  444,  444,  444, 
          444,  444,  444,  444,  444,  444,  444,  444,  444,  444, 
          411,  529,  444,  444,  444,  444,  292,    0,  444,  292, 
            0,  444,  292,  292,  444,  292,  444,  292,  444,  292, 
          444,  292,  444,  444,  444,  444,  444,  444,  444,  292, 
          444,  444,  444,    0,  292,  292,  292,  292,  292,  292, 
            0,    0,  651,  292,    0,  292,  292,    0,  292,  292, 
          292,  292,  651,    0,    0,    0,    0,    0,    0,    0, 
            0,  292,  292,  292,  444,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  181,    0,  181, 
          181,  181,  181,  181,    0,    0,  651,    0,  181,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  530, 
            0,  651,    0,    0,    0,  651,    0,  530,    0,    0, 
            0,  181,  181,    0,    0,    0,    0,    0,    0,  181, 
          181,  181,  181,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  651,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  138,  181,  138,  138,  138,  138, 
          138,    0,    0,    0,    0,  138,    0,  124,    0,    0, 
            0,  332,    0,    0,    0,    0,  519,    0,    0,    0, 
            0,  332,    0,    0,  519,    0,    0,    0,  138,  138, 
            0,    0,    0,    0,    0,    0,  138,  138,  138,  138, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  332,    0,    0,    0,    0, 
            0,    0,    0,    0,  796,    0,    0,    0,  529,    0, 
          332,    0,    0,    0,  332,  529,  529,    0,    0,    0, 
            0,    0,  138,    0,  386,  387,  388,  389,  390,  391, 
          392,  393,  394,  395,  396,    0,  397,  398,    0,    0, 
          399,  400,  332,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  401,    0,  402,    0, 
          403,  404,  405,  406,  407,  408,  409,    0,  410,  292, 
          292,  292,    0,  292,  651,  651,  651,  292,  292,  651, 
          651,  651,  292,  651,  292,  292,  292,  292,  292,  292, 
          292,    0,  651,  651,  651,  292,  292,  292,  292,  292, 
          292,  292,  651,  651,  292,  651,  651,  651,  651,  651, 
            0,  292,    0,    0,  292,  292,  292,    0,  292,  292, 
          292,  292,  292,  292,  292,  292,  292,  292,  292,  651, 
          651,  651,  651,  651,  651,  651,  651,  651,  651,  651, 
          651,  651,  651,  263,    0,  651,  651,  651,  651,  292, 
            0,  651,  292,  263,  651,  292,  292,  651,  292,  651, 
          292,  651,  292,  651,  292,  651,  651,  651,  651,  651, 
          651,  651,  292,  651,  651,  651,    0,  292,  292,  292, 
          292,  292,  292,    0,    0,    0,  292,  263,  292,  292, 
            0,  292,  292,  292,  292,    0,    0,    0,    0,    0, 
            0,  263,  263,    0,  292,  292,  292,  651,  292,  292, 
          292,    0,  292,  332,  332,  332,  292,  292,  332,  332, 
          332,  292,  332,  292,  292,  292,  292,  292,  292,  292, 
            0,    0,  332,  332,  292,  292,  292,  292,  292,  292, 
          292,  332,  332,  292,  332,  332,  332,  332,  332,    0, 
          292,    0,    0,  292,  292,  292,    0,  292,  292,  292, 
          292,  292,  292,  292,  292,  292,  292,  292,  332,  332, 
          332,  332,  332,  332,  332,  332,  332,  332,  332,  332, 
          332,  332,    0,    0,  332,  332,  332,  332,  292,    0, 
          332,  292,    0,  332,  292,  292,  332,  292,  332,  292, 
          332,  292,  332,  292,  332,  332,  332,  332,  332,  332, 
          332,  292,  332,  519,  332,    0,  292,  292,  292,  292, 
          292,  292,    0,  519,    0,  292,    0,  292,  292,    0, 
          292,  292,  292,  292,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  292,  292,  292,  332,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,   95,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  519,    0,  114,    0,  519,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  519,  263,  263,  263,    0,    0, 
          263,  263,  263,    0,  263,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  263,  263,    0,    0,    0,    0, 
            0,    0,    0,  263,  263,    0,  263,  263,  263,  263, 
          263,    0,    0,    0,    0,    0,    0,    0,  263,    0, 
            0,    0,  522,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  522,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  263,    0,    0,  263,    0,    0,  263,    0, 
          263,    0,    0,    0,    0,    0,   98,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  263,    0,  636,  630, 
            0,  522,  637,  117,    0,  522,  188,  189,  263,  190, 
          191,  192,  193,  194,  195,  196,    0,    0,  197,  198, 
            0,    0,    0,  199,  200,  201,  202,    0,  263,    0, 
            0,    0,    0,  522,    0,    0,    0,    0,    0,    0, 
          204,  205,    0,  206,  207,  208,  209,  210,  211,  212, 
          213,  214,  215,  216,  139,    0,  217,    0,    0,    0, 
          622,  622,  622,    0,  622,  519,  519,  519,  622,  622, 
            0,  519,  519,  622,  519,  622,  622,  622,  622,  622, 
          622,  622,    0,  622,    0,    0,  622,  622,  622,  622, 
          622,  622,  622,  519,  519,  622,  519,  519,  519,  519, 
          519,    0,  622,  139,    0,  622,  622,  622,    0,  622, 
          622,  622,  622,  622,  622,  622,  622,  622,  622,  622, 
          519,  519,  519,  519,  519,  519,  519,  519,  519,  519, 
          519,  519,  519,  519,    0,    0,  519,  519,  519,  519, 
          622,  531,  108,  622,  622,  519,  622,  622,    0,  622, 
            0,  622,  519,  622,  519,  622,  519,  519,  519,  519, 
          519,  519,  519,  622,  519,  622,  519,    0,  622,  622, 
          622,  622,  622,  622,    0,    0,  108,  622,    0,  622, 
          622,    0,  622,  622,  622,  622,    0,    0,    0,    0, 
            0,    0,    0,  108,    0,  622,  622,  622,  519,  623, 
          623,  623,    0,  623,  522,  522,  522,  623,  623,    0, 
          522,  522,  623,  522,  623,  623,  623,  623,  623,  623, 
          623,    0,  623,  625,    0,  623,  623,  623,  623,  623, 
          623,  623,  522,  522,  623,  522,  522,  522,  522,  522, 
            0,  623,    0,    0,  623,  623,  623,    0,  623,  623, 
          623,  623,  623,  623,  623,  623,  623,  623,  623,  522, 
          522,  522,  522,  522,  522,  522,  522,  522,  522,  522, 
          522,  522,  522,  411,    0,  522,  522,  522,  522,  623, 
          534,    0,  623,  623,  522,  623,  623,    0,  623,    0, 
          623,  522,  623,  522,  623,  522,  522,  522,  522,  522, 
          522,  522,  623,  522,  623,  522,    0,  623,  623,  623, 
          623,  623,  623,    0,    0,  629,  623,    0,  623,  623, 
            0,  623,  623,  623,  623,  629,    0,    0,    0,    0, 
            0,    0,    0,    0,  623,  623,  623,  522,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          139,    0,  139,  139,  139,  139,  139,    0,    0,  109, 
            0,  139,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  522,    0,  629,    0,  128,    0,  629,    0, 
          522,    0,    0,    0,  139,  139,    0,    0,    0,    0, 
            0,    0,  139,  139,  139,  139,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  629,    0,    0,    0, 
            0,  108,  625,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  139,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  625,    0,    0,    0,    0,    0, 
          625,    0,    0,    0,  625,    0,    0,  625,  625,    0, 
            0,    0,    0,  625,  108,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  625,    0,    0,    0,  108,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  625,    0,  127,    0,  625,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  386,  387,  388, 
          389,  390,  391,  392,  393,  394,  395,  396,    0,  397, 
          398,    0,    0,  399,  400,  625,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  401, 
            0,  402,    0,  403,  404,  405,  406,  407,  408,  409, 
            0,  410,  625,  625,  625,    0,  625,  629,  629,  629, 
          625,  625,    0,  629,  629,  625,  629,  625,  625,  625, 
          625,  625,  625,  625,    0,  629,    0,    0,  625,  625, 
          625,  625,  625,  625,  625,  629,  629,  625,  629,  629, 
          629,  629,  629,    0,  625,    0,    0,  625,  625,  625, 
            0,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  629,  629,  629,  629,  629,  629,  629,  629, 
          629,  629,  629,  629,  629,  629,    0,    0,  629,  629, 
          629,  629,  625,    0,  110,  625,  625,  629,  625,  625, 
            0,  625,    0,  625,  629,  625,  629,  625,  629,  629, 
          629,  629,  629,  629,  629,  625,  629,  629,  629,    0, 
          625,  625,  625,  625,  625,  625,    0,    0,  110,  625, 
            0,  625,  625,    0,  625,  625,  625,  625,    0,    0, 
            0,    0,    0,    0,    0,  110,    0,  625,  625, 
      };
   }

   private static final short[] yyTable2() {
      return new short[] {

          625,  629,  625,  625,  625,    0,  625,  625,  625,  625, 
          625,  625,    0,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,    0,  625,  626,    0,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,    0,  625,    0,    0,  625,  625,  625, 
            0,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,    0,    0,  625,  625, 
          625,  625,  625,    0,    0,  625,  625,  625,  625,  625, 
            0,  625,    0,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,    0, 
          625,  625,  625,  625,  625,  625,    0,    0,  626,  625, 
            0,  625,  625,    0,  625,  625,  625,  625,  626,    0, 
            0,    0,    0,    0,    0,    0,    0,  625,  625,  625, 
          625,    0,    0,  621,  622,    0,    0,  623,    0,    0, 
            0,  188,  189,    0,  190,  191,  192,  193,  194,  195, 
          196,    0,  110,  197,  198,    0,    0,    0,  199,  200, 
          201,  202,    0,    0,    0,    0,    0,  626,  304,  129, 
            0,  626,    0,    0,    0,  204,  205,    0,  206,  207, 
          208,  209,  210,  211,  212,  213,  214,  215,  216,    0, 
            0,  217,    0,    0,    0,    0,    0,    0,    0,  626, 
            0,    0,    0,    0,  110,  626,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  629,    0,    0, 
            0,    0,    0,  626,    0,    0,    0,  629,    0,    0, 
          626,  626,    0,    0,    0,    0,  626,  110,    0,  629, 
          630,    0,    0,  631,    0,    0,    0,  188,  189,    0, 
          190,  191,  192,  193,  194,  195,  196,  626,    0,  197, 
          198,  629,    0,    0,  199,  200,  201,  202,    0,    0, 
            0,    0,    0,    0,  304,    0,  629,    0,  128,    0, 
          629,  204,  205,    0,  206,  207,  208,  209,  210,  211, 
          212,  213,  214,  215,  216,    0,    0,  217,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  629,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  626,  626,  626,    0,  626, 
          626,  626,  626,  626,  626,    0,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,    0,  626,    0, 
            0,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,    0,  626,    0,    0, 
          626,  626,  626,    0,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,    0, 
            0,  626,  626,  626,  626,  626,    0,  127,  626,  626, 
          626,  626,  626,    0,  626,    0,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,    0,  626,  626,  626,  626,  626,  626,    0, 
            0,    0,  626,    0,  626,  626,    0,  626,  626,  626, 
          626,    0,    0,    0,    0,    0,  127,    0,  123,    0, 
          626,  626,  626,  626,  625,  625,  625,    0,  625,  629, 
          629,  629,  625,  625,    0,  629,  629,  625,  629,  625, 
          625,  625,  625,  625,  625,  625,    0,  629,  625,    0, 
          625,  625,  625,  625,  625,  625,  625,  629,  629,  625, 
          629,  629,  629,  629,  629,    0,  625,  123,    0,  625, 
          625,  625,    0,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  629,  629,  629,  629,  629,  629, 
          629,  629,  629,  629,  629,  629,  629,  629,    0,  528, 
          629,  629,  629,  629,  625,    0,    0,  625,  625,  629, 
          625,  625,    0,  625,    0,  625,  629,  625,  629,  625, 
          629,  629,  629,  629,  629,  629,  629,  625,  629,  629, 
          629,    0,  625,  625,  625,  625,  625,  625,    0,    0, 
          625,  625,    0,  625,  625,    0,  625,  625,  625,  625, 
          625,    0,    0,    0,    0,    0,    0,    0,    0,  625, 
          625,  625,  629,    0,    0,  670,  622,    0,    0,  671, 
            0,    0,    0,  188,  189,    0,  190,  191,  192,  193, 
          194,  195,  196,    0,  625,  197,  198,    0,    0,    0, 
          199,  200,  201,  202,    0,    0,    0,    0,    0,  625, 
          304,  127,    0,  625,    0,    0,    0,  204,  205,    0, 
          206,  207,  208,  209,  210,  211,  212,  213,  214,  215, 
          216,    0,    0,  217,  127,    0,    0,    0,    0,    0, 
            0,  625,    0,    0,    0,    0,    0,  625,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  123,    0,    0,    0,  626, 
            0,    0,    0,    0,    0,  625,    0,    0,    0,  626, 
            0,    0,  625,  625,    0,    0,    0,    0,  625,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  625, 
            0,    0,    0,  626,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  528,    0,  626,    0, 
          129,    0,  626,  528,  528,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          626,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  625,  625,  625, 
            0,  625,  625,  625,  625,  625,  625,    0,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,    0, 
          625,    0,    0,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,    0,  625, 
            0,    0,  625,  625,  625,    0,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,    0,    0,  625,  625,  625,  625,  625,    0,  129, 
          625,  625,  625,  625,  625,    0,  625,    0,  625,  625, 
          625,  625,  625,  625,  625,  625,  625,  625,  625,  625, 
          625,  625,  625,  625,    0,  625,  625,  625,  625,  625, 
          625,    0,    0,    0,  625,    0,  625,  625,    0,  625, 
          625,  625,  625,    0,    0,    0,    0,    0,  129,    0, 
          125,    0,  625,  625,  625,  625,  626,  626,  626,    0, 
          626,  626,  626,  626,  626,  626,    0,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,    0,  626, 
          626,    0,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,    0,  626,  125, 
            0,  626,  626,  626,    0,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
            0,  530,  626,  626,  626,  626,  626,    0,    0,  626, 
          626,  626,  626,  626,    0,  626,    0,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  626, 
          626,  626,  626,    0,  626,  626,  626,  626,  626,  626, 
            0,    0,  325,  626,    0,  626,  626,    0,  626,  626, 
          626,  626,  325,    0,    0,  100,    0,    0,    0,    0, 
            0,  626,  626,  626,  626,    0,    0,  636,  630,    0, 
            0,  637,    0,    0,    0,  188,  189,    0,  190,  191, 
          192,  193,  194,  195,  196,    0,  111,  197,  198,  100, 
            0,    0,  199,  200,  201,  202,    0,    0,    0,    0, 
            0,  325,  304,  130,    0,  325,  100,    0,    0,  204, 
          205,    0,  206,  207,  208,  209,  210,  211,  212,  213, 
          214,  215,  216,    0,    0,  217,  129,    0,    0,    0, 
            0,    0,    0,  325,    0,    0,  524,    0,    0,  626, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  125,    0,    0, 
            0,  325,    0,    0,    0,    0,    0,  626,    0,    0, 
            0,  325,    0,    0,  626,  626,    0,    0,    0,    0, 
          626,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  626,    0,    0,    0,  325,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  530,    0, 
          325,    0,  130,    0,  325,  530,  530,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  325,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  626, 
          626,  626,    0,  626,  325,  325,  325,  626,  626,    0, 
          325,  325,  626,  325,  626,  626,  626,  626,  626,  626, 
          626,    0,    0,    0,  100,  626,  626,  626,  626,  626, 
          626,  626,  325,  325,  626,  325,  325,  325,  325,  325, 
            0,  626,    0,    0,  626,  626,  626,    0,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  626,  325, 
          325,  325,  325,  325,  325,  325,  325,  325,  325,  325, 
          325,  325,  325,  524,    0,  325,  325,  325,  325,  626, 
          524,  524,  626,  626,  325,  626,  626,  100,  626,    0, 
          626,  325,  626,  325,  626,  325,  325,  325,  325,  325, 
          325,  325,  626,  325,  651,  325,    0,  626,  626,  626, 
          626,  626,  626,    0,  651,    0,  626,    0,  626,  626, 
            0,  626,  626,  626,  626,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  626,  626,  626,  325,  626,  626, 
          626,    0,  626,  325,  325,  325,  626,  626,    0,  325, 
          325,  626,  325,  626,  626,  626,  626,  626,  626,  626, 
            0,    0,    0,  651,  626,  626,  626,  626,  626,  626, 
          626,  325,  325,  626,  325,  325,  325,  325,  325,    0, 
          626,    0,    0,  626,  626,  626,    0,  626,  626,  626, 
          626,  626,  626,  626,  626,  626,  626,  626,  325,  325, 
          325,  325,  325,  325,  325,  325,  325,  325,  325,  325, 
          325,  325,    0,    0,  325,  325,  325,  325,  626,    0, 
            0,  626,  626,  325,  626,  626,    0,  626,    0,  626, 
          325,  626,  325,  626,  325,  325,  325,  325,  325,  325, 
          325,  626,  325,  445,  325,    0,  626,  626,  626,  626, 
          626,  626,    0,  445,    0,  626,    0,  626,  626,    0, 
          626,  626,  626,  626,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  626,  626,  626,  325,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  445,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  445,  445,    0,  126,    0,  445,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  445,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  126,    0,  116,    0,    0,    0,    0, 
            0,  292,  292,  292,    0,  292,  651,  651,  651,  292, 
          292,  651,  651,  651,  292,  651,  292,  292,  292,  292, 
          292,  292,  292,    0,  651,  651,    0,  292,  292,  292, 
          292,  292,  292,  292,  651,  651,  292,  651,  651,  651, 
          651,  651,  126,  292,  116,    0,  292,  292,  292,    0, 
          292,  292,  292,  292,  292,  292,  292,  292,  292,  292, 
          292,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  651,    0,  445,    0,  521,    0,    0,  651, 
          651,  292,    0,    0,  292,    0,  651,  292,  292,    0, 
          292,    0,  292,    0,  292,    0,  292,    0,  444,    0, 
            0,    0,    0,    0,  292,    0,  651,  651,  444,  292, 
          292,  292,  292,  292,  292,    0,    0,    0,  292,    0, 
          292,  292,    0,  292,  292,  292,  292,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  292,  292,  292,  651, 
            0,    0,  444,    0,    0,  445,  445,  445,    0,    0, 
          445,  445,  445,    0,  445,    0,  444,  444,    0,    0, 
            0,  444,    0,  445,  445,  445,    0,    0,    0,    0, 
            0,    0,    0,  445,  445,    0,  445,  445,  445,  445, 
          445,    0,    0,    0,    0,    0,    0,    0,  445,  444, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          445,  445,  445,  445,  445,  445,  445,  445,  445,  445, 
          445,  445,  445,  445,    0,    0,  445,  445,  445,  445, 
            0,  525,  445,    0,    0,  445,    0,    0,  445,    0, 
          445,  525,  445,    0,  445,    0,  445,  445,  445,  445, 
          445,  445,  445,    0,  445,  445,  445,    0,    0,    0, 
          126,    0,  116,    0,    0,    0,    0,    0,  445,    0, 
            0,    0,    0,  445,    0,  525,    0,    0,    0,    0, 
            0,    0,    0,  115,    0,    0,    0,    0,  445,  525, 
          525,    0,  120,    0,  525,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  445,  525,  521,    0,    0,    0,    0,  445,  445, 
          521,  521,  115,    0,  118,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  524,    0, 
            0,    0,    0,    0,    0,  445,    0,    0,  524,    0, 
            0,    0,    0,    0,  520,    0,    0,    0,    0,    0, 
          444,  444,  444,    0,    0,  444,  444,  444,    0,  444, 
            0,    0,    0,  118,    0,    0,    0,    0,  444,  444, 
          444,    0,  524,    0,    0,    0,    0,    0,  444,  444, 
          132,  444,  444,  444,  444,  444,  524,  524,    0,  119, 
            0,  524,    0,  444,    0,  523,    0,    0,    0,    0, 
            0,    0,    0,  131,    0,  444,  444,  444,  444,  444, 
          444,  444,  444,  444,  444,  444,  444,  444,  444,  524, 
            0,  444,  444,  444,  444,    0,    0,  444,    0,  132, 
          444,    0,    0,  444,    0,  444,    0,  444,  102,  444, 
            0,  444,  444,  444,  444,  444,  444,  444,    0,  444, 
          444,  444,  131,    0,    0,    0,    0,    0,    0,    0, 
            0,  315,    0,  444,    0,    0,    0,    0,    0,    0, 
            0,    0,  102,  525,  525,  525,    0,    0,  525,  525, 
          525,    0,  525,  444,  326,    0,    0,    0,    0,  102, 
            0,    0,  525,  525,    0,    0,    0,    0,    0,    0, 
            0,  525,  525,    0,  525,  525,  525,  525,  525,    0, 
            0,    0,    0,    0,    0,    0,  525,  526,    0,  526, 
          115,    0,    0,    0,    0,    0,    0,  526,  525,  525, 
          525,  525,  525,  525,  525,  525,  525,  525,  525,  525, 
          525,  525,    0,    0,  525,  525,  525,  525,    0,  537, 
          525,    0,    0,  525,    0,    0,  525,    0,  525,    0, 
          525,  526,  525,    0,  525,  525,  525,  525,  525,  525, 
          525,  118,  525,    0,  525,  526,  526,    0,  121,    0, 
          526,  520,    0,    0,    0,    0,  525,    0,  520,  520, 
          524,  524,  524,    0,    0,  524,  524,  524,    0,  524, 
            0,    0,    0,    0,    0,    0,  525,    0,  526,  524, 
          524,    0,    0,    0,    0,    0,    0,    0,  524,  524, 
            0,  524,  524,  524,  524,  524,    0,    0,    0,    0, 
            0,    0,  523,  524,  527,    0,    0,  132,    0,  523, 
          523,    0,    0,    0,  527,  524,  524,  524,  524,  524, 
          524,  524,  524,  524,  524,  524,  524,  524,  524,    0, 
          131,  524,  524,  524,  524,    0,  536,  524,    0,    0, 
          524,    0,    0,  524,    0,  524,    0,  524,  527,  524, 
            0,  524,  524,  524,  524,  524,  524,  524,    0,  524, 
            0,  524,  527,  527,    0,  122,    0,  527,  315,    0, 
            0,    0,    0,  524,    0,  315,  315,  102,    0,  103, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  326,    0,  524,    0,  527,    0,    0,  326,  326, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  103,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  526,  529,    0,    0, 
          103,    0,    0,  526,  526,    0,    0,  529,    0,    0, 
          102,    0,    0,    0,    0,    0,    0,    0,    0,  526, 
          526,  526,    0,    0,  526,  526,  526,    0,  526,    0, 
          527,    0,    0,    0,    0,    0,    0,    0,  526,  526, 
            0,  529,    0,    0,    0,    0,    0,  526,  526,    0, 
          526,  526,  526,  526,  526,  529,  529,    0,  124,    0, 
          529,    0,  526,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  526,  526,  526,  526,  526,  526, 
          526,  526,  526,  526,  526,  526,  526,  526,  529,    0, 
          526,  526,  526,  526,    0,  538,  526,    0,    0,  526, 
          105,    0,  526,    0,  526,    0,  526,    0,  526,    0, 
          526,  526,  526,  526,  526,  526,  526,  434,  526,    0, 
          526,    0,    0,    0,    0,    0,    0,  434,    0,    0, 
            0,    0,  526,    0,  105,    0,  527,  527,  527,    0, 
            0,  527,  527,  527,    0,  527,    0,    0,    0,    0, 
            0,  105,  526,    0,    0,  527,  527,    0,    0,    0, 
            0,    0,    0,    0,  527,  527,    0,  527,  527,  527, 
          527,  527,    0,    0,    0,    0,  434,    0,    0,  527, 
          528,  529,    0,    0,    0,    0,    0,    0,    0,    0, 
          528,  527,  527,  527,  527,  527,  527,  527,  527,  527, 
          527,  527,  527,  527,  527,    0,    0,  527,  527,  527, 
          527,    0,  539,  527,    0,    0,  527,    0,  103,  527, 
            0,  527,    0,  527,  528,  527,    0,  527,  527,  527, 
          527,  527,  527,  527,    0,  527,    0,  527,  528,  528, 
            0,  123,    0,  528,    0,    0,    0,    0,    0,  527, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  529, 
          529,  529,    0,    0,  529,  529,  529,  527,  529,  527, 
            0,  528,    0,    0,  527,  527,    0,    0,  529,  529, 
            0,  103,    0,    0,    0,    0,    0,  529,  529,    0, 
          529,  529,  529,  529,  529,    0,    0,    0,    0,    0, 
            0,    0,  529,  530,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  530,  529,  529,  529,  529,  529,  529, 
          529,  529,  529,  529,  529,  529,  529,  529,    0,    0, 
          529,  529,  529,  529,    0,  541,  529,    0,    0,  529, 
            0,    0,  529,    0,  529,    0,  529,  530,  529,    0, 
          529,  529,  529,  529,  529,  529,  529,    0,  529,    0, 
          529,  530,  530,    0,  125,    0,  530,    0,    0,  105, 
            0,    0,  529,    0,    0,    0,    0,    0,    0,  434, 
          434,  434,    0,    0,  434,  434,  434,    0,  434,    0, 
            0,    0,  529,    0,  530,    0,    0,    0,  434,    0, 
            0,    0,    0,    0,    0,    0,    0,  434,  434,    0, 
          434,  434,  434,  434,  434,    0,    0,    0,  529,    0, 
          521,    0,    0,    0,    0,  529,  529,    0,    0,    0, 
          521,    0,  105,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  528,  528,  528,  434,    0,  528,  528,  528, 
            0,  528,  434,  434,    0,    0,    0,    0,    0,  434, 
            0,  528,  528,    0,  521,    0,    0,    0,    0,    0, 
          528,  528,    0,  528,  528,  528,  528,  528,  521,  521, 
          434,  116,    0,  521,    0,  528,    0,    0,    0,  308, 
            0,    0,    0,    0,    0,    0,    0,  528,  528,  528, 
          528,  528,  528,  528,  528,  528,  528,  528,  528,  528, 
          528,  521,  434,  528,  528,  528,  528,    0,  540,  528, 
            0,    0,  528,    0,    0,  528,    0,  528,    0,  528, 
            0,  528,    0,  528,  528,  528,  528,  528,  528,  528, 
            0,  528,  411,  528,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  528,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  530,  530,  530,    0,    0, 
          530,  530,  530,    0,  530,  528,    0,    0,    0,    0, 
            0,    0,    0,    0,  530,  530,    0,    0,    0,    0, 
            0,    0,    0,  530,  530,    0,  530,  530,  530,  530, 
          530,    0,    0,    0,    0,    0,    0,    0,  530,  520, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  520, 
          530,  530,  530,  530,  530,  530,  530,  530,  530,  530, 
          530,  530,  530,  530,    0,    0,  530,  530,  530,  530, 
            0,  542,  530,    0,    0,  530,    0,    0,  530,    0, 
          530,    0,  530,  520,  530,    0,  530,  530,  530,  530, 
          530,  530,  530,    0,  530,    0,  530,  520,  520,    0, 
          115,    0,  520,    0,    0,    0,    0,    0,  530,    0, 
            0,    0,  521,  521,  521,    0,    0,  521,  521,  521, 
            0,  521,    0,    0,    0,    0,    0,    0,  530,    0, 
          520,  521,  521,    0,    0,    0,    0,    0,    0,    0, 
          521,  521,    0,  521,  521,  521,  521,  521,    0,    0, 
            0,    0,    0,    0,    0,  521,  523,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  523,  521,  521,  521, 
          521,  521,  521,  521,  521,  521,  521,  521,  521,  521, 
          521,    0,    0,  521,  521,  521,  521,    0,  533,  521, 
            0,    0,  521,    0,    0,  521,    0,  521,    0,  521, 
          523,  521,    0,  521,  521,  521,  521,  521,  521,  521, 
            0,  521,    0,  521,  523,  523,    0,  118,    0,  523, 
            0,    0,    0,    0,    0,  521,  386,  387,  388,  389, 
          390,  391,  392,  393,  394,  395,  396,    0,  397,  398, 
            0,    0,  399,  400,    0,  521,    0,  523,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  401,    0, 
          402,    0,  403,  404,  405,  406,  407,  408,  409,    0, 
          410,    0,    0,    0,    0,    0,    0,    0,    0,  651, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  651, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  520,  520,  520,    0,    0,  520,  520,  520,    0, 
          520,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          520,  520,    0,  651,    0,    0,    0,    0,    0,  520, 
          520,    0,  520,  520,  520,  520,  520,  651,  651,    0, 
            0,    0,  651,    0,  520,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  520,  520,  520,  520, 
          520,  520,  520,  520,  520,  520,  520,  520,  520,  520, 
          651,    0,  520,  520,  520,  520,    0,  532,  520,    0, 
            0,  520,  104,    0,  520,    0,  520,    0,  520,    0, 
          520,    0,  520,  520,  520,  520,  520,  520,  520,   53, 
          520,    0,  520,    0,    0,    0,    0,    0,    0,   53, 
            0,    0,    0,    0,  520,    0,  104,    0,  523,  523, 
          523,    0,    0,  523,  523,  523,    0,  523,    0,    0, 
            0,    0,    0,  104,  520,    0,    0,  523,  523,    0, 
            0,    0,    0,    0,    0,    0,  523,  523,    0,  523, 
          523,  523,  523,  523,    0,    0,    0,    0,   53,    0, 
            0,  523,  315,  528,    0,    0,    0,    0,    0,    0, 
            0,    0,  315,  523,  523,  523,  523,  523,  523,  523, 
          523,  523,  523,  523,  523,  523,  523,    0,    0,  523, 
          523,  523,  523,    0,  535,  523,    0,    0,  523,    0, 
            0,  523,    0,  523,    0,  523,  315,  523,    0,  523, 
          523,  523,  523,  523,  523,  523,    0,  523,    0,  523, 
          315,  315,    0,  132,    0,  315,    0,    0,    0,    0, 
            0,  523,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  651,  651,  651,    0,    0,  651,  651,  651,    0, 
          651,  523,    0,  315,    0,    0,    0,    0,    0,  651, 
          651,  651,    0,    0,    0,    0,    0,    0,    0,  651, 
          651,    0,  651,  651,  651,  651,  651,    0,    0,    0, 
            0,    0,    0,    0,  651,  326,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  326,  651,  651,  651,  651, 
          651,  651,  651,  651,  651,  651,  651,  651,  651,  651, 
            0,    0,  651,  651,  651,  651,    0,    0,  651,    0, 
            0,  651,    0,    0,  651,    0,  651,    0,  651,  326, 
          651,    0,  651,  651,  651,  651,  651,  651,  651,    0, 
          651,  651,  651,  326,  326,    0,  131,    0,  326,    0, 
            0,  104,    0,    0,  651,    0,    0,    0,    0,    0, 
            0,   53,   53,   53,    0,    0,   53,   53,   53,    0, 
           53,    0,    0,    0,  651,    0,  326,    0,    0,    0, 
           53,    0,    0,    0,    0,    0,    0,    0,    0,   53, 
           53,    0,   53,   53,   53,   53,   53,    0,    0,    0, 
          528,    0,  468,    0,    0,    0,    0,  528,  528,    0, 
            0,    0,  468,    0,  104,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  315,  315,  315,    0,    0,  315, 
          315,  315,    0,  315,    0,    0,    0,    0,    0,    0, 
            0,   53,    0,  315,  315,    0,  468,    0,    0,    0, 
            0,    0,  315,  315,    0,  315,  315,  315,  315,  315, 
          468,  468,   53,    0,    0,  468,    0,  315,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  315, 
          315,  315,  315,  315,  315,  315,  315,  315,  315,  315, 
          315,  315,  315,  468,   53,  315,  315,  315,  315,    0, 
            0,  315,    0,    0,  315,    0,    0,  315,    0,  315, 
            0,  315,  106,  315,    0,  315,  315,  315,  315,  315, 
          315,  315,    0,  315,    0,  315,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  315,    0,    0, 
            0,    0,    0,    0,    0,    0,  106,  326,  326,  326, 
            0,    0,  326,  326,  326,    0,  326,  315,    0,    0, 
            0,    0,    0,  106,    0,    0,  326,  326,    0,    0, 
            0,    0,    0,    0,    0,  326,  326,    0,  326,  326, 
          326,  326,  326,    0,    0,    0,    0,    0,    0,    0, 
          326,  337,    0,  530,    0,    0,    0,    0,    0,    0, 
            0,  337,  326,  326,  326,  326,  326,  326,  326,  326, 
          326,  326,  326,  326,  326,  326,    0,    0,  326,  326, 
          326,  326,    0,    0,  326,    0,    0,  326,    0,    0, 
          326,    0,  326,    0,  326,  337,  326,    0,  326,  326, 
          326,  326,  326,  326,  326,    0,  326,    0,  326,  337, 
          337,    0,    0,    0,  337,    0,    0,    0,    0,    0, 
          326,    0,    0,    0,  468,  468,  468,    0,    0,  468, 
          468,  468,    0,  468,    0,    0,    0,    0,    0,    0, 
          326,    0,  337,  468,  468,    0,    0,    0,    0,    0, 
            0,    0,  468,  468,    0,  468,  468,  468,  468,  468, 
            0,    0,    0,    0,    0,    0,    0,  468,  265,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  265,  468, 
          468,  468,  468,  468,  468,  468,  468,  468,  468,  468, 
          468,  468,  468,    0,    0,  468,  468,  468,  468,    0, 
            0,  468,    0,    0,  468,    0,    0,  468,    0,  468, 
            0,  468,  265,  468,    0,  468,  468,  468,  468,  468, 
          468,  468,    0,  468,    0,  468,  265,  265,    0,    0, 
            0,  265,    0,    0,    0,    0,    0,  468,    0,    0, 
            0,  106,    0,   97,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  468,    0,  368, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,   97,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          530,  332,    0,    0,   97,    0,    0,  530,  530,    0, 
            0,  332,    0,    0,  106,    0,    0,    0,    0,    0, 
            0,    0,    0,  337,  337,  337,    0,    0,  337,  337, 
          337,    0,  337,    0,  521,    0,    0,    0,    0,    0, 
            0,    0,  337,  337,    0,  332,    0,    0,    0,    0, 
            0,  337,  337,    0,  337,  337,  337,  337,  337,  332, 
          332,    0,    0,    0,  332,    0,  337,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  337,  337, 
          337,  337,  337,  337,  337,  337,  337,  337,  337,  337, 
          337,  337,  332,    0,  337,  337,  337,  337,    0,    0, 
          337,    0,    0,  337,   96,    0,  337,    0,  337,    0, 
          337,    0,  337,    0,  337,  337,  337,  337,  337,  337, 
          337,   58,  337,    0,  337,    0,    0,    0,    0,    0, 
            0,   58,    0,    0,    0,    0,  337,    0,   96,    0, 
          265,  265,  265,    0,    0,  265,  265,  265,    0,  265, 
            0,    0,    0,    0,    0,   96,  337,    0,    0,  265, 
          265,    0,    0,    0,    0,    0,    0,    0,  265,  265, 
            0,  265,  265,  265,  265,  265,    0,    0,    0,    0, 
           58,    0,    0,  265,  514,  520,    0,    0,    0,    0, 
            0,    0,    0,    0,  514,  265,  265,  265,  265,  265, 
          265,  265,  265,  265,  265,  265,  368,  265,  265,    0, 
            0,  265,  265,  368,  368,    0,    0,  265,    0,    0, 
          265,    0,   97,  265,    0,  265,    0,  265,  514,  265, 
            0,  265,  265,  265,  265,  265,  265,  265,    0,  265, 
            0,  265,  514,  514,    0,    0,    0,  514,    0,    0, 
            0,    0,    0,  265,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  332,  332,  332,    0,    0,  332,  332, 
          332,  521,  332,  265,    0,  514,    0,    0,  521,  521, 
            0,    0,  332,  332,    0,   97,    0,    0,    0,    0, 
            0,  332,  332,    0,  332,  332,  332,  332,  332,    0, 
            0,    0,    0,    0,    0,    0,  332,  365,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  365,  332,  332, 
          332,  332,  332,  332,  332,  332,  332,  332,  332,  332, 
          332,  332,    0,    0,  332,  332,  332,  332,    0,    0, 
          332,    0,    0,  332,    0,    0,  332,    0,  332,    0, 
          332,  365,  332,    0,  332,  332,  332,  332,  332,  332, 
          332,    0,  332,    0,  332,    0,  365,    0,    0,    0, 
          365,    0,    0,   96,    0,    0,  332,    0,    0,    0, 
            0,    0,    0,   58,   58,   58,    0,    0,   58,   58, 
           58,    0,   58,    0,    0,    0,  332,    0,  365,    0, 
            0,    0,   58,    0,    0,    0,    0,    0,    0,    0, 
          329,   58,   58,    0,   58,   58,   58,   58,   58,    0, 
          329,    0,  520,    0,    0,    0,    0,    0,    0,  520, 
          520,    0,    0,    0,    0,    0,   96,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  514,  514,  514,    0, 
            0,  514,  514,  514,  329,  514,    0,    0,    0,    0, 
            0,    0,    0,   58,    0,  514,  514,    0,    0,  329, 
            0,    0,    0,  329,  514,  514,    0,  514,  514,  514, 
          514,  514,    0,    0,   58,    0,    0,    0,    0,  514, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  329,  514,  514,  514,  514,  514,  514,  514,  514, 
          514,  514,  514,  514,  514,    0,   58,  514,  514,  514, 
          514,    0,  236,  514,    0,    0,  514,    0,    0,  514, 
            0,  514,  236,  514,    0,  514,    0,  514,  514,  514, 
          514,  514,  514,  514,    0,  514,    0,  514,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  514, 
            0,    0,    0,    0,    0,    0,  236,    0,    0,  365, 
          365,  365,    0,    0,  365,  365,  365,    0,  365,  514, 
          236,  236,    0,    0,    0,  236,    0,    0,  365,  365, 
            0,    0,    0,    0,    0,    0,    0,  365,  365,    0, 
          365,  365,  365,  365,  365,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  365,  365,  365,  365,  365,  365, 
          365,  365,  365,  365,  365,  365,  365,  365,    0,    0, 
          365,  365,  365,  365,    0,    0,  365,    0,    0,  365, 
            0,    0,  365,    0,  365,    0,  365,    0,  365,    0, 
          365,  365,  365,  365,  365,  365,  365,    0,  365,    0, 
          365,    0,  329,  329,  329,    0,    0,  329,  329,  329, 
            0,  329,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  329,  329,    0,    0,    0,    0,    0,    0,    0, 
          329,  329,  365,  329,  329,  329,  329,  329,    0,    0, 
            0,    0,    0,    0,    0,    0,  237,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  237,  329,  329,  329, 
          329,  329,  329,  329,  329,  329,  329,  329,  329,  329, 
          329,    0,    0,  329,  329,  329,  329,    0,    0,  329, 
            0,    0,  329,    0,    0,  329,    0,  329,    0,  329, 
          237,  329,    0,  329,  329,  329,  329,  329,  329,  329, 
            0,  329,    0,  329,  237,  237,    0,    0,    0,  237, 
            0,    0,    0,    0,  236,  236,  236,    0,    0,  236, 
          236,  236,    0,  236,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  236,  236,  329,    0,    0,    0,    0, 
            0,    0,  236,  236,    0,  236,  236,  236,  236,  236, 
            0,    0,    0,    0,    0,    0,    0,  236,    0,    0, 
            0,   99,    0,    0,    0,    0,    0,    0,    0,  236, 
          236,  236,  236,  236,  236,  236,  236,  236,  236,  236, 
          246,  236,  236,    0,    0,  236,  236,    0,    0,    0, 
          246,  236,    0,    0,  236,   99,    0,  236,  113,  236, 
            0,  236,    0,  236,    0,  236,  236,  236,  236,  236, 
          236,  236,   99,  236,    0,  236,    0,    0,    0,    0, 
            0,    0,  112,    0,  246,    0,    0,  236,    0,    0, 
            0,    0,  113,    0,    0,    0,    0,    0,  246,  246, 
            0,    0,  523,  246,    0,    0,    0,  236,    0,  113, 
            0,    0,    0,    0,    0,    0,  112,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  112,    0,    0,    0,    0,    0,  315, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  243, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  243, 
            0,    0,    0,  326,    0,    0,    0,    0,  237,  237, 
          237,    0,    0,  237,  237,  237,    0,  237,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  237,  237,    0, 
            0,    0,    0,  243,    0,    0,  237,  237,    0,  237, 
          237,  237,  237,  237,    0,    0,    0,  243,  243,    0, 
            0,  237,  243,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  237,  237,  237,  237,  237,  237,  237, 
          237,  237,  237,  237,    0,  237,  237,    0,    0,  237, 
          237,    0,    0,    0,    0,  237,    0,    0,  237,    0, 
            0,  237,    0,  237,    0,  237,    0,  237,    0,  237, 
          237,  237,  237,  237,  237,  237,  240,  237,    0,  237, 
            0,    0,    0,    0,    0,    0,  240,    0,    0,    0, 
            0,  237,    0,    0,    0,    0,    0,    0,    0,    0, 
           99,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  237,  246,  246,  246,    0,    0,  246,  246,  246, 
          240,  246,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  246,  246,    0,  240,  240,    0,  113,    0,  240, 
          246,  246,    0,  246,  246,  246,  246,  246,    0,  523, 
            0,    0,    0,    0,    0,  246,  523,  523,    0,    0, 
            0,  112,    0,   99,    0,    0,    0,    0,  246,  246, 
          246,  246,  246,  246,  246,  246,  246,  246,    0,  246, 
          246,    0,    0,  246,  246,    0,  315,    0,    0,  246, 
            0,    0,  246,  315,  315,  246,    0,  246,    0,  246, 
          113,  246,    0,  246,  246,  246,  246,  246,  246,  246, 
          326,  246,    0,  246,    0,    0,    0,  326,  326,    0, 
            0,  243,  243,  243,  112,  246,  243,  243,  243,    0, 
          243,    0,  242,    0,    0,    0,    0,    0,    0,    0, 
          243,  243,  242,    0,    0,  246,    0,    0,    0,  243, 
          243,    0,  243,  243,  243,  243,  243,    0,    0,    0, 
            0,    0,    0,    0,  243,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  242,  243,  243,  243, 
          243,  243,  243,  243,  243,  243,  243,    0,  243,  243, 
          242,  242,  243,  243,    0,  242,    0,    0,  243,    0, 
            0,  243,    0,    0,  243,    0,  243,    0,  243,    0, 
          243,    0,  243,  243,  243,  243,  243,  243,  243,    0, 
          243,    0,  243,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  243,    0,    0,    0,  240,  240, 
          240,    0,    0,  240,  240,  240,    0,  240,    0,  241, 
            0,    0,    0,    0,  243,    0,    0,  240,  240,  241, 
            0,    0,    0,    0,    0,    0,  240,  240,    0,  240, 
          240,  240,  240,  240,    0,    0,    0,    0,    0,    0, 
            0,  240,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  241,  240,  240,  240,  240,  240,  240, 
          240,  240,  240,  240,    0,  240,  240,  241,  241,  240, 
          240,    0,  241,    0,    0,  240,    0,    0,  240,    0, 
            0,  240,    0,  240,    0,  240,    0,  240,    0,  240, 
          240,  240,  240,  240,  240,  240,    0,  240,    0,  240, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  240,    0,    0,    0,    0,    0,    0,  244,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  244,    0, 
            0,  240,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  242,  242,  242,    0,    0,  242, 
          242,  242,  244,  242,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  242,  242,    0,  244,  244,    0,    0, 
            0,  244,  242,  242,    0,  242,  242,  242,  242,  242, 
            0,    0,    0,    0,    0,    0,    0,  242,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          242,  242,  242,  242,  242,  242,  242,  242,  242,  242, 
            0,  242,  242,    0,    0,  242,  242,    0,    0,    0, 
            0,  242,    0,    0,  242,    0,    0,  242,    0,  242, 
            0,  242,    0,  242,    0,  242,  242,  242,  242,  242, 
          242,  242,    0,  242,  445,  242,    0,    0,    0,    0, 
            0,    0,    0,    0,  445,    0,    0,  242,    0,    0, 
            0,  241,  241,  241,    0,    0,  241,  241,  241,    0, 
          241,    0,    0,    0,    0,    0,    0,  242,    0,    0, 
          241,  241,    0,    0,    0,    0,    0,    0,  107,  241, 
          241,    0,  241,  241,  241,  241,  241,    0,    0,    0, 
            0,    0,    0,  445,  241,  126,    0,  445,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  241,  241,  241, 
          241,  241,  241,  241,  241,  241,  241,    0,  241,  241, 
            0,    0,  241,  241,    0,  445,    0,    0,  241,    0, 
            0,  241,    0,    0,  241,    0,  241,    0,  241,    0, 
          241,    0,  241,  241,  241,  241,  241,  241,  241,   60, 
          241,    0,  241,    0,    0,    0,    0,    0,    0,   60, 
          244,  244,  244,    0,  241,  244,  244,  244,    0,  244, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  244, 
          244,    0,    0,    0,  241,    0,    0,    0,  244,  244, 
            0,  244,  244,  244,  244,  244,    0,    0,    0,    0, 
            0,    0,    0,  244,  445,    0,    0,    0,   60,    0, 
            0,    0,    0,    0,  445,    0,  244,  244,  244,  244, 
          244,  244,  244,  244,  244,  244,    0,  244,  244,    0, 
            0,  244,  244,    0,    0,    0,    0,  244,    0,    0, 
          244,    0,    0,  244,    0,  244,    0,  244,  445,  244, 
            0,  244,  244,  244,  244,  244,  244,  244,    0,  244, 
            0,  244,    0,  445,    0,  126,    0,  445,    0,    0, 
            0,    0,    0,  244,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  244,    0,  445,  445,  445,  445,    0, 
            0,    0,  445,  445,    0,  445,    0,    0,    0,    0, 
            0,    0,    0,    0,  445,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  445,  445,    0,  445,  445,  445, 
          445,  445,    0,    0,    0,    0,    0,    0,    0,    0, 
          525,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          525,  445,  445,  445,  445,  445,  445,  445,  445,  445, 
          445,  445,  445,  445,  445,    0,    0,  445,  445,  445, 
          445,    0,    0,    0,    0,    0,  445,    0,    0,    0, 
            0,    0,    0,  445,  101,  445,    0,  445,  445,  445, 
      };
   }

   private static final short[] yyTable3() {
      return new short[] {

          445,  445,  445,  445,    0,  445,  445,  445,    0,  525, 
            0,  120,    0,  525,    0,    0,    0,    0,    0,    0, 
            0,   60,   60,   60,    0,    0,   60,   60,   60,    0, 
           60,    0,    0,    0,    0,    0,    0,    0,    0,  445, 
           60,  525,    0,    0,    0,    0,    0,    0,    0,   60, 
           60,    0,   60,   60,   60,   60,   60,    0,    0,    0, 
            0,    0,  524,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  524,    0,    0,    0,  445,  445,  445,    0, 
            0,    0,  445,  445,    0,  445,    0,    0,    0,    0, 
            0,    0,    0,    0,  445,    0,    0,    0,    0,    0, 
            0,   60,    0,    0,  445,  445,  100,  445,  445,  445, 
          445,  445,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  524,   60,  119,    0,  524,    0,    0,    0,    0, 
            0,  445,  445,  445,  445,  445,  445,  445,  445,  445, 
          445,  445,  445,  445,  445,    0,    0,  445,  445,  445, 
          445,    0,    0,  524,   60,    0,  445,    0,    0,    0, 
            0,    0,    0,  445,    0,  445,    0,  445,  445,  445, 
          445,  445,  445,  445,    0,  445,  445,  445,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          526,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          526,    0,    0,    0,    0,    0,    0,    0,    0,  445, 
            0,    0,  525,  525,  525,    0,    0,    0,  525,  525, 
            0,  525,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  102,    0,    0,    0,    0,    0, 
          525,  525,    0,  525,  525,  525,  525,  525,    0,  526, 
            0,  121,    0,  526,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  525,  525,  525, 
          525,  525,  525,  525,  525,  525,  525,  525,  525,  525, 
          525,  526,    0,  525,  525,  525,  525,    0,  537,    0, 
            0,    0,  525,    0,    0,    0,    0,    0,    0,  525, 
            0,  525,  527,  525,  525,  525,  525,  525,  525,  525, 
            0,  525,  527,  525,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  524,  524,  524,    0,    0,    0, 
          524,  524,    0,  524,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  525,  103,    0,    0,    0, 
            0,    0,  524,  524,    0,  524,  524,  524,  524,  524, 
            0,  527,    0,  122,    0,  527,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  524, 
          524,  524,  524,  524,  524,  524,  524,  524,  524,  524, 
          524,  524,  524,  527,    0,  524,  524,  524,  524,    0, 
          536,    0,    0,    0,  524,    0,    0,    0,    0,    0, 
            0,  524,  529,  524,    0,  524,  524,  524,  524,  524, 
          524,  524,  529,  524,    0,  524,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  526,  526,  526,    0,  105,  524,  526,  526, 
            0,  526,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  529,    0,  124,    0,  529,    0,    0,    0,    0, 
          526,  526,    0,  526,  526,  526,  526,  526,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  529,    0,    0,    0,  526,  526,  526, 
          526,  526,  526,  526,  526,  526,  526,  526,  526,  526, 
          526,    0,    0,  526,  526,  526,  526,    0,  538,    0, 
            0,    0,  526,    0,    0,    0,    0,    0,    0,  526, 
            0,  526,  528,  526,  526,  526,  526,  526,  526,  526, 
            0,  526,  528,  526,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  527,  527,  527,    0,    0,    0, 
          527,  527,    0,  527,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  526,  104,    0,    0,    0, 
            0,    0,  527,  527,    0,  527,  527,  527,  527,  527, 
            0,  528,    0,  123,    0,  528,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  527, 
          527,  527,  527,  527,  527,  527,  527,  527,  527,  527, 
          527,  527,  527,  528,    0,  527,  527,  527,  527,    0, 
          539,    0,    0,    0,  527,    0,    0,    0,    0,    0, 
            0,  527,  530,  527,    0,  527,  527,  527,  527,  527, 
          527,  527,  530,  527,    0,  527,    0,    0,    0,    0, 
            0,    0,    0,    0,  529,  529,  529,    0,    0,    0, 
          529,  529,    0,  529,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  106,  527,    0,    0, 
            0,    0,  529,  529,    0,  529,  529,  529,  529,  529, 
            0,  530,    0,  125,    0,  530,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  529, 
          529,  529,  529,  529,  529,  529,  529,  529,  529,  529, 
          529,  529,  529,  530,    0,  529,  529,  529,  529,    0, 
          541,    0,    0,    0,  529,    0,    0,    0,    0,    0, 
            0,  529,    0,  529,  521,  529,  529,  529,  529,  529, 
          529,  529,    0,  529,  521,  529,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  528,  528,  528,  529,   97,    0, 
          528,  528,    0,  528,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  521,    0,  116,    0,  521,    0,    0, 
            0,    0,  528,  528,    0,  528,  528,  528,  528,  528, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  521,    0,    0,    0,  528, 
          528,  528,  528,  528,  528,  528,  528,  528,  528,  528, 
          528,  528,  528,  520,    0,  528,  528,  528,  528,    0, 
          540,    0,    0,  520,  528,    0,    0,    0,    0,    0, 
            0,  528,    0,  528,    0,  528,  528,  528,  528,  528, 
          528,  528,    0,  528,    0,  528,    0,    0,    0,    0, 
            0,    0,    0,    0,  530,  530,  530,   96,    0,    0, 
          530,  530,    0,  530,    0,    0,    0,    0,    0,    0, 
            0,    0,  520,    0,  115,    0,  520,  528,    0,    0, 
            0,    0,  530,  530,    0,  530,  530,  530,  530,  530, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  520,    0,    0,    0,    0,  530, 
          530,  530,  530,  530,  530,  530,  530,  530,  530,  530, 
          530,  530,  530,    0,    0,  530,  530,  530,  530,    0, 
          542,    0,    0,    0,  530,    0,    0,    0,    0,    0, 
            0,  530,    0,  530,    0,  530,  530,  530,  530,  530, 
          530,  530,  523,  530,    0,  530,    0,    0,    0,    0, 
            0,    0,  523,    0,    0,    0,  521,  521,  521,    0, 
            0,    0,  521,  521,    0,  521,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  530,    0,    0, 
            0,    0,    0,    0,  521,  521,   99,  521,  521,  521, 
          521,  521,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  523,    0,  118,    0,  523,    0,    0,    0,    0, 
            0,  521,  521,  521,  521,  521,  521,  521,  521,  521, 
          521,  521,  521,  521,  521,    0,    0,  521,  521,  521, 
          521,    0,  533,  523,    0,    0,  521,    0,    0,    0, 
            0,    0,    0,  521,    0,  521,  305,  521,  521,  521, 
          521,  521,  521,  521,    0,  521,  305,  521,    0,    0, 
            0,    0,    0,    0,    0,  520,  520,  520,    0,    0, 
            0,  520,  520,    0,  520,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  521, 
          298,    0,    0,  520,  520,    0,  520,  520,  520,  520, 
          520,    0,    0,    0,    0,  305,    0,    0,    0,  251, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  251, 
          520,  520,  520,  520,  520,  520,  520,  520,  520,  520, 
          520,  520,  520,  520,    0,    0,  520,  520,  520,  520, 
            0,  532,    0,    0,    0,  520,    0,    0,    0,    0, 
            0,    0,  520,  251,  520,    0,  520,  520,  520,  520, 
          520,  520,  520,    0,  520,    0,  520,  251,  251,    0, 
            0,    0,  251,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  520,    0, 
            0,    0,    0,    0,  523,  523,  523,    0,    0,    0, 
          523,  523,    0,  523,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  523,  523,    0,  523,  523,  523,  523,  523, 
            0,    0,    0,  315,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  315,    0,    0,    0,    0,    0,  523, 
          523,  523,  523,  523,  523,  523,  523,  523,  523,  523, 
          523,  523,  523,    0,    0,  523,  523,  523,  523,    0, 
          535,    0,    0,    0,  523,    0,    0,  113,    0,    0, 
            0,  523,    0,  523,    0,  523,  523,  523,  523,  523, 
          523,  523,  315,  523,  132,  523,  315,    0,  305,  305, 
          305,    0,    0,  305,  305,  305,    0,  305,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  315,    0,    0,  523,    0,  305, 
          305,  305,  305,  305,    0,    0,    0,    0,    0,  326, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  326, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  251,  251,  251,    0,    0,  251,  251,  251,    0, 
          251,    0,    0,    0,    0,  305,    0,    0,  305,    0, 
          251,  251,    0,  112,    0,    0,    0,    0,    0,  251, 
          251,    0,  251,  251,  251,  251,  251,    0,  326,  305, 
          131,    0,  326,    0,  251,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  251,  251,  251,  251, 
          251,    0,    0,  251,  251,  251,  251,    0,  251,  251, 
          326,  305,  251,  251,    0,    0,    0,    0,  251,    0, 
            0,  251,    0,    0,  251,    0,  251,  238,  251,    0, 
          251,    0,  251,  251,  251,  251,    0,  238,  251,    0, 
          251,    0,  251,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  251,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  238,    0,    0,  251,  315,  315,  315,    0,    0, 
            0,  315,  315,    0,  315,  238,  238,    0,    0,    0, 
          238,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  315,  315,    0,  315,  315,  315,  315, 
          315,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          315,  315,  315,  315,  315,  315,  315,  315,  315,  315, 
          315,  315,  315,  315,    0,    0,  315,  315,  315,  315, 
            0,    0,    0,    0,    0,  315,    0,    0,    0,    0, 
            0,    0,  315,  239,  315,    0,  315,  315,  315,  315, 
          315,  315,  315,  239,  315,    0,  315,    0,    0,    0, 
            0,  326,  326,  326,    0,    0,    0,  326,  326,    0, 
          326,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  239,  315,  326, 
          326,    0,  326,  326,  326,  326,  326,    0,    0,    0, 
            0,  239,  239,    0,    0,    0,  239,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  326,  326,  326,  326, 
          326,  326,  326,  326,  326,  326,  326,  326,  326,  326, 
            0,    0,  326,  326,  326,  326,    0,    0,    0,    0, 
            0,  326,    0,    0,    0,    0,    0,    0,  326,    0, 
          326,    0,  326,  326,  326,  326,  326,  326,  326,    0, 
          326,    0,  326,    0,    0,    0,    0,    0,    0,  238, 
          238,  238,    0,    0,  238,  238,  238,    0,  238,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  238,  238, 
            0,    0,    0,    0,  326,    0,    0,  238,  238,    0, 
          238,  238,  238,  238,  238,    0,    0,    0,    0,    0, 
            0,    0,  238,  315,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  315,    0,  238,  238,  238,  238,  238, 
          238,  238,  238,  238,  238,    0,  238,  238,    0,    0, 
          238,  238,    0,    0,    0,    0,  238,    0,    0,  238, 
            0,    0,  238,    0,  238,    0,    0,  315,  238,    0, 
            0,    0,  238,  238,  238,  238,  238,    0,  238,    0, 
          238,    0,  315,    0,  132,    0,  315,    0,    0,    0, 
            0,    0,  238,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  238,    0,  315,  239,  239,  239,    0,    0, 
          239,  239,  239,    0,  239,    0,  259,    0,    0,    0, 
            0,    0,    0,    0,  239,  239,  259,    0,    0,    0, 
            0,    0,    0,  239,  239,    0,  239,  239,  239,  239, 
          239,    0,    0,    0,    0,    0,    0,    0,  239,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          259,  239,  239,  239,  239,  239,  239,  239,  239,  239, 
          239,    0,  239,  239,  259,  259,  239,  239,    0,  259, 
            0,    0,  239,    0,    0,  239,    0,    0,  239,    0, 
          239,    0,    0,    0,  239,    0,    0,    0,  239,  239, 
          239,  239,  239,    0,  239,    0,  239,    0,    0,    0, 
            0,    0,    0,    0,    0,  688,  622,    0,  239,  689, 
            0,    0,    0,  188,  189,    0,  190,  191,  192,  193, 
          194,  195,  196,    0,  260,  197,  198,    0,  239,    0, 
          199,  200,  201,  202,  260,    0,    0,    0,    0,    0, 
          304,    0,    0,    0,    0,    0,    0,  204,  205,    0, 
          206,  207,  208,  209,  210,  211,  212,  213,  214,  215, 
          216,    0,    0,  217,    0,  315,  315,  315,  260,    0, 
            0,  315,  315,    0,  315,    0,    0,    0,    0,    0, 
            0,    0,  260,  260,    0,    0,    0,  260,    0,    0, 
            0,    0,    0,  315,  315,    0,  315,  315,  315,  315, 
          315,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          315,  315,  315,  315,  315,  315,  315,  315,  315,  315, 
          315,  315,  315,  315,    0,    0,  315,  315,  315,  315, 
            0,    0,    0,    0,    0,  315,    0,    0,    0,    0, 
            0,    0,  315,    0,  315,    0,  315,  315,  315,  315, 
          315,  315,  315,    0,  315,    0,  315,    0,  259,  259, 
          259,    0,    0,  259,  259,  259,    0,  259,    0,  249, 
            0,    0,    0,    0,    0,    0,    0,  259,  259,  249, 
            0,    0,    0,    0,    0,    0,  259,  259,  315,  259, 
          259,  259,  259,  259,    0,    0,    0,    0,    0,    0, 
            0,  259,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  249,  259,  259,  259,  259,  259,  259, 
          259,  259,  259,  259,    0,  259,  259,  249,  249,  259, 
          259,    0,  249,    0,    0,  259,    0,    0,  259,    0, 
            0,  259,    0,  259,    0,    0,    0,  259,    0,    0, 
            0,    0,    0,  259,  259,  259,    0,  259,    0,  259, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  259,    0,    0,    0,    0,  260,  260,  260,    0, 
            0,  260,  260,  260,    0,  260,    0,  247,    0,    0, 
            0,  259,    0,    0,    0,  260,  260,  247,    0,    0, 
            0,    0,    0,    0,  260,  260,    0,  260,  260,  260, 
          260,  260,    0,    0,    0,    0,    0,    0,    0,  260, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  247,  260,  260,  260,  260,  260,  260,  260,  260, 
          260,  260,    0,  260,  260,  247,  247,  260,  260,    0, 
          247,    0,    0,  260,    0,    0,  260,    0,    0,  260, 
            0,  260,    0,    0,    0,  260,    0,    0,    0,    0, 
            0,  260,  260,  260,    0,  260,    0,  260,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  260, 
          248,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          248,    0,    0,    0,    0,    0,    0,    0,    0,  260, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  249,  249,  249,  248,    0,  249,  249,  249,    0, 
          249,    0,    0,    0,    0,    0,    0,    0,  248,  248, 
          249,  249,    0,  248,    0,    0,    0,    0,    0,  249, 
          249,    0,  249,  249,  249,  249,  249,    0,    0,    0, 
            0,    0,    0,    0,  249,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  249,  249,  249, 
          249,  249,  249,  249,  249,  249,  249,    0,  249,  249, 
            0,    0,    0,    0,    0,    0,    0,    0,  249,    0, 
            0,  249,    0,  270,  249,    0,  249,    0,    0,    0, 
          249,    0,    0,  270,    0,    0,  249,  249,  249,    0, 
          249,    0,  249,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  249,    0,    0,    0,    0,  247, 
          247,  247,    0,    0,  247,  247,  247,  270,  247,    0, 
            0,    0,    0,    0,  249,    0,    0,    0,  247,  247, 
            0,  270,  270,    0,    0,    0,  270,  247,  247,    0, 
          247,  247,  247,  247,  247,    0,    0,    0,    0,    0, 
            0,    0,  247,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  247,  247,  247,  247,  247, 
          247,  247,  247,  247,  247,    0,  247,  247,    0,    0, 
            0,    0,    0,    0,    0,    0,  247,    0,    0,  247, 
          271,    0,  247,    0,  247,    0,    0,    0,    0,    0, 
          271,    0,    0,    0,  247,  247,  247,    0,  247,    0, 
          247,    0,  248,  248,  248,    0,    0,  248,  248,  248, 
            0,  248,  247,    0,    0,    0,    0,    0,    0,    0, 
            0,  248,  248,    0,  271,    0,    0,    0,    0,    0, 
          248,  248,  247,  248,  248,  248,  248,  248,  271,  271, 
            0,    0,    0,  271,    0,  248,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  248,  248, 
          248,  248,  248,  248,  248,  248,  248,  248,    0,  248, 
          248,    0,    0,    0,    0,    0,    0,    0,    0,  248, 
            0,    0,  248,    0,    0,  248,    0,  248,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  248,  248,  248, 
            0,  248,    0,  248,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  248,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  270,  270,  270,    0,    0, 
          270,  270,  270,    0,  270,  248,  250,    0,    0,    0, 
            0,    0,    0,    0,  270,  270,  250,    0,    0,    0, 
            0,    0,    0,  270,  270,    0,  270,  270,  270,  270, 
          270,    0,    0,    0,    0,    0,    0,    0,  270,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          250,  270,  270,  270,  270,  270,  270,  270,  270,  270, 
          270,    0,  270,  270,  250,  250,    0,    0,    0,  250, 
            0,    0,  270,    0,    0,  270,    0,    0,  270,    0, 
          270,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          270,  270,    0,  252,    0,    0,  270,    0,    0,    0, 
            0,    0,    0,  252,    0,    0,    0,    0,  270,    0, 
            0,    0,  271,  271,  271,    0,    0,  271,  271,  271, 
            0,  271,    0,    0,    0,    0,    0,    0,  270,    0, 
            0,  271,  271,    0,    0,    0,    0,  252,    0,    0, 
          271,  271,    0,  271,  271,  271,  271,  271,    0,    0, 
            0,  252,  252,    0,    0,  271,  252,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  271,  271, 
          271,  271,  271,  271,  271,  271,  271,  271,    0,  271, 
          271,    0,    0,    0,  253,    0,    0,    0,    0,  271, 
            0,    0,  271,    0,  253,  271,    0,  271,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  271,  271,    0, 
            0,    0,    0,  271,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  271,    0,    0,  253,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  253,  253,    0,  271,    0,  253,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  250,  250, 
          250,    0,    0,  250,  250,  250,    0,  250,    0,    0, 
            0,  254,    0,    0,    0,    0,    0,  250,  250,    0, 
            0,  254,    0,    0,    0,    0,  250,  250,    0,  250, 
          250,  250,  250,  250,    0,    0,    0,    0,    0,    0, 
            0,  250,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  254,    0,    0,    0,    0, 
          250,  250,    0,    0,    0,  250,  250,    0,    0,  254, 
          254,    0,    0,    0,  254,  250,    0,    0,  250,    0, 
            0,  250,    0,  250,    0,  252,  252,  252,    0,    0, 
          252,  252,  252,    0,  252,    0,    0,  261,    0,  250, 
            0,    0,    0,    0,  252,  252,    0,  261,    0,    0, 
            0,  250,    0,  252,  252,    0,  252,  252,  252,  252, 
          252,    0,    0,    0,    0,    0,    0,    0,  252,    0, 
            0,  250,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  261,    0,    0,    0,    0,    0,  252,  252,    0, 
            0,    0,  252,  252,    0,  261,  261,    0,    0,    0, 
          261,    0,  252,    0,    0,  252,    0,    0,  252,    0, 
          252,    0,    0,    0,    0,    0,  253,  253,  253,    0, 
            0,  253,  253,  253,    0,  253,  252,    0,    0,    0, 
            0,    0,    0,    0,    0,  253,  253,    0,  252,    0, 
            0,    0,    0,    0,  253,  253,  255,  253,  253,  253, 
          253,  253,    0,    0,    0,    0,  255,    0,  252,  253, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  253,  253, 
            0,    0,    0,  253,  253,    0,    0,    0,    0,    0, 
          255,    0,    0,  253,    0,    0,  253,    0,    0,  253, 
            0,  253,    0,    0,  255,  255,    0,    0,    0,  255, 
            0,    0,    0,  254,  254,  254,    0,  253,  254,  254, 
          254,    0,  254,    0,    0,    0,    0,    0,    0,  253, 
            0,    0,  254,  254,    0,    0,    0,    0,    0,    0, 
            0,  254,  254,    0,  254,  254,  254,  254,  254,  253, 
            0,    0,    0,    0,    0,    0,  254,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  254,  254,    0,    0,    0, 
          254,  254,    0,    0,    0,  256,    0,    0,    0,    0, 
          254,    0,    0,  254,    0,  256,  254,    0,  254,  261, 
          261,  261,    0,    0,  261,  261,  261,    0,  261,    0, 
            0,    0,    0,    0,  254,    0,    0,    0,  261,  261, 
            0,    0,    0,    0,    0,    0,  254,  261,  261,  256, 
          261,  261,  261,  261,  261,    0,    0,    0,    0,    0, 
            0,    0,  261,  256,  256,    0,  254,    0,  256,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  261,  261,    0,    0,    0,  261,  261,    0,    0, 
            0,    0,    0,    0,    0,    0,  261,    0,    0,  261, 
            0,    0,  261,    0,  261,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  262,    0,    0,    0,    0,    0, 
          261,    0,    0,    0,  262,    0,    0,    0,  255,  255, 
          255,    0,  261,  255,  255,  255,    0,  255,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  255,  255,    0, 
            0,    0,  261,    0,    0,    0,  255,  255,  262,  255, 
          255,  255,  255,  255,    0,    0,    0,    0,    0,    0, 
            0,  255,  262,  262,    0,    0,    0,  262,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          255,  255,    0,    0,    0,  255,  255,    0,    0,    0, 
          235,    0,    0,    0,    0,  255,    0,    0,  255,    0, 
          235,  255,    0,  255,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  255, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  255,    0,    0,  235,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  235,  235, 
            0,  255,    0,  235,    0,    0,    0,  256,  256,  256, 
            0,    0,  256,  256,  256,    0,  256,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  256,  256,    0,    0, 
            0,    0,    0,    0,    0,  256,  256,  278,  256,  256, 
          256,  256,  256,    0,    0,    0,    0,  278,    0,    0, 
          256,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  256, 
          256,    0,    0,    0,  256,  256,    0,    0,    0,    0, 
            0,  278,    0,    0,  256,    0,    0,  256,    0,    0, 
          256,    0,  256,    0,    0,  278,  278,    0,    0,  264, 
            0,    0,    0,    0,    0,    0,    0,    0,  256,  264, 
            0,    0,    0,    0,    0,    0,  262,  262,  262,    0, 
          256,  262,  262,  262,    0,  262,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  262,  262,    0,    0,    0, 
          256,    0,    0,  264,  262,  262,    0,  262,  262,  262, 
          262,  262,    0,    0,    0,    0,    0,  264,  264,  262, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  262, 
            0,    0,    0,  262,  262,    0,    0,    0,  227,    0, 
            0,    0,    0,  262,    0,    0,  262,    0,  227,  262, 
            0,  262,  235,  235,  235,    0,    0,  235,  235,  235, 
            0,  235,    0,    0,    0,    0,    0,  262,    0,    0, 
            0,  235,  235,    0,    0,    0,    0,    0,    0,  262, 
          235,  235,  227,  235,  235,  235,  235,  235,    0,    0, 
            0,    0,    0,    0,    0,  235,  227,  227,    0,  262, 
          272,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          272,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  235, 
            0,    0,  235,    0,    0,  235,    0,  235,    0,    0, 
            0,    0,    0,    0,  272,    0,    0,    0,    0,  278, 
          278,  278,    0,  235,  278,  278,  278,    0,  278,  272, 
            0,    0,    0,    0,    0,  235,    0,    0,  278,  278, 
            0,    0,    0,    0,    0,    0,    0,  278,  278,    0, 
          278,  278,  278,  278,  278,  235,    0,    0,    0,  297, 
            0,    0,  278,    0,    0,    0,    0,    0,    0,  297, 
            0,    0,    0,    0,    0,  303,    0,    0,    0,    0, 
            0,  264,  264,  264,    0,  303,  264,  264,  264,    0, 
          264,    0,    0,    0,    0,  299,  278,    0,    0,  278, 
          264,  264,  278,  297,  278,  299,    0,    0,    0,  264, 
          264,    0,  264,  264,  264,  264,  264,    0,  297,  299, 
          278,    0,    0,    0,  264,    0,    0,    0,    0,    0, 
            0,    0,  278,    0,  303,    0,    0,    0,    0,  299, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  278,    0,  299,    0,    0,    0,  264,    0, 
            0,  264,    0,    0,  264,    0,  264,    0,    0,    0, 
          227,  227,  227,    0,    0,  227,  227,  227,    0,  227, 
            0,   68,  264,    0,    0,    0,    0,    0,    0,  227, 
          227,   68,    0,    0,  264,    0,    0,    0,  227,  227, 
            0,  227,  227,  227,  227,  227,    0,    0,    0,    0, 
            0,    0,    0,  227,  264,    0,    0,   66,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,   66,    0,    0, 
            0,    0,  272,  272,  272,    0,    0,  272,  272,  272, 
           68,  272,    0,    0,    0,    0,    0,  227,    0,    0, 
          227,  272,  272,  227,    0,  227,    0,    0,    0,    0, 
          272,  272,    0,  272,  272,  272,  272,  272,    0,    0, 
            0,  227,    0,    0,    0,  272,   66,    0,    0,    0, 
            0,    0,    0,  227,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,   61,    0,  227,    0,    0,    0,    0,    0,  272, 
            0,   61,  272,    0,    0,  272,    0,  272,    0,    0, 
            0,  297,  297,  297,    0,    0,  297,  297,  297,    0, 
          297,    0,    0,  272,    0,    0,    0,  303,  303,  303, 
          297,  297,  303,  303,  303,  272,  303,    0,    0,  297, 
          297,  304,  297,  297,  297,  297,  297,  299,  299,  299, 
           61,  304,  299,  299,  299,  272,  299,    0,  303,  303, 
          303,  303,  303,    0,    0,    0,  299,  299,    0,    0, 
            0,    0,    0,    0,    0,  299,  299,   54,  299,  299, 
          299,  299,  299,    0,    0,  300,    0,   54,    0,    0, 
            0,  297,    0,    0,  297,    0,  297,    0,    0,    0, 
          304,    0,    0,    0,  303,    0,    0,  303,    0,    0, 
            0,    0,  297,  302,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  302,    0,    0,    0,  299,  303,    0, 
          299,   36,  299,   68,   68,   68,   54,    0,   68,   68, 
           68,   36,   68,    0,  297,    0,    0,    0,  299,   25, 
            0,    0,   68,   68,    0,    0,    0,  297,    0,   25, 
          303,   68,   68,    0,   68,   68,   68,   68,   68,   66, 
           66,   66,  302,    0,   66,   66,   66,    0,   66,    0, 
          299,    0,    0,    0,    0,    0,    0,    0,   66,   66, 
           36,    0,    0,    0,    0,    0,    0,   66,   66,    0, 
           66,   66,   66,   66,   66,    0,    0,    0,   25,    0, 
            0,    0,    0,   68,    0,    0,   68,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,   68,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,   66, 
            0,    0,   66,   61,   61,   61,    0,    0,   61,   61, 
           61,    0,   61,    0,    0,    0,   68,    0,    0,    0, 
           66,    0,   61,  308,    0,    0,    0,    0,    0,    0, 
            0,   61,   61,    0,   61,   61,   61,   61,   61,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,   66,  304,  304,  304,    0,    0,  304,  304, 
          304,    0,  304,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,   61,  304,  304,  304,  304,  304,   54, 
           54,   54,    0,    0,   54,   54,   54,    0,   54,    0, 
            0,    0,    0,    0,   61,    0,    0,    0,   54,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
           54,   54,   54,   54,   54,  302,  302,  302,    0,    0, 
          304,  302,  302,  304,  302,    0,   61,    0,    0,    0, 
            0,    0,    0,   36,   36,   36,    0,    0,    0,   36, 
           36,    0,   36,    0,  304,    0,  302,  302,  302,  302, 
          302,   25,   25,   25,  133,    0,    0,   25,   25,   54, 
           25,    0,    0,    0,   36,   36,   36,   36,   36,    0, 
            0,    0,    0,    0,    0,    0,  304,    0,    0,    0, 
           54,    0,   25,   25,   25,   25,   25,    0,    0,    0, 
            0,    0,    0,    0,    0,  302,    0,    0,    0,    0, 
            0,    0,    0,  133,    0,    0,    0,    0,    0,    0, 
            0,    0,   54,   36,    0,    0,  302,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,   25,    0,    0,   36,  519,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  302,    0, 
            0,    0,   25,    0,    0,    0,    0,    0,    0,    0, 
            4,    5,    6,    0,    8,    0,   36,    0,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,   15,   16, 
           17,   18,    0,    0,   25,    0,   19,   20,   21,  239, 
          240,  241,  242,  134,    0,  118,    0,    0,    0,    0, 
            0,    0,   28,    0,    0,  243,  244,  245,    0,  246, 
           35,  247,  248,  249,  250,  285,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          251,    0,  134,  131,    0,    0,   46,   47,    0,   48, 
            0,  286,    0,  287,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,  288,    0,    0,    0,    0,   52,  289, 
           54,   55,   56,   57,  522,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,   65,   66,  290,    0,    0, 
          133,  133,  133,  133,  133,  133,  133,  133,  133,  133, 
          133,    0,    0,  133,  133,    0,  133,  133,  133,  133, 
          133,  133,  133,    0,  622,    0,    0,  133,  133,  133, 
          133,  133,  133,  133,  133,    0,  133,    0,    0,    0, 
            0,    0,  133,  133,    0,  133,  133,  133,  133,    0, 
          133,  133,  133,  133,  133,  133,    0,  133,  133,  133, 
          133,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  519,    0,    0,    0,    0,    0,    0,  519, 
          519,  133,    0,  133,  133,  622,    0,  133,  133,    0, 
          133,    0,  133,    0,    0,    0,  133,    0,    0,    0, 
            0,  133,    0,    0,  133,    0,  622,    0,    0,  133, 
          133,  133,  133,  133,  133,  629,    0,    0,  133,    0, 
          133,  133,    0,  133,  133,  133,  133,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  133,  133,    0,  134, 
          134,  134,  134,  134,  134,  134,  134,  134,  134,  134, 
            0,    0,  134,  134,    0,  134,  134,  134,  134,  134, 
          134,  134,    0,  623,    0,    0,  134,  134,  134,  134, 
          134,  134,  134,  134,    0,  134,    0,    0,    0,    0, 
            0,  134,  134,    0,  134,  134,  134,  134,    0,  134, 
          134,  134,  134,  134,  134,    0,  134,  134,  134,  134, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  522,    0,    0,    0,    0,    0,    0,  522,  522, 
          134,    0,  134,  134,  623,    0,  134,  134,    0,  134, 
            0,  134,    0,    0,    0,  134,    0,    0,    0,    0, 
          134,    0,    0,  134,    0,  623,    0,    0,  134,  134, 
          134,  134,  134,  134,  325,    0,    0,  134,    0,  134, 
          134,    0,  134,  134,  134,  134,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  134,  134,    0,    0,    0, 
          133,  133,  133,  133,  133,  133,  133,  133,  133,  133, 
          133,    0,    0,  133,  133,    0,  133,  133,  133,  133, 
          133,  133,  133,    0,  629,    0,    0,  133,  133,  133, 
          133,  133,  133,  133,  134,    0,  133,    0,    0,    0, 
            0,    0,  133,  133,    0,  133,  133,  133,  133,    0, 
          133,  133,  133,  133,  133,  133,    0,  133,  133,  133, 
          133,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  629,    0,    0,    0,    0,    0,    0,  629, 
          629,  133,    0,  134,  133,  625,    0,  133,  133,    0, 
          133,    0,  133,    0,    0,    0,  133,    0,    0,    0, 
            0,  133,    0,    0,  133,    0,  629,    0,    0,  133, 
          133,  133,  133,  133,  133,  326,    0,    0,  133,    0, 
          133,  133,    0,  133,  133,  133,  133,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  133,  133,    0,  134, 
          134,  134,  134,  134,  134,  134,  134,  134,  134,  134, 
            0,    0,  134,  134,    0,  134,  134,  134,  134,  134, 
          134,  134,    0,    0,    0,    0,  134,  134,  134,  134, 
          134,  134,  134,  369,    0,  134,    0,    0,    0,    0, 
            0,  134,  134,    0,  134,  134,  134,  134,    0,  134, 
          134,  134,  134,  134,  134,    0,  134,  134,  134,  134, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  325,    0,    0,    0,    0,    0,    0,  325,  325, 
          134,    0,  369,  134,  626,    0,  134,  134,    0,  134, 
            0,  134,    0,    0,    0,  134,    0,    0,    0,    0, 
          134,    0,    0,  134,    0,    0,    0,    0,  134,  134, 
          134,  134,  134,  134,    0,    0,    0,  134,    0,  134, 
          134,    0,  134,  134,  134,  134,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  134,  134,    0,    0,    0, 
          134,  134,  134,  134,  134,  134,  134,  134,  134,  134, 
          134,    0,    0,  134,  134,    0,  134,  134,  134,  134, 
          134,  134,  134,    0,    0,    0,    0,  134,  134,  134, 
          134,  134,  134,  134,  317,    0,  134,    0,    0,    0, 
            0,    0,  134,  134,    0,  134,  134,  134,  134,    0, 
          134,  134,  134,  134,  134,  134,    0,  134,  134,  134, 
          134,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  326,    0,    0,    0,    0,    0,    0,  326, 
          326,  134,    0,  317,  134,    0,    0,  134,  134,    0, 
          134,    0,  134,    0,    0,    0,  134,    0,    0,    0, 
            0,  134,    0,    0,  134,    0,    0,    0,    0,  134, 
          134,  134,  134,  134,  134,    0,    0,    0,  134,    0, 
          134,  134,    0,  134,  134,  134,  134,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  134,  134,    0,  369, 
          369,  369,  369,  369,  369,  369,  369,  369,  369,  369, 
            0,  369,  369,  369,  369,  369,  369,  369,  369,  369, 
          369,  369,    0,    0,    0,    0,  369,  369,  369,  369, 
          369,  369,  369,  547,    0,  369,    0,    0,    0,    0, 
            0,  369,  369,    0,  369,  369,  369,  369,    0,  369, 
          369,  369,  369,  369,  369,    0,  369,  369,  369,  369, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          369,    0,  547,  369,    0,    0,  369,  369,    0,  369, 
            0,  369,    0,    0,    0,  369,    0,    0,    0,    0, 
            0,    0,    0,  369,    0,    0,    0,    0,  369,  369, 
          369,  369,  369,  369,    0,    0,    0,  369,    0,  369, 
          369,    0,  369,  369,  369,  369,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  369,  369,    0,    0,    0, 
          317,  317,  317,  317,  317,  317,  317,  317,  317,  317, 
          317,    0,    0,  317,  317,    0,  317,  317,  317,  317, 
          317,  317,  317,    0,    0,    0,    0,  317,  317,  317, 
          317,  317,  317,  317,  651,    0,  317,    0,    0,    0, 
            0,    0,  317,  317,    0,  317,  317,  317,  317,    0, 
          317,  317,  317,  317,  317,  317,    0,  317,  317,  317, 
          317,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  317,    0,  651,  317,    0,    0,  317,  317,    0, 
          317,    0,  317,    0,    0,    0,  317,    0,    0,    0, 
            0,    0,    0,    0,  317,    0,   18,    0,    0,  317, 
          317,  317,  317,  317,  317,    0,    0,    0,  317,    0, 
          317,  317,    0,  317,  317,  317,  317,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  317,  317,    0,  547, 
          547,  547,  547,  547,  547,  547,  547,  547,  547,  547, 
            0,    0,  547,  547,    0,  547,  547,  547,  547,  547, 
          547,  547,    0,    0,  651,    0,  547,  547,  547,  547, 
          547,  547,  547,    0,    0,  547,    0,    0,    0,    0, 
            0,  547,  547,    0,  547,  547,  547,  547,    0,  547, 
          547,  547,  547,  547,  547,    0,  547,  547,  547,  547, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  651,    0,    0,    0,    0,    0,    0, 
          547,    0,    0,  547,    0,    0,  547,  547,    0,  547, 
            0,  547,    0,    0,    0,  547,    0,    0,    0,    0, 
            0,    0,    0,  547,    0,    0,    0,    0,  547,  547, 
          547,  547,  547,  547,    0,    0,    0,  547,    0,  547, 
          547,    0,  547,  547,  547,  547,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  547,  547,    0,    0,    0, 
          651,  651,  651,  651,  651,  651,  651,  651,  651,  651, 
          651,    0,    0,  651,  651,    0,  651,  651,  651,  651, 
          651,  651,  651,  651,    0,    0,    0,  651,  651,  651, 
          651,  651,  651,  651,    0,    0,  651,    0,    0,    0, 
            0,    0,  651,  651,    0,  651,  651,  651,  651,    0, 
          651,  651,  651,  651,  651,  651,    0,  651,  651,  651, 
          651,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  651,    0,    0,    0,    0,    0,    0,    0, 
            0,  651,    0,    0,  651,    0,    0,  651,  651,    0, 
          651,    0,  651,    0,    0,    0,  651,    0,    0,    0, 
            0,    0,    0,    0,  651,    0,    0,    0,    0,  651, 
          651,  651,  651,  651,  651,    0,    0,    0,  651,    0, 
      };
   }

   private static final short[] yyTable4() {
      return new short[] {

          651,  651,    0,  651,  651,  651,  651,    0,    0,    0, 
          651,  651,  651,  651,  651,  651,  651,  651,  651,  651, 
          651,    0,    0,    0,  651,    0,  651,  651,  651,  651, 
          651,  651,  651,  319,    0,    0,    0,  651,  651,  651, 
          651,  651,  651,  651,    0,    0,  651,    0,    0,    0, 
            0,    0,  651,  651,    0,  651,  651,  651,  651,    0, 
          651,  651,  651,  651,  651,  651,    0,  651,  651,  651, 
          651,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  651,    0,    0,  651,    0,    0,  651,  651,    0, 
          651,    0,  651,    0,    0,    0,  651,    0,    0,    0, 
            0,    0,    0,    0,  651,    0,    0,    0,    0,  651, 
          651,  651,  651,  651,  651,    0,    0,    0,  651,    0, 
          651,  651,    0,  651,  651,  651,  651,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  651,  651,    0,  651, 
          651,  651,  651,  651,  651,    0,    0,    0,  651,  651, 
            0,    0,  260,  651,    0,  651,  651,  651,  651,  651, 
          651,  651,    0,    0,    0,    0,  651,  651,  651,  651, 
          651,  651,  651,    0,    0,  651,    0,    0,    0,    0, 
            0,  651,  651,    0,  651,  651,  651,  651,    0,  651, 
          651,  651,  651,  651,  651,    0,  651,  651,  651,  651, 
            0,  259,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          651,    0,    0,  651,    0,    0,  651,  651,    0,  651, 
            0,  651,    0,    0,    0,  651,    0,    0,    0,    0, 
            0,    0,    0,  651,    0,    0,  651,    0,  651,  651, 
          651,  651,  651,  651,    0,    0,    0,  651,    0,  651, 
          651,    0,  651,  651,  651,  651,    0,    0,    0,    0, 
          321,  321,  321,  321,  321,  651,  651,    0,  321,  321, 
            0,  308,    0,  321,    0,  321,  321,  321,  321,  321, 
          321,  321,    0,    0,    0,    0,  321,  321,  321,  321, 
          321,  321,  321,    0,    0,  321,    0,    0,    0,    0, 
            0,  321,  321,    0,  321,  321,  321,  321,    0,  321, 
          321,  321,  321,  321,  321,    0,  321,  321,  321,  321, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          321,    0,    0,  321,    0,  319,  321,  321,    0,  321, 
            0,  321,    0,    0,    0,  321,    0,    0,    0,    0, 
            0,    0,    0,  321,    0,    0,    0,    0,  321,  321, 
          321,  321,  321,  321,    0,    0,    0,  321,    0,  321, 
          321,    0,  321,  321,  321,  321,    0,    0,    0,    4, 
            5,    6,    0,    8,    0,  321,  321,    9,   10,    0, 
            0,    0,   11,    0,   12,   13,   14,   15,   16,   17, 
           18,    0,    0,    0,    0,   19,   20,   21,  239,  240, 
          241,  242,    0,    0,   26,    0,    0,    0,    0,    0, 
            0,   28,    0,    0,  243,  244,  245,    0,  246,   35, 
          247,  248,  249,  250,  353,   40,   41,   42,   43,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  251, 
            0,    0,  131,    0,    0,   46,   47,    0,   48,    0, 
            0,    0,    0,    0,   50,    0,    0,    0,    0,    0, 
            0,    0,   51,    0,    0,    0,    0,   52,   53,   54, 
           55,   56,   57,    0,    0,    0,   58,    0,   59,   60, 
            0,   61,   62,   63,   64,    0,    0,    0,    4,    5, 
            6,    0,    8,    0,   65,   66,    9,   10,    0,    0, 
            0,   11,    0,   12,   13,   14,   15,   16,   17,   18, 
            0,    0,    0,    0,   19,   20,   21,  239,  240,  241, 
          242,    0,    0,   26,    0,    0,    0,    0,    0,    0, 
           28,    0,    0,  243,  244,  245,    0,  246,   35,  247, 
          248,  249,  250,  655,   40,   41,   42,   43,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  251,    0, 
            0,  131,    0,    0,   46,   47,    0,   48,    0,    0, 
            0,    0,    0,   50,    0,    0,    0,    0,    0,    0, 
            0,   51,    0,    0,    0,    0,   52,   53,   54,   55, 
           56,   57,    0,    0,    0,   58,    0,   59,   60,    0, 
           61,   62,   63,   64,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,   65,   66,    0,    0,    4,    5,    6, 
            0,    8,    0,    0,    0,    9,   10,    0,    0,    0, 
           11,    0,   12,   13,   14,  109,  110,   17,   18,    0, 
            0,    0,    0,  111,  112,  113,  340,  341,  342,  343, 
            0,    0,  118,    0,    0,    0,    0,    0,    0,  119, 
            0,    0,  344,  345,  346,    0,  347,   35,  348,  349, 
          350,  351,    0,   40,    0,    0,  128,  471,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  352,    0,    0, 
          131,    0,    0,   46,   47,    0,   48,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,   52,   53,   54,   55,   56, 
           57,    0,    0,    0,   58,    0,   59,   60,    0,   61, 
           62,   63,   64,    0,    0,    0,    4,    5,    6,    0, 
            8,    0,   65,   66,    9,   10,    0,    0,    0,   11, 
            0,   12,   13,   14,  109,  110,   17,   18,    0,    0, 
            0,    0,  111,  112,  113,  340,  341,  342,  343,    0, 
            0,  118,    0,    0,    0,    0,    0,    0,  119,    0, 
            0,  344,  345,  346,    0,  347,   35,  348,  349,  350, 
          351,    0,   40,    0,    0,  128,  471,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  352,    0,    0,  131, 
            0,    0,   46,   47,    0,   48,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,   52,   53,   54,   55,   56,   57, 
            0,    0,    0,   58,    0,   59,   60,    0,   61,   62, 
           63,   64,    0,    0,    0,    4,    5,    6,    7,    8, 
            0,   65,   66,    9,   10,    0,    0,    0,   11,    0, 
           12,   13,   14,   15,   16,   17,   18,    0,    0,    0, 
            0,   19,   20,   21,   22,   23,   24,   25,    0,    0, 
           26,    0,    0,    0,    0,    0,   27,   28,   29,   30, 
           31,   32,   33,    0,   34,   35,   36,   37,   38,   39, 
            0,   40,   41,   42,   43,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,   44,    0,    0,   45,    0, 
            0,   46,   47,    0,   48,    0,   49,    0,    0,    0, 
           50,    0,    0,    0,    0,    0,    0,    0,   51,    0, 
            0,    0,    0,   52,   53,   54,   55,   56,   57,    0, 
            0,    0,   58,    0,   59,   60,    0,   61,   62,   63, 
           64,    0,    0,    0,    4,    5,    6,    7,  321,    0, 
           65,   66,    9,   10,    0,    0,    0,   11,    0,   12, 
           13,   14,   15,   16,   17,   18,    0,    0,    0,    0, 
           19,   20,   21,   22,   23,   24,   25,    0,    0,   26, 
            0,    0,    0,    0,    0,   27,   28,    0,   30,   31, 
           32,   33,    0,   34,   35,   36,   37,   38,   39,    0, 
           40,   41,   42,   43,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,   44,    0,    0,   45,    0,    0, 
           46,   47,    0,   48,    0,   49,    0,    0,    0,   50, 
            0,    0,    0,    0,    0,    0,    0,   51,    0,    0, 
            0,    0,   52,   53,   54,   55,   56,   57,    0,    0, 
            0,   58,    0,   59,   60,    0,   61,   62,   63,   64, 
            0,    0,    3,    4,    5,    6,    7,    8,    0,   65, 
           66,    9,   10,    0,    0,    0,   11,    0,   12,   13, 
           14,   15,   16,   17,   18,    0,    0,    0,    0,   19, 
           20,   21,   22,   23,   24,   25,    0,    0,   26,    0, 
            0,    0,    0,    0,   27,   28,   29,   30,   31,   32, 
           33,    0,   34,   35,   36,   37,   38,   39,    0,   40, 
           41,   42,   43,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,   44,    0,    0,   45,    0,    0,   46, 
           47,    0,   48,    0,   49,    0,    0,    0,   50,    0, 
            0,    0,    0,    0,    0,    0,   51,    0,    0,    0, 
            0,   52,   53,   54,   55,   56,   57,    0,    0,    0, 
           58,    0,   59,   60,    0,   61,   62,   63,   64,    0, 
            0,  320,    4,    5,    6,    7,  321,    0,   65,   66, 
            9,   10,    0,    0,    0,   11,    0,   12,   13,   14, 
           15,   16,   17,   18,    0,    0,    0,    0,   19,   20, 
           21,   22,   23,   24,   25,    0,    0,   26,    0,    0, 
            0,    0,    0,   27,   28,    0,   30,   31,   32,   33, 
            0,   34,   35,   36,   37,   38,   39,    0,   40,   41, 
           42,   43,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,   44,    0,    0,  322,    0,    0,   46,   47, 
            0,   48,    0,   49,    0,    0,    0,   50,    0,    0, 
            0,    0,    0,    0,    0,   51,    0,    0,    0,    0, 
           52,   53,   54,   55,   56,   57,    0,    0,    0,   58, 
            0,   59,   60,    0,   61,   62,   63,   64,    0,    0, 
          320,    4,    5,    6,    7,  321,    0,   65,   66,    9, 
           10,    0,    0,    0,   11,    0,   12,   13,   14,   15, 
           16,   17,   18,    0,    0,    0,    0,   19,   20,   21, 
           22,   23,   24,   25,    0,    0,   26,    0,    0,    0, 
            0,    0,   27,   28,    0,   30,   31,   32,   33,    0, 
           34,   35,   36,   37,   38,   39,    0,   40,   41,   42, 
           43,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,   44,    0,    0,   45,    0,    0,   46,   47,    0, 
           48,    0,   49,    0,    0,    0,   50,    0,    0,    0, 
            0,    0,    0,    0,   51,    0,    0,    0,    0,   52, 
           53,   54,   55,   56,   57,    0,    0,    0,   58,    0, 
           59,   60,    0,   61,   62,   63,   64,    0,    0,    0, 
            4,    5,    6,    0,    8,    0,   65,   66,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,   15,   16, 
           17,   18,    0,    0,    0,    0,   19,   20,   21,  239, 
          240,  241,  242,    0,    0,  118,    0,    0,    0,    0, 
            0,    0,   28,    0,    0,  243,  244,  245,    0,  246, 
           35,  247,  248,  249,  250,  285,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          251,    0,    0,  131,    0,    0,   46,   47,    0,   48, 
            0,  286,    0,  287,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,  288,    0,    0,    0,    0,   52,  289, 
           54,   55,   56,   57,    0,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,    0, 
            4,    5,    6,    0,    8,   65,   66,  290,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,  109,  110, 
           17,   18,    0,    0,    0,    0,  111,  112,  113,  239, 
          240,  241,  242,    0,    0,  118,    0,    0,    0,    0, 
            0,    0,   28,    0,    0,  243,  244,  245,    0,  246, 
           35,  247,  248,  249,  250,  285,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          251,    0,    0,  131,    0,    0,   46,   47,    0,   48, 
            0,  708,    0,  287,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,  288,    0,    0,    0,    0,   52,  289, 
           54,   55,   56,   57,    0,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,    0, 
          292,  292,  292,    0,  292,   65,   66,  290,  292,  292, 
            0,    0,    0,  292,    0,  292,  292,  292,  292,  292, 
          292,  292,    0,    0,    0,    0,  292,  292,  292,  292, 
          292,  292,  292,    0,    0,  292,    0,    0,    0,    0, 
            0,    0,  292,    0,    0,  292,  292,  292,    0,  292, 
          292,  292,  292,  292,  292,  292,  292,  292,  292,  292, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          292,    0,    0,  292,    0,    0,  292,  292,    0,  292, 
            0,  292,    0,  292,    0,  292,    0,    0,    0,    0, 
            0,    0,    0,  292,    0,    0,    0,    0,  292,  292, 
          292,  292,  292,  292,    0,    0,    0,  292,    0,  292, 
          292,    0,  292,  292,  292,  292,    0,    0,    0,    0, 
            4,    5,    6,    0,    8,  292,  292,  292,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,  109,  110, 
           17,   18,    0,    0,    0,    0,  111,  112,  113,  239, 
          240,  241,  242,    0,    0,  118,    0,    0,    0,    0, 
            0,    0,   28,    0,    0,  243,  244,  245,    0,  246, 
           35,  247,  248,  249,  250,  285,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          251,    0,    0,  131,    0,    0,   46,   47,    0,   48, 
            0,  286,    0,    0,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,  288,    0,    0,    0,    0,   52,  289, 
           54,   55,   56,   57,    0,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,    0, 
            4,    5,    6,    0,    8,   65,   66,  290,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,  109,  110, 
           17,   18,    0,    0,    0,    0,  111,  112,  113,  239, 
          240,  241,  242,    0,    0,  118,    0,    0,    0,    0, 
            0,    0,   28,    0,    0,  243,  244,  245,    0,  246, 
           35,  247,  248,  249,  250,  285,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          251,    0,    0,  131,    0,    0,   46,   47,    0,   48, 
            0,    0,    0,  287,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,  288,    0,    0,    0,    0,   52,  289, 
           54,   55,   56,   57,    0,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,    0, 
            4,    5,    6,    0,    8,   65,   66,  290,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,  109,  110, 
           17,   18,    0,    0,    0,    0,  111,  112,  113,  239, 
          240,  241,  242,    0,    0,  118,    0,    0,    0,    0, 
            0,    0,   28,    0,    0,  243,  244,  245,    0,  246, 
           35,  247,  248,  249,  250,  285,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          251,    0,    0,  131,    0,    0,   46,   47,    0,   48, 
            0,  708,    0,    0,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,  288,    0,    0,    0,    0,   52,  289, 
           54,   55,   56,   57,    0,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,    0, 
            4,    5,    6,    0,    8,   65,   66,  290,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,  109,  110, 
           17,   18,    0,    0,    0,    0,  111,  112,  113,  239, 
          240,  241,  242,    0,    0,  118,    0,    0,    0,    0, 
            0,    0,   28,    0,    0,  243,  244,  245,    0,  246, 
           35,  247,  248,  249,  250,  285,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          251,    0,    0,  131,    0,    0,   46,   47,    0,   48, 
            0,    0,    0,    0,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,  288,    0,    0,    0,    0,   52,  289, 
           54,   55,   56,   57,    0,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,    0, 
            4,    5,    6,    7,    8,   65,   66,  290,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,   15,   16, 
           17,   18,    0,    0,    0,    0,   19,   20,   21,   22, 
           23,   24,   25,    0,    0,   26,    0,    0,    0,    0, 
            0,   27,   28,   29,   30,   31,   32,   33,    0,   34, 
           35,   36,   37,   38,   39,    0,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
           44,    0,    0,   45,    0,    0,   46,   47,    0,   48, 
            0,   49,    0,    0,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,   51,    0,    0,    0,    0,   52,   53, 
           54,   55,   56,   57,    0,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,    4, 
            5,    6,    7,    8,    0,   65,   66,    9,   10,    0, 
            0,    0,   11,    0,   12,   13,   14,   15,   16,   17, 
           18,    0,    0,    0,    0,   19,   20,   21,   22,   23, 
           24,   25,    0,    0,   26,    0,    0,    0,    0,    0, 
           27,   28,    0,   30,   31,   32,   33,    0,   34,   35, 
           36,   37,   38,   39,    0,   40,   41,   42,   43,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,   44, 
            0,    0,   45,    0,    0,   46,   47,    0,   48,    0, 
           49,    0,    0,    0,   50,    0,    0,    0,    0,    0, 
            0,    0,   51,    0,    0,    0,    0,   52,   53,   54, 
           55,   56,   57,    0,    0,    0,   58,    0,   59,   60, 
            0,   61,   62,   63,   64,    0,    0,    0,    4,    5, 
            6,    0,    8,    0,   65,   66,    9,   10,    0,    0, 
            0,   11,    0,   12,   13,   14,  109,  110,   17,   18, 
            0,    0,    0,    0,  111,  112,  113,  239,  240,  241, 
          242,    0,    0,  118,    0,    0,    0,    0,    0,    0, 
           28,    0,    0,  243,  244,  245,    0,  246,   35,  247, 
          248,  249,  250,    0,   40,   41,   42,   43,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  251,    0, 
            0,  131,  507,    0,   46,   47,    0,   48,    0,    0, 
            0,    0,    0,   50,    0,    0,    0,    0,    0,    0, 
            0,  288,    0,    0,    0,    0,   52,   53,   54,   55, 
           56,   57,    0,    0,    0,   58,    0,   59,   60,    0, 
           61,   62,   63,   64,    0,    0,    0,    4,    5,    6, 
            0,    8,    0,   65,   66,    9,   10,    0,    0,    0, 
           11,    0,   12,   13,   14,   15,   16,   17,   18,    0, 
            0,    0,    0,   19,   20,   21,  239,  240,  241,  242, 
            0,    0,  118,    0,    0,    0,    0,    0,    0,   28, 
            0,    0,  243,  244,  245,    0,  246,   35,  247,  248, 
          249,  250,    0,   40,   41,   42,   43,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,  251,    0,    0, 
          131,    0,    0,   46,   47,    0,   48,    0,  647,    0, 
            0,    0,   50,    0,    0,    0,    0,    0,    0,    0, 
          288,    0,    0,    0,    0,   52,   53,   54,   55,   56, 
           57,    0,    0,    0,   58,    0,   59,   60,    0,   61, 
           62,   63,   64,    0,    0,    0,    4,    5,    6,    0, 
            8,    0,   65,   66,    9,   10,    0,    0,    0,   11, 
            0,   12,   13,   14,  109,  110,   17,   18,    0,    0, 
            0,    0,  111,  112,  113,  239,  240,  241,  242,    0, 
            0,  118,    0,    0,    0,    0,    0,    0,   28,    0, 
            0,  243,  244,  245,    0,  246,   35,  247,  248,  249, 
          250,    0,   40,   41,   42,   43,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  251,    0,    0,  131, 
            0,    0,   46,   47,    0,   48,    0,  286,    0,    0, 
            0,   50,    0,    0,    0,    0,    0,    0,    0,  288, 
            0,    0,    0,    0,   52,   53,   54,   55,   56,   57, 
            0,    0,    0,   58,    0,   59,   60,    0,   61,   62, 
           63,   64,    0,    0,    0,    4,    5,    6,    0,    8, 
            0,   65,   66,    9,   10,    0,    0,    0,   11,    0, 
           12,   13,   14,  109,  110,   17,   18,    0,    0,    0, 
            0,  111,  112,  113,  239,  240,  241,  242,    0,    0, 
          118,    0,    0,    0,    0,    0,    0,   28,    0,    0, 
          243,  244,  245,    0,  246,   35,  247,  248,  249,  250, 
            0,   40,   41,   42,   43,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  251,    0,    0,  131,    0, 
            0,   46,   47,    0,   48,    0,  647,    0,    0,    0, 
           50,    0,    0,    0,    0,    0,    0,    0,  288,    0, 
            0,    0,    0,   52,   53,   54,   55,   56,   57,    0, 
            0,    0,   58,    0,   59,   60,    0,   61,   62,   63, 
           64,    0,    0,    0,    4,    5,    6,    0,    8,    0, 
           65,   66,    9,   10,    0,    0,    0,   11,    0,   12, 
           13,   14,  109,  110,   17,   18,    0,    0,    0,    0, 
          111,  112,  113,  239,  240,  241,  242,    0,    0,  118, 
            0,    0,    0,    0,    0,    0,   28,    0,    0,  243, 
          244,  245,    0,  246,   35,  247,  248,  249,  250,    0, 
           40,   41,   42,   43,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  251,    0,    0,  131,    0,    0, 
           46,   47,    0,   48,    0,  927,    0,    0,    0,   50, 
            0,    0,    0,    0,    0,    0,    0,  288,    0,    0, 
            0,    0,   52,   53,   54,   55,   56,   57,    0,    0, 
            0,   58,    0,   59,   60,    0,   61,   62,   63,   64, 
            0,    0,    0,    4,    5,    6,    0,    8,    0,   65, 
           66,    9,   10,    0,    0,    0,   11,    0,   12,   13, 
           14,  109,  110,   17,   18,    0,    0,    0,    0,  111, 
          112,  113,  239,  240,  241,  242,    0,    0,  118,    0, 
            0,    0,    0,    0,    0,   28,    0,    0,  243,  244, 
          245,    0,  246,   35,  247,  248,  249,  250,    0,   40, 
           41,   42,   43,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  251,    0,    0,  131,    0,    0,   46, 
           47,    0,   48,    0,  708,    0,    0,    0,   50,    0, 
            0,    0,    0,    0,    0,    0,  288,    0,    0,    0, 
            0,   52,   53,   54,   55,   56,   57,    0,    0,    0, 
           58,    0,   59,   60,    0,   61,   62,   63,   64,    0, 
            0,    0,  640,  640,  640,    0,  640,    0,   65,   66, 
          640,  640,    0,    0,    0,  640,    0,  640,  640,  640, 
          640,  640,  640,  640,    0,    0,    0,    0,  640,  640, 
          640,  640,  640,  640,  640,    0,    0,  640,    0,    0, 
            0,    0,    0,    0,  640,    0,    0,  640,  640,  640, 
            0,  640,  640,  640,  640,  640,  640,    0,  640,  640, 
          640,  640,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  640,    0,    0,  640,  640,    0,  640,  640, 
            0,  640,    0,    0,    0,    0,    0,  640,    0,    0, 
            0,    0,    0,    0,    0,  640,    0,    0,    0,    0, 
          640,  640,  640,  640,  640,  640,    0,    0,    0,  640, 
            0,  640,  640,    0,  640,  640,  640,  640,    0,    0, 
            0,    4,    5,    6,    0,    8,    0,  640,  640,    9, 
           10,    0,    0,    0,   11,    0,   12,   13,   14,   15, 
           16,   17,   18,    0,    0,    0,    0,   19,   20,   21, 
          239,  240,  241,  242,    0,    0,   26,    0,    0,    0, 
            0,    0,    0,   28,    0,    0,  243,  244,  245,    0, 
          246,   35,  247,  248,  249,  250,    0,   40,   41,   42, 
           43,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,  251,    0,    0,  131,    0,    0,   46,   47,    0, 
           48,    0,    0,    0,    0,    0,   50,    0,    0,    0, 
            0,    0,    0,    0,   51,    0,    0,    0,    0,   52, 
           53,   54,   55,   56,   57,    0,    0,    0,   58,    0, 
           59,   60,    0,   61,   62,   63,   64,    0,    0,    0, 
            4,    5,    6,    0,    8,    0,   65,   66,    9,   10, 
            0,    0,    0,   11,    0,   12,   13,   14,  109,  110, 
           17,   18,    0,    0,    0,    0,  111,  112,  113,  239, 
          240,  241,  242,    0,    0,  118,    0,    0,    0,    0, 
            0,    0,   28,    0,    0,  243,  244,  245,    0,  246, 
           35,  247,  248,  249,  250,    0,   40,   41,   42,   43, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
          251,    0,    0,  131,    0,    0,   46,   47,    0,   48, 
            0,    0,    0,    0,    0,   50,    0,    0,    0,    0, 
            0,    0,    0,  288,    0,    0,    0,    0,   52,   53, 
           54,   55,   56,   57,    0,    0,    0,   58,    0,   59, 
           60,    0,   61,   62,   63,   64,    0,    0,    0,    4, 
            5,    6,    0,    8,    0,   65,   66,    9,   10,    0, 
            0,    0,   11,    0,   12,   13,   14,   15,   16,   17, 
           18,    0,    0,    0,    0,   19,   20,   21,  239,  240, 
          241,  242,    0,    0,  118,    0,    0,    0,    0,    0, 
            0,   28,    0,    0,  243,  244,  245,    0,  246,   35, 
          247,  248,  249,  250,    0,   40,   41,   42,   43,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  251, 
            0,    0,  131,    0,    0,   46,   47,    0,   48,    0, 
            0,    0,    0,    0,   50,    0,    0,    0,    0,    0, 
            0,    0,  288,    0,    0,    0,    0,   52,   53,   54, 
           55,   56,   57,    0,    0,    0,   58,    0,   59,   60, 
            0,   61,   62,   63,   64,    0,    0,    0,  640,  640, 
          640,    0,  640,    0,   65,   66,  640,  640,    0,    0, 
            0,  640,    0,  640,  640,  640,  640,  640,  640,  640, 
            0,    0,    0,    0,  640,  640,  640,  640,  640,  640, 
          640,    0,    0,  640,    0,    0,    0,    0,    0,    0, 
          640,    0,    0,  640,  640,  640,    0,  640,  640,  640, 
          640,  640,  640,    0,  640,  640,  640,  640,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  640,    0, 
            0,  640,    0,    0,  640,  640,    0,  640,    0,    0, 
            0,    0,    0,  640,    0,    0,    0,    0,    0,    0, 
            0,  640,    0,    0,    0,    0,  640,  640,  640,  640, 
          640,  640,    0,    0,    0,  640,    0,  640,  640,    0, 
          640,  640,  640,  640,    0,    0,    0,    4,    5,    6, 
            0,    8,    0,  640,  640,    9,   10,    0,    0,    0, 
           11,    0,   12,   13,   14,  109,  110,   17,   18,    0, 
            0,    0,    0,  111,  112,  113,  114,  115,  116,  117, 
            0,    0,  118,    0,    0,    0,    0,    0,    0,  119, 
            0,    0,  120,  121,  122,    0,  123,   35,  124,  125, 
          126,  127,    0,   40,    0,    0,  128,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  129,    0,    0,    0,  130,    0,    0, 
          131,    0,    0,   46,   47,    0,   48,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,   52,   53,   54,   55,   56, 
           57,    0,    0,    0,   58,    0,   59,   60,    0,   61, 
           62,   63,   64,    0,    0,    0,    4,    5,    6,    0, 
            8,    0,   65,   66,    9,   10,    0,    0,    0,   11, 
            0,   12,   13,   14,  109,  110,   17,   18,    0,    0, 
            0,    0,  111,  112,  113,  267,  268,  269,  270,    0, 
            0,  118,    0,    0,    0,    0,    0,    0,  119,    0, 
            0,  271,  272,  273,    0,  274,   35,  275,  276,  277, 
          278,    0,   40,    0,    0,  128,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,  279,    0,    0,   45, 
            0,    0,   46,   47,    0,   48,    0,   49,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,   52,   53,   54,   55,   56,   57, 
            0,    0,    0,   58,    0,   59,   60,    0,   61,   62, 
           63,   64,    0,    0,    0,    4,    5,    6,    0,    8, 
            0,   65,   66,    9,   10,    0,    0,    0,   11,    0, 
           12,   13,   14,  109,  110,   17,   18,    0,    0,    0, 
            0,  111,  112,  113,  340,  341,  342,  343,    0,    0, 
          118,    0,    0,    0,    0,    0,    0,  119,    0,    0, 
          344,  345,  346,    0,  347,   35,  348,  349,  350,  351, 
            0,   40,    0,    0,  128,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  352,    0,    0,  427,    0, 
            0,   46,   47,    0,   48,    0,  428,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,   52,   53,   54,   55,   56,   57,    0, 
            0,    0,   58,    0,   59,   60,    0,   61,   62,   63, 
           64,    0,    0,    0,    4,    5,    6,    0,    8,    0, 
           65,   66,    9,   10,    0,    0,    0,   11,    0,   12, 
           13,   14,  109,  110,   17,   18,    0,    0,    0,    0, 
          111,  112,  113,  114,  115,  116,  117,    0,    0,  118, 
            0,    0,    0,    0,    0,    0,  119,    0,    0,  120, 
          121,  122,    0,  123,   35,  124,  125,  126,  127,    0, 
           40,    0,    0,  128,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,  130,    0,    0,  131,    0,    0, 
           46,   47,    0,   48,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,   52,   53,   54,   55,   56,   57,    0,    0, 
            0,   58,    0,   59,   60,    0,   61,   62,   63,   64, 
            0,    0,    0,    4,    5,    6,    0,    8,    0,   65, 
           66,    9,   10,    0,    0,    0,   11,    0,   12,   13, 
           14,  109,  110,   17,   18,    0,    0,    0,    0,  111, 
          112,  113,  340,  341,  342,  343,    0,    0,  118,    0, 
            0,    0,    0,    0,    0,  119,    0,    0,  344,  345, 
          346,    0,  347,   35,  348,  349,  350,  351,    0,   40, 
            0,    0,  128,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  352,    0,    0,  427,    0,    0,   46, 
           47,    0,   48,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,   52,   53,   54,   55,   56,   57,    0,    0,    0, 
           58,    0,   59,   60,    0,   61,   62,   63,   64,    0, 
            0,    0,    4,    5,    6,    0,    8,    0,   65,   66, 
            9,   10,    0,    0,    0,   11,    0,   12,   13,   14, 
          109,  110,   17,   18,    0,    0,    0,    0,  111,  112, 
          113,  114,  115,  116,  117,    0,    0,  118,    0,    0, 
            0,    0,    0,    0,  119,    0,    0,  120,  121,  122, 
            0,  988,   35,  124,  125,  989,  127,    0,   40,    0, 
            0,  128,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,  990,    0,    0,  131,    0,    0,   46,   47, 
            0,   48,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
           52,   53,   54,   55,   56,   57,    0,    0,    0,   58, 
            0,   59,   60,    0,   61,   62,   63,   64,    0,    0, 
            0,    4,    5,    6,    0,    8,    0,   65,   66,    9, 
           10,    0,    0,    0,   11,    0,   12,   13,   14,  109, 
          110,   17,   18,    0,    0,    0,    0,  111,  112,  113, 
         1001, 1002, 1003, 1004,    0,    0,  118,    0,    0,    0, 
            0,    0,    0,  119,    0,    0, 1005, 1006, 1007,    0, 
         1008,   35, 1009, 1010, 1011, 1012,    0,   40,    0,    0, 
          128,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0, 1013,    0,    0,  131,    0,    0,   46,   47,    0, 
           48,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,   52, 
           53,   54,   55,   56,   57,    0,    0,    0,   58,    0, 
           59,   60,    0,   61,   62,   63,   64,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,   65,   66,  139,  140, 
          141,  142,  143,  144,  145,  146,  147,  148,  149,  150, 
          151,  152,  153,  154,  155,  156,  157,  158,  159,  160, 
          161,  162,    0,    0,  163,  164,  165,  223,  224,  225, 
          226,  170,  171,  172,  173,  174,  175,  176,  177,  178, 
          179,  180,  181,  227,  228,  229,    0,  230,  186,  310, 
            0,  231,    0,    0,    0,  188,  189,    0,  190,  191, 
          192,  193,  194,  195,  196,    0,    0,  197,  198,    0, 
            0,    0,  199,  200,  201,  202,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,  204, 
          205,    0,  206,  207,  208,  209,  210,  211,  212,  213, 
          214,  215,  216,    0,    0,  217,   52,  139,  140,  141, 
          142,  143,  144,  145,  146,  147,  148,  149,  150,  151, 
          152,  153,  154,  155,  156,  157,  158,  159,  160,  161, 
          162,    0,    0,  163,  164,  165,  223,  224,  225,  226, 
          170,  171,  172,  173,  174,  175,  176,  177,  178,  179, 
          180,  181,  227,  228,  229,    0,  230,  186,    0,    0, 
          231,    0,    0,    0,  188,  189,    0,  190,  191,  192, 
          193,  194,  195,  196,    0,    0,  197,  198,    0,    0, 
            0,  199,  200,  201,  202,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,    0,    0,    0,  204,  205, 
            0,  206,  207,  208,  209,  210,  211,  212,  213,  214, 
          215,  216,    0,    0,  217,   52,  139,  140,  141,  142, 
          143,  144,  145,  146,  147,  148,  149,  150,  151,  152, 
          153,  154,  155,  156,  157,  158,  159,  160,  161,  162, 
            0,    0,  163,  164,  165,  166,  167,  168,  169,  170, 
          171,  172,  173,  174,  175,  176,  177,  178,  179,  180, 
          181,  182,  183,  184,    0,  185,  186,  124,  125,  187, 
          127,    0,    0,  188,  189,    0,  190,  191,  192,  193, 
          194,  195,  196,    0,    0,  197,  198,    0,    0,    0, 
          199,  200,  201,  202,    0,    0,    0,    0,    0,    0, 
          203,    0,    0,    0,    0,    0,    0,  204,  205,    0, 
          206,  207,  208,  209,  210,  211,  212,  213,  214,  215, 
          216,    0,    0,  217,  139,  140,  141,  142,  143,  144, 
          145,  146,  147,  148,  149,  150,  151,  152,  153,  154, 
          155,  156,  157,  158,  159,  160,  161,  162,    0,    0, 
          163,  164,  165,  223,  224,  225,  226,  170,  171,  172, 
          173,  174,  175,  176,  177,  178,  179,  180,  181,  227, 
          228,  229,    0,  230,  186,  360,  361,  231,  362,    0, 
            0,  188,  189,    0,  190,  191,  192,  193,  194,  195, 
          196,    0,    0,  197,  198,    0,    0,    0,  199,  200, 
          201,  202,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,    0,    0,  204,  205,    0,  206,  207, 
          208,  209,  210,  211,  212,  213,  214,  215,  216,    0, 
            0,  217,  139,  140,  141,  142,  143,  144,  145,  146, 
          147,  148,  149,  150,  151,  152,  153,  154,  155,  156, 
          157,  158,  159,  160,  161,  162,    0,    0,  163,  164, 
          165,  223,  224,  225,  226,  170,  171,  172,  173,  174, 
          175,  176,  177,  178,  179,  180,  181,  227,  228,  229, 
            0,  230,  186,    0,    0,  231,    0,    0,    0,  188, 
          189,    0,  190,  191,  192,  193,  194,  195,  196,    0, 
            0,  197,  198,    0,    0,    0,  199,  200,  201,  202, 
            0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 
            0,    0,    0,  204,  205,    0,  206,  207,  208,  209, 
          210,  211,  212,  213,  214,  215,  216,  691,  630,  217, 
            0,  692,    0,    0,    0,  188,  189,    0,  190,  191, 
          192,  193,  194,  195,  196,    0,    0,  197,  198,    0, 
            0,    0,  199,  200,  201,  202,    0,    0,    0,    0, 
            0,    0,  304,    0,    0,    0,    0,    0,    0,  204, 
          205,    0,  206,  207,  208,  209,  210,  211,  212,  213, 
          214,  215,  216,  700,  622,  217,    0,  701,    0,    0, 
            0,  188,  189,    0,  190,  191,  192,  193,  194,  195, 
          196,    0,    0,  197,  198,    0,    0,    0,  199,  200, 
          201,  202,    0,    0,    0,    0,    0,    0,  304,    0, 
            0,    0,    0,    0,    0,  204,  205,    0,  206,  207, 
          208,  209,  210,  211,  212,  213,  214,  215,  216,  703, 
          630,  217,    0,  704,    0,    0,    0,  188,  189,    0, 
          190,  191,  192,  193,  194,  195,  196,    0,    0,  197, 
          198,    0,    0,    0,  199,  200,  201,  202,    0,    0, 
            0,    0,    0,    0,  304,    0,    0,    0,    0,    0, 
            0,  204,  205,    0,  206,  207,  208,  209,  210,  211, 
          212,  213,  214,  215,  216,  731,  622,  217,    0,  732, 
            0,    0,    0,  188,  189,    0,  190,  191,  192,  193, 
          194,  195,  196,    0,    0,  197,  198,    0,    0,    0, 
          199,  200,  201,  202,    0,    0,    0,    0,    0,    0, 
          304,    0,    0,    0,    0,    0,    0,  204,  205,    0, 
          206,  207,  208,  209,  210,  211,  212,  213,  214,  215, 
          216,  734,  630,  217,    0,  735,    0,    0,    0,  188, 
          189,    0,  190,  191,  192,  193,  194,  195,  196,    0, 
            0,  197,  198,    0,    0,    0,  199,  200,  201,  202, 
            0,    0,    0,    0,    0,    0,  304,    0,    0,    0, 
            0,    0,    0,  204,  205,    0,  206,  207,  208,  209, 
          210,  211,  212,  213,  214,  215,  216,  922,  622,  217, 
            0,  923,    0,    0,    0,  188,  189,    0,  190,  191, 
          192,  193,  194,  195,  196,    0,    0,  197,  198,    0, 
            0,    0,  199,  200,  201,  202,    0,    0,    0,    0, 
            0,    0,  304,    0,    0,    0,    0,    0,    0,  204, 
          205,    0,  206,  207,  208,  209,  210,  211,  212,  213, 
          214,  215,  216,  925,  630,  217,    0,  926,    0,    0, 
            0,  188,  189,    0,  190,  191,  192,  193,  194,  195, 
          196,    0,    0,  197,  198,    0,    0,    0,  199,  200, 
          201,  202,    0,    0,    0,    0,    0,    0,  304,    0, 
            0,    0,    0,    0,    0,  204,  205,    0,  206,  207, 
          208,  209,  210,  211,  212,  213,  214,  215,  216, 1080, 
          622,  217,    0, 1081,    0,    0,    0,  188,  189,    0, 
          190,  191,  192,  193,  194,  195,  196,    0,    0,  197, 
          198,    0,    0,    0,  199,  200,  201,  202,    0,    0, 
            0,    0,    0,    0,  304,    0,    0,    0,    0,    0, 
            0,  204,  205,    0,  206,  207,  208,  209,  210,  211, 
          212,  213,  214,  215,  216, 1083,  622,  217,    0, 1084, 
            0,    0,    0,  188,  189,    0,  190,  191,  192,  193, 
          194,  195,  196,    0,    0,  197,  198,    0,    0,    0, 
          199,  200,  201,  202,    0,    0,    0,    0,    0,    0, 
          304,    0,    0,    0,    0,    0,    0,  204,  205,    0, 
          206,  207,  208,  209,  210,  211,  212,  213,  214,  215, 
          216, 1086,  630,  217,    0, 1087,    0,    0,    0,  188, 
          189,    0,  190,  191,  192,  193,  194,  195,  196,    0, 
            0,  197,  198,    0,    0,    0,  199,  200,  201,  202, 
            0,    0,    0,    0,    0,    0,  304,    0,    0,    0, 
            0,    0,    0,  204,  205,    0,  206,  207,  208,  209, 
          210,  211,  212,  213,  214,  215,  216,    0,    0,  217, 
      };
   }

   private static final short[] yyCheck1() {
      return new short[] {

           26,   74,   28,   88,    2,   49,   91,   27,   10,  100, 
            3,   10,   11,    2,    3,    4,    5,  258,  296,  468, 
            9,   10,   11,   45,   51,   14,   15,   16,    6,  448, 
           19,   11,  130,  418,    2,    3,    6,    7,    4,    5, 
           15,   16,   21,  266,   19,    7,  430,    0,   14,   47, 
           48,    7,  581,   10,   59,  489,   45,   27,    0,  105, 
           49,  446,   51,   15,   16,   27,  486,   19,   32,  489, 
           10,   27,  335,   10,   52,  758,  339,  462,   10,   10, 
           53,  772,   52,   49,   10,   15,   16,   44,  538,   19, 
           10,  684,  477,  119,  528,   47,   48,  509,   10,   10, 
          693,  417,   59,  419,   10,    0,  559,  423,  674,  867, 
           10,  100,   10,  756,   21,   10,  280,   47,   44,    9, 
           10,   11,   59,   10,  690,  105,   10,   59,  107,  367, 
          368,  447,   44,   59,  100,   10,  702,   10,   44,   44, 
          129,  424,  131,   10,   10,  461,   44,  463,  431,   61, 
          322,  536,   10,   10,  984,   61,   61,   44,   10,  305, 
           44,   59,  478,   59,   59,   10,   10,  733,   10,  254, 
          305,   10,   59,    0,   10,   59,   10,   44,   10,   91, 
          964,   10,   10,   10,   59,  345,  306,   10,   61,   43, 
          310,  280,   59,   59,   10,  280,  512,   10,  362,  890, 
          107,   59,   10,  363,  341,  655,  113,   59,  312,   10, 
           44,  325,   32,  317,   59,   59,   44,   59,  572,  573, 
          498,  537,   10,  373,  308,  678,  341,  680,  378,  375, 
          137,  316,   59,   10,   10,  424,   59,  426,   10,  129, 
          375,   10,  431,   59, 1074,  359,  350,  351,   10,  238, 
          669,   59,  341,   10,  588,  427,  373,   10,   59,  784, 
          679,  378,  353,   10,   10,  790,  317, 1025,   10,  354, 
           10,  372,   44,  362,  128,   10,   10,  342,  324, 1063, 
          362,  261,  716,  269,   10,  271,  620,   59,  971,   10, 
          394,   45,  376,  377,  359,  715,  716,   59,  349,  683, 
          322,  310,  353,  354,   61,   32,  304,  290,  291,   10, 
           44,  549,  301,  725,  303,  304,  289,  342,  307,   59, 
          309,   32,  965,   44,   59,  314,  301,   32,  303,  304, 
           32,  320,  366,  322,  359,   61,  282,  980,   59,   61, 
          342,  338,  331,  345,  324,   44,  584,  445,  305,  301, 
          338,  303,  320,  307,  359,  805,  885,  950,  924,  341, 
           61,  780,  357,  331,  353,  894,   61,  279,  280,   44, 
          899,  301,  310,  303,  279,  329,  330,  131,  328,  338, 
          379,  380,  381,  382,  428,  342,  336,  353,  807,   61, 
          379,  380,  381,  382,  383,  384,  385,  424,  279,  426, 
          267,  341,  359,  430,  431,  427,   10,  371,  372,  267, 
          363,  342,  271,  798,  378,  383,  328,  307,  375,  309, 
          418,  363,  342,  335,  336,  345,  363, 1070,  558,  418, 
          342,  342,  710,  359,  345,  424,  342,  426,  427,  428, 
          433,  430,  431,  418,  433,  345,  342,  538,  446,  419, 
          362,  290,  291,  423,  695,   59,  264,  446,  559,  448, 
          449,  359,  428,  395,  462,  433,  729,  342,  363,  342, 
          459,  446,  359,  462,  473,  359,  342,  447,  467,  477, 
          281,  797,  705,  799,  473,  342,  506,  462,  477,  379, 
          380,  381,  382,  463,  384,  385,  426,  338,  781,  782, 
          430,  950,  477,  741,  787,  788,  342,  359,  478,  345, 
          342,  509,  419,  342,  359,  359, 1045,  871,  507, 1085, 
          509,  955,  279,  877,  878,   10,  880,  267,  882,  342, 
          768,  520,  267,  305,  486,  955,  506,  491,  536,  469, 
          447,  509,  512,  305,  506,  499,  500,  536,   10,  538, 
          506,  371,  372,  279,  342,   10,  463,  642,  378,  449, 
          314,  536,   59,  517,  655,  342,  342,  537,  322,  847, 
          342,  478,  538,  525,  310,   10,   61,  678,  279,  680, 
          342, 1000,   44,  473,  305,  342,   10,  359,  586,  342, 
          588,  617,  781,  782,  105,  342,  342,   59,  787,  788, 
          342,  339,  628,  375,  339,  512,   61,  268,  269,   44, 
           44,  800,   10,  375,  744,  745,  342,  507,    2,    3, 
          372,  342,  620,   10,   59,    9,   10,   11,  911,   44, 
          537,   15,   16,  362,  632,   19,  317,   61,  317,  730, 
          638,  342,  328,   10,  338,  624,  263,  264,  674,  335, 
          336,  342,  269,  632,  375,  756,   91,  758,  345,  638, 
          363,   45,  682,   61,  690,  310,  655,   51,   44,  668, 
         1055,  772,   59,  427,  350,  351,  702,  675,   44,  668, 
          669, 1035, 1036, 1037, 1038,  683,  684,  667,  686,  655, 
          679,  263,  264,   10,   61,  693,   44,   10,  939,  347, 
          654,  305,  341,  681,  945,  459,  865,  733,  697,   10, 
          699,  681,  682,  467,  805,  341,  262,  624,  698,   48, 
          682,  794,  911,  677,  750,  632,  682,  725,  264,  267, 
           10,  638,   10,  264,  342,  865,  725,  867,  342, 1093, 
         1056,  730,   44,   91,   61,  129,  772,  131,   61,   44, 
          261,  363,  362,  683,  781,  782,   10,  725,   59,   10, 
          787,  788,   44,  715,  730,  672,  720,  306,  675,  772, 
           10,  375,  306,  800,  728,  424,   44,   44,  668,   59, 
           44,   59,  431,   44,   44,  774,   44,   91,  777,  306, 
           61,  780,  781,  782,  279,  749,  363,  264,  787,  788, 
          798,  960,  322,  323,   44,   59,  795,  559,   59,  798, 
          338,  800,  338,  324,  338,  464,  805,  466,  807,   59, 
          338,  297,  820,  798,  279,  823,  825,  795,  808,  799, 
          960,  961,   61,  822,   44,  815,  825,  357,  358,  805, 
           15,   16,   44,  305,   19,  280,  264,  836,  269,  371, 
          372,  373,   44,  264,  238,  279,  378,  342,   91,  813, 
           10,  338,  363,  338,  890,  514,   41,   42, 1027,  338, 
          971,   32,   47,   48,  396,   50,   51,  264,  338,  980, 
          342,  279,  269,  984,  911,  874,  875,  342,   44,   32, 
           91,   58,  799,  328,   44, 1025,  338, 1027,  924,  345, 
          335,  336,  363, 1062,   15,   16,  341,  264,   19,   59, 
          864,  900,  279,  375,   91,   10,  339,  301,  342,  303, 
          304,  264,  911,  307,  359,  309,  678,  362,  680, 1014, 
          314,  279, 1062,  263,  900,  825,  320,  264,  322,  371, 
          372,   91,  338,  697,  342,  699,  378,  331,  264,  938, 
          264,  383,  950,  942,  264,   44,  285,  286,  287,  363, 
           44,  290,  279,  264,   59,  954,  279,  342,  269,   59, 
          271,   44,  938, 1074,  977,  342,   44,  980,   44,  982, 
          328,  984,  262,  263,  264,   44,  264,  335,  336,  269, 
          268,  269,   44,  341,   44,  379,  380,  381,  382,  383, 
          384,  385,   44,  992, 1030,   44,  359,   61,  262,  263, 
          264, 1000,  263,  264,  268,  269,  264,  271,  264,   44, 
          774,  264,   91,  777,  328,  342,  992,  267, 1017,  342, 
         1019,  335,  336,  359,  418, 1024,  310,   44,   91,  363, 
          424,   10,  426,  427,  310,  264,  430,  431,  306,  433, 
          308,  309,  310,  311,   44,   44,   44, 1055,   44, 1085, 
         1058,   44,  446,   44,  448,  449, 1055, 1070,  822, 1072, 
          264, 1074,  395,   44, 1077,  459,  271,  363,  462,  341, 
         1055,  271,  836,  467,   91,  342, 1056,  363,  342,  473, 
           59,   58,  756,  477,  758,  328,  345,  426, 1101,  339, 
           44,  430,  335,  336,  379,  380,  381,  382,  772,  363, 
          285,  286,  287,  288,  306,  290,  308,  309,  310,  311, 
          874,  875,   91,  507,  345,  509,  301,  328,  303,  304, 
          345,  345,  781,  782,  335,  336,  520,  312,  787,  788, 
          469,  395,  264,   70,   91,  584,    5,   91,  520, 1056, 
         1058,  328,  536,    6,  638,   10,  485,  486,  335,  336, 
          489,  756,  937,    2,    3,    4,    5,  262,  263,  264, 
          772,  760,  982,  268,  269,   14,  271,  980,  328,  828, 
          829,   14,  831,  832,  793,  335,  336,   91,  942,   91, 
          301,  341,  303,   10,  665,  760,  525, 1057,  763,  528, 
          954,  371,  372,  373,   59,   89,   45,  772,  378,  359, 
           49,  386,  387,  388,  389,  390, 1063,  744,  393,  394, 
          395,  396,  397,  398,  399,  400,  401,  402,  403,  404, 
          405,  406,   -1,   -1,  409,  410,  411,  412,   -1,  371, 
          372,  373,   59,  418,   -1,  420,  378,  342,   -1,  424, 
           -1,  426,   -1,  582,   -1,  430,  431,  308,  309,  328, 
          311,  100,   -1, 1017,   -1, 1019,  335,  336,  363,   -1, 
         1024,  446,  279,  280,   91,  328,   -1,   -1,   -1,  371, 
          372,  373,  335,  336,  668,  669,  378,  462,   -1,  464, 
          465,  466,  131,   10,  469,  679,   -1,   -1,  267,   -1, 
          395,  306,  477,  308,  309,  310,  311,  971,   -1,    0, 
          485,  486,   -1,  697,  489,  699,  980,   -1,  647,   10, 
          984,  328,   -1,   -1,   -1,  376,  377,   44,  335,  336, 
           -1,   -1,    2,    3,  341,  279,  280,   -1,  513,  514, 
           -1,  725,   59,   -1,   14,   -1,   -1,   44,   -1,   -1, 
          525,   -1,   -1,  528,  683,  362,  371,  372,  373,  328, 
           -1,  536,   -1,  378,   -1,   -1,  335,  336,   59,  371, 
          372,  373,  341,   -1,   91,   45,  378,  706,   -1,  708, 
           -1,  328,   44,   -1,  328,   -1,  715,  716,  335,  336, 
          774,  335,  336,  777,   91,   -1,  780,  781,  782,  238, 
           -1,   -1,   -1,  787,  788,   -1,   -1,  582,   -1,   -1, 
         1074,  795,  977,   -1,  798,  980,  800,  982,  362,  984, 
          371,  372,  373,  807,  328,    0,  328,  378,   -1,   91, 
           -1,  335,  336,  335,  336,   10,   -1,  766,  822,   -1, 
          306,  825,  308,  309,  310,  311,    2,    3,   -1,   -1, 
          267,  306,  836,  308,  309,  310,  311,  312,   -1,   -1, 
          753,  131,  317,  280,  306,  758,  308,  309,  310,  311, 
           -1,   -1,  647,  802,  340,  314,   10,   -1,   -1,  772, 
           -1,  320,  348,  322,   59,  340,   -1,   -1,   -1,   45, 
          874,  875,  331,  348,  349,  350,  351,   -1,  340,  293, 
          294,  295,  296,  297,   -1, 1070,  835, 1072,  683, 1074, 
           44,  328, 1077,   -1,  353,   -1,   10,   -1,  335,  336, 
           -1,   -1,   -1,   -1,  341,   59,   -1,  911,   -1,   -1, 
           -1,  706,   -1,  708,   -1,   -1, 1101,   -1,   -1,  394, 
          715,  716,    0,   -1,  383,  362,   -1,   -1,   -1,   -1, 
           44,   -1,   10,   -1,   -1,   10,   -1,   91,  942,   -1, 
           -1,   -1,   -1,  280,   -1,   -1,   -1,   61,  238,   -1, 
          954,  262,  263,  264,   -1,  131,   -1,  268,  269,   -1, 
          271,   -1,  279,   -1,   -1,  424,   44,  426,  427,  428, 
           -1,  766,  431,   -1,  433,   -1,   -1,   91,  927,   -1, 
           58,   59,  293,  294,  295,  296,  781,  782,   -1,  448, 
           -1,  328,  787,  788,   -1,   -1, 1000,  279,  335,  336, 
          459,  796,   -1,  798,  341,  800,  955,  802,  467,   -1, 
           -1,  328,   -1, 1017,   -1, 1019,   -1,   -1,  335,  336, 
         1024,   -1,  359,   -1,  314,  362,   -1,   -1,   -1,   10, 
          320,  342,  322,  828,  829,   -1,  831,  832,   -1,   10, 
          835,  331,   -1,   44,   -1,   -1,  328,   44,   -1,   -1, 
          509, 1055,  363,  335,  336,   -1,  969,   -1,  971,   -1, 
          973,  520,  238,   44,  977,   -1,   -1,  262,  263,  264, 
          760,  984,  267,  268,  269,   -1,  271,   -1,   -1,  538, 
           61,   -1,  772,   -1,  395,   -1,  281,  282,   59,   -1, 
           91,  317,   -1,  383,   91,  290,  291,   -1,  293,  294, 
          295,  296,  297,   -1,    0,   -1,   -1,  333,  334,  904, 
           91,   -1,   -1,   -1,   10,   -1,  911,   -1,   -1,   -1, 
           91,   -1,   -1,  349,   -1,  351,  280,  353,  354,  355, 
          356,   -1,  927,  359,  424,  361,  426,  427,  314,   -1, 
           -1,  431,   -1,  433,  320,   -1,  322,  342,   44,   -1, 
          345,   -1,  947, 1066,   -1,  331,   -1,   -1,  448, 1072, 
          955, 1074,   -1,   59, 1077,  279,   -1,   63,  363,  459, 
           -1,   -1,   -1,   -1,  328,   -1,   -1,  467,   -1,   -1, 
           -1,  335,  336,   -1,  262,  263,  264,  341, 1101,  267, 
          268,  269,   -1,  271,   -1,   91,  655,   -1,   -1,   -1, 
          395,   -1,   -1,  281,  282,  359,   -1,  383,  362,   -1, 
          669,   -1,  290,  291,  328,  293,  294,  295,  296,  509, 
          679,  335,  336,   -1,   -1,   -1,   -1,  305,  342,   -1, 
          520,  306,   -1,  308,  309,  310,  311,  312,  697,   -1, 
          699,   -1,  317,   -1,    0,   -1,   -1,   -1,  424,   -1, 
          426,  427,   -1,   -1,   10,  431,   -1,  433,   -1,   -1, 
         1055,  339,   -1,   -1,  342,  340,  725,  345,   -1,  347, 
           -1,  730,  448,  348,  349,  350,  351,   -1,  279,  738, 
           -1,   -1,  279,  459,   -1,  363,   -1,  977,   44,   -1, 
          980,  467,  982,   -1,  984,   -1,  267,  375,  279,  280, 
            0,   -1,   10,   59,   -1,   -1,   -1,   -1,   -1,  280, 
           10,   -1,   -1,   -1,   -1,  774,   -1,  395,  777,  394, 
           -1,  780,  781,  782,   -1,   -1,   -1,  328,  787,  788, 
           -1,  328,   -1,  509,  335,  336,  795,   -1,  335,  336, 
           -1,  800,   -1,   -1,  520,   -1,  805,  328,  807,   -1, 
           -1,   59,   -1,   -1,  335,  336,   -1,  328,   -1,   59, 
          341,  342,   -1,  822,  335,  336,  262,  263,  264,   -1, 
          341,  267,  268,  269,   -1,  271,   -1,  836,   -1,  669, 
         1070,  362, 1072,   91, 1074,  281,  282, 1077,   -1,  679, 
           -1,  362,   -1,   -1,  290,  291,   -1,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,   -1,  697,   -1,  699, 
            0, 1101,   -1,   -1,   -1,  874,  875,   -1,   -1,   -1, 
           10,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,  725,   -1,  333,  334,  335, 
          336,  900,   -1,  339,   -1,   -1,  342,   -1,   -1,  345, 
           -1,  347,  911,  349,   44,  351,   -1,  353,  354,  355, 
          356,  357,  358,  359,   -1,  361,   -1,  363,   58,   59, 
           -1,   -1,   -1,   63,   -1,   10,   -1,   44,   -1,  938, 
           -1,   -1,   -1,  942,  774,   10,   -1,  777,   -1,   -1, 
          780,  781,  782,  669,   10,  954,   -1,  787,  788,  395, 
           -1,   -1,   -1,  679,   -1,  795,  262,  263,  264,   44, 
          800,   -1,  268,  269,  891,  271,   -1,  807,  895,   44, 
           -1,  697,   -1,  699,   91,   -1,   61,   -1,   44,   -1, 
           -1,   -1,  822,  992,  290,  291,   61,  293,  294,  295, 
          296, 1000,   -1,   -1,   -1,   61,  836,   -1,   -1,  725, 
           -1,   -1,   -1,   44,   -1,   -1,   91,   -1, 1017,  267, 
         1019,   -1,  262,  263,  264, 1024,   91,  267,  268,  269, 
           -1,  271,  280,   -1,  306,   91,  308,  309,  310,  311, 
           -1,  281,  282,   -1,  874,  875,  342,   -1,   -1,   -1, 
          290,  291,    0,  293,  294,  295,  296,  297,  774,   -1, 
           91,  777,   10,   -1,  780,  781,  782,  363,  340,   -1, 
           -1,  787,  788,   -1,   -1,   -1,  348,   -1,   -1,  795, 
          328,  911,   -1,   -1,  800,   -1,   -1,  335,  336,   -1, 
           -1,  807,   -1,  341,   -1,   -1,   44,   -1,   -1,  395, 
           -1,   -1,  342,   -1,   -1,  345,  822,  347,  938,   -1, 
           58,   59,  942,   61,  362,   63,   -1,   -1,   -1,   -1, 
          836,   -1, 1039,  363,  954, 1042, 1043,   -1,   -1, 1046, 
         1047,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           -1,  271,   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  281,  282,   -1,   -1,  395,   -1,   -1,  874,  875, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   -1, 
         1000,   -1,  279,  280,   -1,  305,   -1, 1094, 1095, 1096, 
         1097,   -1,   -1,   -1,   -1,   -1,   -1, 1017, 1105, 1019, 
           -1,    0,   -1,   -1, 1024,  911,   -1,   -1,   -1,   -1, 
           -1,   10,   -1,   -1,  279,  280,   -1,   -1,   -1,  339, 
           -1,   -1,  342,   -1,  279,  345,   -1,  347,   -1,   -1, 
           -1,  328,   -1,  279,  280,   -1,  942,   -1,  335,  336, 
           -1,   -1,   -1,  363,  341,   44,   -1,   -1,  954,   -1, 
           -1,   -1,   -1,   -1,   -1,  375,   -1,   -1,  279,   58, 
           59,   -1,   61,  328,   63,  362,   -1,  317,   -1,   -1, 
          335,  336,   -1,  328,   -1,  395,  341,  342,   -1,   -1, 
          335,  336,  328,  333,  334,   -1,  341,  342,   -1,  335, 
          336,   -1,   91,   -1, 1000,  341,  342,  362,   -1,  349, 
           -1,  351,   -1,  353,  354,  355,  356,  328,   -1,   -1, 
           -1, 1017,   -1, 1019,  335,  336,  362,   -1, 1024,  257, 
          258,  259,   -1,  261,  262,  263,  264,  265,  266,  267, 
          268,  269,  270,  271,  272,  273,  274,  275,  276,  277, 
          278,   -1,  280,  281,  282,  283,  284,  285,  286,  287, 
          288,  289,  290,  291,  292,  293,  294,  295,  296,  297, 
           -1,  299,   -1,   -1,  302,  303,  304,  305,  306,  307, 
          308,  309,  310,  311,  312,  313,  314,  315,  316,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   44,   -1,  333,  334,  335,  336,  337, 
          338,  339,  340,  341,  342,  343,  344,  345,  346,  347, 
          348,  349,  350,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,  362,  363,   -1,  365,  366,  367, 
          368,  369,  370,   44,   -1,   -1,  374,  375,  376,  377, 
           91,  379,  380,  381,  382,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,   -1,  392,  393,  394,  395,  257,  258, 
          259,   -1,  261,  262,  263,  264,  265,  266,  267,  268, 
          269,  270,  271,  272,  273,  274,  275,  276,  277,  278, 
           91,  280,  281,  282,  283,  284,  285,  286,  287,  288, 
          289,  290,  291,  292,  293,  294,  295,  296,  297,   -1, 
          299,   59,   -1,  302,  303,  304,  305,  306,  307,  308, 
          309,  310,  311,  312,  313,  314,  315,  316,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   44,   91,  333,  334,  335,  336,  337,  338, 
          339,  340,  341,  342,  343,  344,  345,  346,  347,  348, 
          349,  350,  351,  352,  353,  354,  355,  356,  357,  358, 
          359,  360,  361,  362,  363,   -1,  365,  366,  367,  368, 
          369,  370,   -1,   -1,    0,  374,  375,  376,  377,   91, 
          379,  380,  381,  382,   10,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  392,  393,  394,  395,   -1,  262,  263, 
          264,   -1,   -1,   -1,  268,  269,  306,  271,  308,  309, 
          310,  311,  312,  262,  263,  264,   -1,  317,   44,  268, 
          269,   -1,  271,   -1,   -1,   -1,   -1,   -1,  279,  280, 
           -1,   -1,   58,   59,  317,   61,   -1,   63,   -1,   -1, 
          340,  341,   -1,   -1,  317,   -1,   -1,   -1,  348,  349, 
          350,  351,   -1,  306,   -1,  308,  309,  310,  311,  312, 
          333,  334,   -1,   -1,  317,   91,  349,   -1,  279,  280, 
          353,  354,  355,  356,   -1,   -1,  349,  328,  342,   -1, 
          353,  354,  355,  356,  335,  336,   -1,  340,   -1,   -1, 
          341,   -1,   -1,  342,  394,  348,  349,  350,  351,  363, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  267, 
           -1,  362,   -1,    0,  363,   -1,   -1,  328,   -1,   -1, 
           -1,   -1,  317,   10,  335,  336,   -1,  322,  323,   -1, 
          341,  395,   -1,   -1,   -1,   -1,   -1,   -1,  333,  334, 
           -1,  394,   -1,   -1,   -1,   -1,  395,   -1,   -1,   -1, 
           -1,  362,   -1,   -1,  349,   -1,  351,   44,  353,  354, 
          355,  356,  357,  358,  359,   -1,  361,  279,   -1,   -1, 
          328,   58,   59,   -1,   61,   -1,   63,  335,  336,   -1, 
           -1,   -1,   -1,   -1,  306,   -1,  308,  309,  310,  311, 
          312,   -1,   -1,   -1,   -1,  317,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   91,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  328,   -1,  340,   -1, 
           -1,   -1,   -1,  335,  336,   10,   -1,   -1,  350,  351, 
           -1,  257,  258,  259,   -1,  261,  262,  263,  264,  265, 
          266,  267,  268,  269,  270,  271,  272,  273,  274,  275, 
          276,  277,  278,   -1,  280,  281,  282,  283,  284,  285, 
          286,  287,  288,  289,  290,  291,  292,  293,  294,  295, 
          296,  297,  394,  299,   59,   -1,  302,  303,  304,  305, 
          306,  307,  308,  309,  310,  311,  312,  313,  314,  315, 
          316,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,   44,   -1,  333,  334,  335, 
          336,  337,   10,  339,  340,  341,  342,  343,  344,  345, 
          346,  347,  348,  349,  350,  351,  352,  353,  354,  355, 
          356,  357,  358,  359,  360,  361,  362,  363,   -1,  365, 
          366,  367,  368,  369,  370,   44,   -1,   -1,  374,  375, 
          376,  377,   91,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,   59,   -1,   -1,   10,   -1,  392,  393,  394,  395, 
          257,  258,  259,   -1,  261,  262,  263,  264,  265,  266, 
          267,  268,  269,  270,  271,  272,  273,  274,  275,  276, 
          277,  278,   91,  280,  281,  282,  283,  284,  285,  286, 
          287,  288,  289,  290,  291,  292,  293,  294,  295,  296, 
          297,   -1,  299,   59,   -1,  302,  303,  304,  305,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   44,   91,  333,  334,  335,  336, 
          337,   -1,  339,  340,  341,  342,  343,  344,  345,  346, 
          347,  348,  349,  350,  351,  352,  353,  354,  355,  356, 
          357,  358,  359,  360,  361,  362,  363,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,    0,  374,  375,  376, 
          377,   91,  379,  380,  381,  382,   10,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  392,  393,  394,  395,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  306,   -1,  308,  309,  310,  311,  312,   -1,   -1, 
           44,   -1,  317,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          279,  280,   -1,  328,   58,   59,   -1,   61,   -1,   63, 
           -1,  336,   -1,   -1,   -1,  340,  341,   -1,   -1,   -1, 
           -1,   -1,   -1,  348,  349,  350,  351,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   -1, 
          279,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  328, 
           -1,   -1,   -1,   -1,   -1,   -1,  335,  336,   -1,   -1, 
           -1,   -1,  341,   -1,   -1,   -1,   -1,   -1,  306,  394, 
          308,  309,  310,  311,  312,   -1,   -1,   -1,   -1,  317, 
           -1,  267,   -1,  362,   -1,    0,   -1,   -1,   -1,  328, 
          328,   -1,   -1,   -1,   -1,   10,  335,  336,  336,   -1, 
           -1,   -1,  340,  341,   -1,   -1,   -1,   -1,   -1,   -1, 
          348,  349,  350,  351,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  317,  318,  319,  320,  321,  322,  323,  324,   44, 
          326,  327,   -1,   -1,   -1,   -1,   -1,  333,  334,  279, 
           -1,   -1,  328,   58,   59,   -1,   -1,   -1,   63,  335, 
          336,   -1,   -1,  349,   -1,  351,  394,  353,  354,  355, 
          356,  357,  358,  359,   -1,  361,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  328,   -1, 
           -1,   -1,   -1,   -1,   -1,  335,  336,   10,   -1,   -1, 
           -1,   -1,   -1,  257,  258,  259,   -1,  261,  262,  263, 
          264,  265,  266,  267,  268,  269,  270,  271,  272,  273, 
          274,  275,  276,  277,  278,   -1,  280,  281,  282,  283, 
          284,  285,  286,  287,  288,  289,  290,  291,  292,  293, 
          294,  295,  296,  297,   -1,  299,   59,   -1,  302,  303, 
          304,  305,  306,  307,  308,  309,  310,  311,  312,  313, 
          314,  315,  316,  317,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,  328,  329,  330,   44,   -1,  333, 
          334,  335,  336,  337,   10,  339,  340,  341,  342,  343, 
          344,  345,  346,  347,  348,  349,  350,  351,  352,  353, 
          354,  355,  356,  357,  358,  359,  360,  361,  362,  363, 
           -1,  365,  366,  367,  368,  369,  370,   44,   -1,   -1, 
          374,  375,  376,  377,   91,  379,  380,  381,  382,   -1, 
           -1,   -1,   -1,   59,   -1,   -1,   10,   -1,  392,  393, 
          394,  395,  257,  258,  259,   -1,  261,  262,  263,  264, 
          265,  266,  267,  268,  269,  270,  271,  272,  273,  274, 
          275,  276,  277,  278,   91,  280,  281,  282,  283,  284, 
          285,  286,  287,  288,  289,  290,  291,  292,  293,  294, 
          295,  296,  297,   -1,  299,   59,   -1,  302,  303,  304, 
          305,  306,  307,  308,  309,  310,  311,  312,  313,  314, 
          315,  316,  317,  318,  319,  320,  321,  322,  323,  324, 
          325,  326,  327,  328,  329,  330,   44,   91,  333,  334, 
          335,  336,  337,   -1,  339,  340,  341,  342,  343,  344, 
          345,  346,  347,  348,  349,  350,  351,  352,  353,  354, 
          355,  356,  357,  358,  359,  360,  361,  362,  363,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,    0,  374, 
          375,  376,  377,   91,  379,  380,  381,  382,   10,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  392,  393,  394, 
          395,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  306,   -1,  308,  309,  310,  311,  312, 
           -1,   -1,   44,   -1,  317,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  279,   -1,   -1,  328,   58,   59,   -1,   -1, 
           -1,   63,   -1,  336,   -1,   -1,   -1,  340,  341,   -1, 
           -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91, 
           -1,   -1,  279,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  328,   -1,   -1,   -1,   -1,   -1,   -1,  335,  336, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          306,  394,  308,  309,  310,  311,  312,   -1,   -1,   -1, 
           -1,  317,   -1,  267,   -1,   -1,   -1,    0,   -1,   -1, 
           -1,  328,  328,   -1,   -1,   -1,   -1,   10,  335,  336, 
          336,   -1,   -1,   -1,  340,  341,   -1,   -1,   -1,   -1, 
           -1,   -1,  348,  349,  350,  351,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  317,  318,  319,  320,  321,  322,  323, 
           -1,   44,  326,  327,   -1,   -1,   -1,   -1,   -1,  333, 
          334,  279,   -1,   -1,  328,   58,   59,   -1,   -1,   -1, 
           63,  335,  336,   -1,   -1,  349,   -1,  351,  394,  353, 
          354,  355,  356,  357,  358,  359,   -1,  361,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          328,   -1,   -1,   -1,   -1,   -1,   -1,  335,  336,   10, 
           -1,   -1,   -1,   -1,   -1,  257,  258,  259,   -1,  261, 
          262,  263,  264,  265,  266,  267,  268,  269,  270,  271, 
          272,  273,  274,  275,  276,  277,  278,   -1,  280,  281, 
          282,  283,  284,  285,  286,  287,  288,  289,  290,  291, 
          292,  293,  294,  295,  296,  297,   -1,  299,   59,   -1, 
          302,  303,  304,  305,  306,  307,  308,  309,  310,  311, 
          312,  313,  314,  315,  316,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,  328,  329,  330,   44, 
           -1,  333,  334,  335,  336,  337,   10,  339,  340,  341, 
          342,  343,  344,  345,  346,  347,  348,  349,  350,  351, 
          352,  353,  354,  355,  356,  357,  358,  359,  360,  361, 
          362,  363,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,  375,  376,  377,   91,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   59,   -1,   -1,   10,   -1, 
          392,  393,  394,  395,  257,  258,  259,   -1,  261,  262, 
          263,  264,  265,  266,  267,  268,  269,  270,  271,  272, 
          273,  274,  275,  276,  277,  278,   -1,  280,  281,  282, 
          283,  284,  285,  286,  287,  288,  289,  290,  291,  292, 
          293,  294,  295,  296,  297,   -1,  299,   59,   -1,  302, 
          303,  304,  305,  306,  307,  308,  309,  310,  311,  312, 
          313,  314,  315,  316,  317,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,  328,  329,  330,   -1,   91, 
          333,  334,  335,  336,  337,   -1,  339,  340,  341,  342, 
          343,  344,  345,  346,  347,  348,  349,  350,  351,  352, 
          353,  354,  355,  356,  357,  358,  359,  360,  361,  362, 
          363,   -1,  365,  366,  367,  368,  369,  370,   -1,   -1, 
            0,  374,  375,  376,  377,   -1,  379,  380,  381,  382, 
           10,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  392, 
          393,  394,  395,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  306,   -1,  308,  309,  310, 
          311,  312,   -1,   -1,   44,   -1,  317,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  279,   -1,   -1,  328,   58,   59, 
           -1,   61,   -1,   63,   -1,  336,   -1,   -1,   -1,  340, 
          341,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350, 
          351,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  328,   -1,   -1,   -1,   -1,   -1,   -1, 
          335,  336,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  306,  394,  308,  309,  310,  311,  312,   -1, 
           -1,   -1,   -1,  317,   -1,  267,   -1,   -1,   -1,    0, 
           -1,   -1,   -1,   -1,  328,   -1,   -1,   -1,   -1,   10, 
           -1,   -1,  336,   -1,   -1,   -1,  340,  341,   -1,   -1, 
           -1,   -1,   -1,   -1,  348,  349,  350,  351,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   44,  317,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,   -1,  328,   -1,   59,   -1, 
          333,  334,   63,  335,  336,   -1,   -1,   -1,   -1,   -1, 
          394,   -1,   -1,   -1,   -1,   -1,  349,   -1,  351,   -1, 
          353,  354,  355,  356,  357,  358,  359,   -1,  361,   -1, 
           91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   10,   -1,   -1,   -1,   -1,   -1,  257,  258,  259, 
           -1,  261,  262,  263,  264,  265,  266,  267,  268,  269, 
          270,  271,  272,  273,  274,  275,  276,  277,  278,   -1, 
           -1,  281,  282,  283,  284,  285,  286,  287,  288,  289, 
          290,  291,  292,  293,  294,  295,  296,  297,   -1,  299, 
           59,   -1,  302,  303,  304,  305,  306,  307,  308,  309, 
          310,  311,  312,  313,  314,  315,  316,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   -1,   -1,  333,  334,  335,  336,  337,   10,  339, 
          340,  341,  342,  343,  344,  345,  346,  347,  348,  349, 
          350,  351,  352,  353,  354,  355,  356,  357,  358,  359, 
          360,  361,   -1,  363,   -1,  365,  366,  367,  368,  369, 
          370,   -1,   -1,   -1,  374,  375,  376,  377,   -1,  379, 
          380,  381,  382,   -1,   -1,   -1,   -1,   59,   -1,   -1, 
           10,   -1,  392,  393,  394,  395,  257,  258,  259,   -1, 
          261,  262,  263,  264,  265,  266,  267,  268,  269,  270, 
          271,  272,  273,  274,  275,  276,  277,  278,   -1,  280, 
          281,  282,  283,  284,  285,  286,  287,  288,  289,  290, 
          291,  292,  293,  294,  295,  296,  297,   -1,  299,   59, 
           -1,  302,  303,  304,   -1,  306,  307,  308,  309,  310, 
          311,  312,  313,  314,  315,  316,  317,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,  328,  329,  330, 
           63,   91,  333,  334,  335,  336,  337,   -1,  339,  340, 
           -1,  342,  343,  344,  345,  346,  347,  348,  349,  350, 
          351,  352,  353,  354,  355,  356,  357,  358,  359,  360, 
          361,  362,  363,   -1,  365,  366,  367,  368,  369,  370, 
           -1,   -1,    0,  374,   -1,  376,  377,   -1,  379,  380, 
          381,  382,   10,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  392,  393,  394,  395,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  306,   -1,  308, 
          309,  310,  311,  312,   -1,   -1,   44,   -1,  317,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  328, 
           -1,   59,   -1,   -1,   -1,   63,   -1,  336,   -1,   -1, 
           -1,  340,  341,   -1,   -1,   -1,   -1,   -1,   -1,  348, 
          349,  350,  351,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  306,  394,  308,  309,  310,  311, 
          312,   -1,   -1,   -1,   -1,  317,   -1,  267,   -1,   -1, 
           -1,    0,   -1,   -1,   -1,   -1,  328,   -1,   -1,   -1, 
           -1,   10,   -1,   -1,  336,   -1,   -1,   -1,  340,  341, 
           -1,   -1,   -1,   -1,   -1,   -1,  348,  349,  350,  351, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   44,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  297,   -1,   -1,   -1,  328,   -1, 
           59,   -1,   -1,   -1,   63,  335,  336,   -1,   -1,   -1, 
           -1,   -1,  394,   -1,  317,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,   -1,  329,  330,   -1,   -1, 
          333,  334,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  349,   -1,  351,   -1, 
          353,  354,  355,  356,  357,  358,  359,   -1,  361,  257, 
          258,  259,   -1,  261,  262,  263,  264,  265,  266,  267, 
          268,  269,  270,  271,  272,  273,  274,  275,  276,  277, 
          278,   -1,  280,  281,  282,  283,  284,  285,  286,  287, 
          288,  289,  290,  291,  292,  293,  294,  295,  296,  297, 
           -1,  299,   -1,   -1,  302,  303,  304,   -1,  306,  307, 
          308,  309,  310,  311,  312,  313,  314,  315,  316,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,    0,   -1,  333,  334,  335,  336,  337, 
           -1,  339,  340,   10,  342,  343,  344,  345,  346,  347, 
          348,  349,  350,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,  362,  363,   -1,  365,  366,  367, 
          368,  369,  370,   -1,   -1,   -1,  374,   44,  376,  377, 
           -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1,   -1, 
           -1,   58,   59,   -1,  392,  393,  394,  395,  257,  258, 
          259,   -1,  261,  262,  263,  264,  265,  266,  267,  268, 
          269,  270,  271,  272,  273,  274,  275,  276,  277,  278, 
           -1,   -1,  281,  282,  283,  284,  285,  286,  287,  288, 
          289,  290,  291,  292,  293,  294,  295,  296,  297,   -1, 
          299,   -1,   -1,  302,  303,  304,   -1,  306,  307,  308, 
          309,  310,  311,  312,  313,  314,  315,  316,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   -1,   -1,  333,  334,  335,  336,  337,   -1, 
          339,  340,   -1,  342,  343,  344,  345,  346,  347,  348, 
          349,  350,  351,  352,  353,  354,  355,  356,  357,  358, 
          359,  360,  361,    0,  363,   -1,  365,  366,  367,  368, 
          369,  370,   -1,   10,   -1,  374,   -1,  376,  377,   -1, 
          379,  380,  381,  382,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  392,  393,  394,  395,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   44,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   59,   -1,   61,   -1,   63,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   91,  262,  263,  264,   -1,   -1, 
          267,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  281,  282,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  290,  291,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,   -1, 
           -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1, 
          347,   -1,   -1,   -1,   -1,   -1,   44,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  363,   -1,  306,  307, 
           -1,   59,  310,   61,   -1,   63,  314,  315,  375,  317, 
          318,  319,  320,  321,  322,  323,   -1,   -1,  326,  327, 
           -1,   -1,   -1,  331,  332,  333,  334,   -1,  395,   -1, 
           -1,   -1,   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1, 
          348,  349,   -1,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,   10,   -1,  364,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,  262,  263,  264,  265,  266, 
           -1,  268,  269,  270,  271,  272,  273,  274,  275,  276, 
          277,  278,   -1,  280,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,  290,  291,  292,  293,  294,  295,  296, 
          297,   -1,  299,   59,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   -1,   -1,  333,  334,  335,  336, 
          337,  338,   10,  340,  341,  342,  343,  344,   -1,  346, 
           -1,  348,  349,  350,  351,  352,  353,  354,  355,  356, 
          357,  358,  359,  360,  361,  362,  363,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   44,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   61,   -1,  392,  393,  394,  395,  257, 
          258,  259,   -1,  261,  262,  263,  264,  265,  266,   -1, 
          268,  269,  270,  271,  272,  273,  274,  275,  276,  277, 
          278,   -1,  280,   91,   -1,  283,  284,  285,  286,  287, 
          288,  289,  290,  291,  292,  293,  294,  295,  296,  297, 
           -1,  299,   -1,   -1,  302,  303,  304,   -1,  306,  307, 
          308,  309,  310,  311,  312,  313,  314,  315,  316,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   63,   -1,  333,  334,  335,  336,  337, 
          338,   -1,  340,  341,  342,  343,  344,   -1,  346,   -1, 
          348,  349,  350,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,  362,  363,   -1,  365,  366,  367, 
          368,  369,  370,   -1,   -1,    0,  374,   -1,  376,  377, 
           -1,  379,  380,  381,  382,   10,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  392,  393,  394,  395,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          306,   -1,  308,  309,  310,  311,  312,   -1,   -1,   44, 
           -1,  317,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  328,   -1,   59,   -1,   61,   -1,   63,   -1, 
          336,   -1,   -1,   -1,  340,  341,   -1,   -1,   -1,   -1, 
           -1,   -1,  348,  349,  350,  351,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   91,   -1,   -1,   -1, 
           -1,  279,  280,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  394,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,    0,   -1,   -1,   -1,   -1,   -1, 
          328,   -1,   -1,   -1,   10,   -1,   -1,  335,  336,   -1, 
           -1,   -1,   -1,  341,  342,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  362,   -1,   -1,   -1,   44,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   59,   -1,   61,   -1,   63,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,   -1,  329, 
          330,   -1,   -1,  333,  334,   91,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  349, 
           -1,  351,   -1,  353,  354,  355,  356,  357,  358,  359, 
           -1,  361,  257,  258,  259,   -1,  261,  262,  263,  264, 
          265,  266,   -1,  268,  269,  270,  271,  272,  273,  274, 
          275,  276,  277,  278,   -1,  280,   -1,   -1,  283,  284, 
          285,  286,  287,  288,  289,  290,  291,  292,  293,  294, 
          295,  296,  297,   -1,  299,   -1,   -1,  302,  303,  304, 
           -1,  306,  307,  308,  309,  310,  311,  312,  313,  314, 
          315,  316,  317,  318,  319,  320,  321,  322,  323,  324, 
          325,  326,  327,  328,  329,  330,   -1,   -1,  333,  334, 
          335,  336,  337,   -1,   10,  340,  341,  342,  343,  344, 
           -1,  346,   -1,  348,  349,  350,  351,  352,  353,  354, 
          355,  356,  357,  358,  359,  360,  361,  362,  363,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,   44,  374, 
           -1,  376,  377,   -1,  379,  380,  381,  382,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   61,   -1,  392,  393, 
      };
   }

   private static final short[] yyCheck2() {
      return new short[] {

          394,  395,  257,  258,  259,   -1,  261,  262,  263,  264, 
          265,  266,   -1,  268,  269,  270,  271,  272,  273,  274, 
          275,  276,  277,  278,   -1,  280,   91,   -1,  283,  284, 
          285,  286,  287,  288,  289,  290,  291,  292,  293,  294, 
          295,  296,  297,   -1,  299,   -1,   -1,  302,  303,  304, 
           -1,  306,  307,  308,  309,  310,  311,  312,  313,  314, 
          315,  316,  317,  318,  319,  320,  321,  322,  323,  324, 
          325,  326,  327,  328,  329,  330,   -1,   -1,  333,  334, 
          335,  336,  337,   -1,   -1,  340,  341,  342,  343,  344, 
           -1,  346,   -1,  348,  349,  350,  351,  352,  353,  354, 
          355,  356,  357,  358,  359,  360,  361,  362,  363,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,    0,  374, 
           -1,  376,  377,   -1,  379,  380,  381,  382,   10,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  392,  393,  394, 
          395,   -1,   -1,  306,  307,   -1,   -1,  310,   -1,   -1, 
           -1,  314,  315,   -1,  317,  318,  319,  320,  321,  322, 
          323,   -1,   44,  326,  327,   -1,   -1,   -1,  331,  332, 
          333,  334,   -1,   -1,   -1,   -1,   -1,   59,  341,   61, 
           -1,   63,   -1,   -1,   -1,  348,  349,   -1,  351,  352, 
          353,  354,  355,  356,  357,  358,  359,  360,  361,   -1, 
           -1,  364,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91, 
           -1,   -1,   -1,   -1,  279,  280,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,    0,   -1,   -1, 
           -1,   -1,   -1,  328,   -1,   -1,   -1,   10,   -1,   -1, 
          335,  336,   -1,   -1,   -1,   -1,  341,  342,   -1,  306, 
          307,   -1,   -1,  310,   -1,   -1,   -1,  314,  315,   -1, 
          317,  318,  319,  320,  321,  322,  323,  362,   -1,  326, 
          327,   44,   -1,   -1,  331,  332,  333,  334,   -1,   -1, 
           -1,   -1,   -1,   -1,  341,   -1,   59,   -1,   61,   -1, 
           63,  348,  349,   -1,  351,  352,  353,  354,  355,  356, 
          357,  358,  359,  360,  361,   -1,   -1,  364,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   91,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  257,  258,  259,   -1,  261, 
          262,  263,  264,  265,  266,   -1,  268,  269,  270,  271, 
          272,  273,  274,  275,  276,  277,  278,   -1,  280,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,  290,  291, 
          292,  293,  294,  295,  296,  297,   -1,  299,   -1,   -1, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
          312,  313,  314,  315,  316,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,  328,  329,  330,   -1, 
           -1,  333,  334,  335,  336,  337,   -1,   10,  340,  341, 
          342,  343,  344,   -1,  346,   -1,  348,  349,  350,  351, 
          352,  353,  354,  355,  356,  357,  358,  359,  360,  361, 
          362,  363,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,   -1,   -1,   59,   -1,   10,   -1, 
          392,  393,  394,  395,  257,  258,  259,   -1,  261,  262, 
          263,  264,  265,  266,   -1,  268,  269,  270,  271,  272, 
          273,  274,  275,  276,  277,  278,   -1,  280,   91,   -1, 
          283,  284,  285,  286,  287,  288,  289,  290,  291,  292, 
          293,  294,  295,  296,  297,   -1,  299,   59,   -1,  302, 
          303,  304,   -1,  306,  307,  308,  309,  310,  311,  312, 
          313,  314,  315,  316,  317,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,  328,  329,  330,   -1,   91, 
          333,  334,  335,  336,  337,   -1,   -1,  340,  341,  342, 
          343,  344,   -1,  346,   -1,  348,  349,  350,  351,  352, 
          353,  354,  355,  356,  357,  358,  359,  360,  361,  362, 
          363,   -1,  365,  366,  367,  368,  369,  370,   -1,   -1, 
            0,  374,   -1,  376,  377,   -1,  379,  380,  381,  382, 
           10,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  392, 
          393,  394,  395,   -1,   -1,  306,  307,   -1,   -1,  310, 
           -1,   -1,   -1,  314,  315,   -1,  317,  318,  319,  320, 
          321,  322,  323,   -1,   44,  326,  327,   -1,   -1,   -1, 
          331,  332,  333,  334,   -1,   -1,   -1,   -1,   -1,   59, 
          341,   61,   -1,   63,   -1,   -1,   -1,  348,  349,   -1, 
          351,  352,  353,  354,  355,  356,  357,  358,  359,  360, 
          361,   -1,   -1,  364,  267,   -1,   -1,   -1,   -1,   -1, 
           -1,   91,   -1,   -1,   -1,   -1,   -1,  280,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  267,   -1,   -1,   -1,    0, 
           -1,   -1,   -1,   -1,   -1,  328,   -1,   -1,   -1,   10, 
           -1,   -1,  335,  336,   -1,   -1,   -1,   -1,  341,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362, 
           -1,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  328,   -1,   59,   -1, 
           61,   -1,   63,  335,  336,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  257,  258,  259, 
           -1,  261,  262,  263,  264,  265,  266,   -1,  268,  269, 
          270,  271,  272,  273,  274,  275,  276,  277,  278,   -1, 
          280,   -1,   -1,  283,  284,  285,  286,  287,  288,  289, 
          290,  291,  292,  293,  294,  295,  296,  297,   -1,  299, 
           -1,   -1,  302,  303,  304,   -1,  306,  307,  308,  309, 
          310,  311,  312,  313,  314,  315,  316,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   -1,   -1,  333,  334,  335,  336,  337,   -1,   10, 
          340,  341,  342,  343,  344,   -1,  346,   -1,  348,  349, 
          350,  351,  352,  353,  354,  355,  356,  357,  358,  359, 
          360,  361,  362,  363,   -1,  365,  366,  367,  368,  369, 
          370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379, 
          380,  381,  382,   -1,   -1,   -1,   -1,   -1,   59,   -1, 
           10,   -1,  392,  393,  394,  395,  257,  258,  259,   -1, 
          261,  262,  263,  264,  265,  266,   -1,  268,  269,  270, 
          271,  272,  273,  274,  275,  276,  277,  278,   -1,  280, 
           91,   -1,  283,  284,  285,  286,  287,  288,  289,  290, 
          291,  292,  293,  294,  295,  296,  297,   -1,  299,   59, 
           -1,  302,  303,  304,   -1,  306,  307,  308,  309,  310, 
          311,  312,  313,  314,  315,  316,  317,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,  328,  329,  330, 
           -1,   91,  333,  334,  335,  336,  337,   -1,   -1,  340, 
          341,  342,  343,  344,   -1,  346,   -1,  348,  349,  350, 
          351,  352,  353,  354,  355,  356,  357,  358,  359,  360, 
          361,  362,  363,   -1,  365,  366,  367,  368,  369,  370, 
           -1,   -1,    0,  374,   -1,  376,  377,   -1,  379,  380, 
          381,  382,   10,   -1,   -1,   10,   -1,   -1,   -1,   -1, 
           -1,  392,  393,  394,  395,   -1,   -1,  306,  307,   -1, 
           -1,  310,   -1,   -1,   -1,  314,  315,   -1,  317,  318, 
          319,  320,  321,  322,  323,   -1,   44,  326,  327,   44, 
           -1,   -1,  331,  332,  333,  334,   -1,   -1,   -1,   -1, 
           -1,   59,  341,   61,   -1,   63,   61,   -1,   -1,  348, 
          349,   -1,  351,  352,  353,  354,  355,  356,  357,  358, 
          359,  360,  361,   -1,   -1,  364,  267,   -1,   -1,   -1, 
           -1,   -1,   -1,   91,   -1,   -1,   91,   -1,   -1,  280, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  267,   -1,   -1, 
           -1,    0,   -1,   -1,   -1,   -1,   -1,  328,   -1,   -1, 
           -1,   10,   -1,   -1,  335,  336,   -1,   -1,   -1,   -1, 
          341,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  362,   -1,   -1,   -1,   44,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  328,   -1, 
           59,   -1,   61,   -1,   63,  335,  336,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  257, 
          258,  259,   -1,  261,  262,  263,  264,  265,  266,   -1, 
          268,  269,  270,  271,  272,  273,  274,  275,  276,  277, 
          278,   -1,   -1,   -1,  279,  283,  284,  285,  286,  287, 
          288,  289,  290,  291,  292,  293,  294,  295,  296,  297, 
           -1,  299,   -1,   -1,  302,  303,  304,   -1,  306,  307, 
          308,  309,  310,  311,  312,  313,  314,  315,  316,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,  328,   -1,  333,  334,  335,  336,  337, 
          335,  336,  340,  341,  342,  343,  344,  342,  346,   -1, 
          348,  349,  350,  351,  352,  353,  354,  355,  356,  357, 
          358,  359,  360,  361,    0,  363,   -1,  365,  366,  367, 
          368,  369,  370,   -1,   10,   -1,  374,   -1,  376,  377, 
           -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  392,  393,  394,  395,  257,  258, 
          259,   -1,  261,  262,  263,  264,  265,  266,   -1,  268, 
          269,  270,  271,  272,  273,  274,  275,  276,  277,  278, 
           -1,   -1,   -1,   59,  283,  284,  285,  286,  287,  288, 
          289,  290,  291,  292,  293,  294,  295,  296,  297,   -1, 
          299,   -1,   -1,  302,  303,  304,   -1,  306,  307,  308, 
          309,  310,  311,  312,  313,  314,  315,  316,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   -1,   -1,  333,  334,  335,  336,  337,   -1, 
           -1,  340,  341,  342,  343,  344,   -1,  346,   -1,  348, 
          349,  350,  351,  352,  353,  354,  355,  356,  357,  358, 
          359,  360,  361,    0,  363,   -1,  365,  366,  367,  368, 
          369,  370,   -1,   10,   -1,  374,   -1,  376,  377,   -1, 
          379,  380,  381,  382,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  392,  393,  394,  395,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   44,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   58,   59,   -1,   61,   -1,   63,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   91,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   10,   -1,   10,   -1,   -1,   -1,   -1, 
           -1,  257,  258,  259,   -1,  261,  262,  263,  264,  265, 
          266,  267,  268,  269,  270,  271,  272,  273,  274,  275, 
          276,  277,  278,   -1,  280,  281,   -1,  283,  284,  285, 
          286,  287,  288,  289,  290,  291,  292,  293,  294,  295, 
          296,  297,   59,  299,   59,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,  312,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  328,   -1,   91,   -1,   91,   -1,   -1,  335, 
          336,  337,   -1,   -1,  340,   -1,  342,  343,  344,   -1, 
          346,   -1,  348,   -1,  350,   -1,  352,   -1,    0,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,  362,  363,   10,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  392,  393,  394,  395, 
           -1,   -1,   44,   -1,   -1,  262,  263,  264,   -1,   -1, 
          267,  268,  269,   -1,  271,   -1,   58,   59,   -1,   -1, 
           -1,   63,   -1,  280,  281,  282,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  290,  291,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,   91, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   -1,   -1,  333,  334,  335,  336, 
           -1,    0,  339,   -1,   -1,  342,   -1,   -1,  345,   -1, 
          347,   10,  349,   -1,  351,   -1,  353,  354,  355,  356, 
          357,  358,  359,   -1,  361,  362,  363,   -1,   -1,   -1, 
          267,   -1,  267,   -1,   -1,   -1,   -1,   -1,  375,   -1, 
           -1,   -1,   -1,  280,   -1,   44,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   10,   -1,   -1,   -1,   -1,  395,   58, 
           59,   -1,   61,   -1,   63,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  328,   91,  328,   -1,   -1,   -1,   -1,  335,  336, 
          335,  336,   59,   -1,   10,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,    0,   -1, 
           -1,   -1,   -1,   -1,   -1,  362,   -1,   -1,   10,   -1, 
           -1,   -1,   -1,   -1,   91,   -1,   -1,   -1,   -1,   -1, 
          262,  263,  264,   -1,   -1,  267,  268,  269,   -1,  271, 
           -1,   -1,   -1,   59,   -1,   -1,   -1,   -1,  280,  281, 
          282,   -1,   44,   -1,   -1,   -1,   -1,   -1,  290,  291, 
           10,  293,  294,  295,  296,  297,   58,   59,   -1,   61, 
           -1,   63,   -1,  305,   -1,   91,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   10,   -1,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,  328,  329,  330,   91, 
           -1,  333,  334,  335,  336,   -1,   -1,  339,   -1,   59, 
          342,   -1,   -1,  345,   -1,  347,   -1,  349,   10,  351, 
           -1,  353,  354,  355,  356,  357,  358,  359,   -1,  361, 
          362,  363,   59,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   91,   -1,  375,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   44,  262,  263,  264,   -1,   -1,  267,  268, 
          269,   -1,  271,  395,   91,   -1,   -1,   -1,   -1,   61, 
           -1,   -1,  281,  282,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  305,    0,   -1,   91, 
          267,   -1,   -1,   -1,   -1,   -1,   -1,   10,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   -1,   -1,  333,  334,  335,  336,   -1,  338, 
          339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1, 
          349,   44,  351,   -1,  353,  354,  355,  356,  357,  358, 
          359,  267,  361,   -1,  363,   58,   59,   -1,   61,   -1, 
           63,  328,   -1,   -1,   -1,   -1,  375,   -1,  335,  336, 
          262,  263,  264,   -1,   -1,  267,  268,  269,   -1,  271, 
           -1,   -1,   -1,   -1,   -1,   -1,  395,   -1,   91,  281, 
          282,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,   -1,   -1,   -1,   -1, 
           -1,   -1,  328,  305,    0,   -1,   -1,  267,   -1,  335, 
          336,   -1,   -1,   -1,   10,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,  328,  329,  330,   -1, 
          267,  333,  334,  335,  336,   -1,  338,  339,   -1,   -1, 
          342,   -1,   -1,  345,   -1,  347,   -1,  349,   44,  351, 
           -1,  353,  354,  355,  356,  357,  358,  359,   -1,  361, 
           -1,  363,   58,   59,   -1,   61,   -1,   63,  328,   -1, 
           -1,   -1,   -1,  375,   -1,  335,  336,  279,   -1,   10, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  328,   -1,  395,   -1,   91,   -1,   -1,  335,  336, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  328,    0,   -1,   -1, 
           61,   -1,   -1,  335,  336,   -1,   -1,   10,   -1,   -1, 
          342,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262, 
          263,  264,   -1,   -1,  267,  268,  269,   -1,  271,   -1, 
           91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  281,  282, 
           -1,   44,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1, 
          293,  294,  295,  296,  297,   58,   59,   -1,   61,   -1, 
           63,   -1,  305,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  317,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,  328,  329,  330,   91,   -1, 
          333,  334,  335,  336,   -1,  338,  339,   -1,   -1,  342, 
           10,   -1,  345,   -1,  347,   -1,  349,   -1,  351,   -1, 
          353,  354,  355,  356,  357,  358,  359,    0,  361,   -1, 
          363,   -1,   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1, 
           -1,   -1,  375,   -1,   44,   -1,  262,  263,  264,   -1, 
           -1,  267,  268,  269,   -1,  271,   -1,   -1,   -1,   -1, 
           -1,   61,  395,   -1,   -1,  281,  282,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,   59,   -1,   -1,  305, 
            0,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           10,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,   -1,   -1,  333,  334,  335, 
          336,   -1,  338,  339,   -1,   -1,  342,   -1,  279,  345, 
           -1,  347,   -1,  349,   44,  351,   -1,  353,  354,  355, 
          356,  357,  358,  359,   -1,  361,   -1,  363,   58,   59, 
           -1,   61,   -1,   63,   -1,   -1,   -1,   -1,   -1,  375, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262, 
          263,  264,   -1,   -1,  267,  268,  269,  328,  271,  395, 
           -1,   91,   -1,   -1,  335,  336,   -1,   -1,  281,  282, 
           -1,  342,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1, 
          293,  294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  305,    0,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   10,  317,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,  328,  329,  330,   -1,   -1, 
          333,  334,  335,  336,   -1,  338,  339,   -1,   -1,  342, 
           -1,   -1,  345,   -1,  347,   -1,  349,   44,  351,   -1, 
          353,  354,  355,  356,  357,  358,  359,   -1,  361,   -1, 
          363,   58,   59,   -1,   61,   -1,   63,   -1,   -1,  279, 
           -1,   -1,  375,   -1,   -1,   -1,   -1,   -1,   -1,  262, 
          263,  264,   -1,   -1,  267,  268,  269,   -1,  271,   -1, 
           -1,   -1,  395,   -1,   91,   -1,   -1,   -1,  281,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1, 
          293,  294,  295,  296,  297,   -1,   -1,   -1,  328,   -1, 
            0,   -1,   -1,   -1,   -1,  335,  336,   -1,   -1,   -1, 
           10,   -1,  342,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  262,  263,  264,  328,   -1,  267,  268,  269, 
           -1,  271,  335,  336,   -1,   -1,   -1,   -1,   -1,  342, 
           -1,  281,  282,   -1,   44,   -1,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   58,   59, 
          363,   61,   -1,   63,   -1,  305,   -1,   -1,   -1,   10, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   91,  395,  333,  334,  335,  336,   -1,  338,  339, 
           -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1,  349, 
           -1,  351,   -1,  353,  354,  355,  356,  357,  358,  359, 
           -1,  361,   63,  363,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  375,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1, 
          267,  268,  269,   -1,  271,  395,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  281,  282,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  290,  291,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   -1,   -1,  333,  334,  335,  336, 
           -1,  338,  339,   -1,   -1,  342,   -1,   -1,  345,   -1, 
          347,   -1,  349,   44,  351,   -1,  353,  354,  355,  356, 
          357,  358,  359,   -1,  361,   -1,  363,   58,   59,   -1, 
           61,   -1,   63,   -1,   -1,   -1,   -1,   -1,  375,   -1, 
           -1,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,  395,   -1, 
           91,  281,  282,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  305,    0,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   10,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   -1,   -1,  333,  334,  335,  336,   -1,  338,  339, 
           -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1,  349, 
           44,  351,   -1,  353,  354,  355,  356,  357,  358,  359, 
           -1,  361,   -1,  363,   58,   59,   -1,   61,   -1,   63, 
           -1,   -1,   -1,   -1,   -1,  375,  317,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,   -1,  329,  330, 
           -1,   -1,  333,  334,   -1,  395,   -1,   91,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  349,   -1, 
          351,   -1,  353,  354,  355,  356,  357,  358,  359,   -1, 
          361,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          281,  282,   -1,   44,   -1,   -1,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   58,   59,   -1, 
           -1,   -1,   63,   -1,  305,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  317,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,  328,  329,  330, 
           91,   -1,  333,  334,  335,  336,   -1,  338,  339,   -1, 
           -1,  342,   10,   -1,  345,   -1,  347,   -1,  349,   -1, 
          351,   -1,  353,  354,  355,  356,  357,  358,  359,    0, 
          361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
           -1,   -1,   -1,   -1,  375,   -1,   44,   -1,  262,  263, 
          264,   -1,   -1,  267,  268,  269,   -1,  271,   -1,   -1, 
           -1,   -1,   -1,   61,  395,   -1,   -1,  281,  282,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   59,   -1, 
           -1,  305,    0,   91,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,  317,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,  328,  329,  330,   -1,   -1,  333, 
          334,  335,  336,   -1,  338,  339,   -1,   -1,  342,   -1, 
           -1,  345,   -1,  347,   -1,  349,   44,  351,   -1,  353, 
          354,  355,  356,  357,  358,  359,   -1,  361,   -1,  363, 
           58,   59,   -1,   61,   -1,   63,   -1,   -1,   -1,   -1, 
           -1,  375,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,  395,   -1,   91,   -1,   -1,   -1,   -1,   -1,  280, 
          281,  282,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  305,    0,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   10,  317,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,  328,  329,  330, 
           -1,   -1,  333,  334,  335,  336,   -1,   -1,  339,   -1, 
           -1,  342,   -1,   -1,  345,   -1,  347,   -1,  349,   44, 
          351,   -1,  353,  354,  355,  356,  357,  358,  359,   -1, 
          361,  362,  363,   58,   59,   -1,   61,   -1,   63,   -1, 
           -1,  279,   -1,   -1,  375,   -1,   -1,   -1,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,   -1,   -1,   -1,  395,   -1,   91,   -1,   -1,   -1, 
          281,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
          328,   -1,    0,   -1,   -1,   -1,   -1,  335,  336,   -1, 
           -1,   -1,   10,   -1,  342,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  342,   -1,  281,  282,   -1,   44,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           58,   59,  363,   -1,   -1,   63,   -1,  305,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   91,  395,  333,  334,  335,  336,   -1, 
           -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347, 
           -1,  349,   10,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,   -1,  361,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  375,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,  262,  263,  264, 
           -1,   -1,  267,  268,  269,   -1,  271,  395,   -1,   -1, 
           -1,   -1,   -1,   61,   -1,   -1,  281,  282,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294, 
          295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          305,    0,   -1,   91,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   10,  317,  318,  319,  320,  321,  322,  323,  324, 
          325,  326,  327,  328,  329,  330,   -1,   -1,  333,  334, 
          335,  336,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1, 
          345,   -1,  347,   -1,  349,   44,  351,   -1,  353,  354, 
          355,  356,  357,  358,  359,   -1,  361,   -1,  363,   58, 
           59,   -1,   -1,   -1,   63,   -1,   -1,   -1,   -1,   -1, 
          375,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
          395,   -1,   91,  281,  282,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,    0,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   -1,   -1,  333,  334,  335,  336,   -1, 
           -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347, 
           -1,  349,   44,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,   -1,  361,   -1,  363,   58,   59,   -1,   -1, 
           -1,   63,   -1,   -1,   -1,   -1,   -1,  375,   -1,   -1, 
           -1,  279,   -1,   10,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  395,   -1,   91, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   44,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          328,    0,   -1,   -1,   61,   -1,   -1,  335,  336,   -1, 
           -1,   10,   -1,   -1,  342,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267,  268, 
          269,   -1,  271,   -1,   91,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  281,  282,   -1,   44,   -1,   -1,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   58, 
           59,   -1,   -1,   -1,   63,   -1,  305,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   91,   -1,  333,  334,  335,  336,   -1,   -1, 
          339,   -1,   -1,  342,   10,   -1,  345,   -1,  347,   -1, 
          349,   -1,  351,   -1,  353,  354,  355,  356,  357,  358, 
          359,    0,  361,   -1,  363,   -1,   -1,   -1,   -1,   -1, 
           -1,   10,   -1,   -1,   -1,   -1,  375,   -1,   44,   -1, 
          262,  263,  264,   -1,   -1,  267,  268,  269,   -1,  271, 
           -1,   -1,   -1,   -1,   -1,   61,  395,   -1,   -1,  281, 
          282,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,   -1,   -1,   -1,   -1, 
           59,   -1,   -1,  305,    0,   91,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   10,  317,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,  328,  329,  330,   -1, 
           -1,  333,  334,  335,  336,   -1,   -1,  339,   -1,   -1, 
          342,   -1,  279,  345,   -1,  347,   -1,  349,   44,  351, 
           -1,  353,  354,  355,  356,  357,  358,  359,   -1,  361, 
           -1,  363,   58,   59,   -1,   -1,   -1,   63,   -1,   -1, 
           -1,   -1,   -1,  375,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267,  268, 
          269,  328,  271,  395,   -1,   91,   -1,   -1,  335,  336, 
           -1,   -1,  281,  282,   -1,  342,   -1,   -1,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  305,    0,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,  317,  318, 
          319,  320,  321,  322,  323,  324,  325,  326,  327,  328, 
          329,  330,   -1,   -1,  333,  334,  335,  336,   -1,   -1, 
          339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1, 
          349,   44,  351,   -1,  353,  354,  355,  356,  357,  358, 
          359,   -1,  361,   -1,  363,   -1,   59,   -1,   -1,   -1, 
           63,   -1,   -1,  279,   -1,   -1,  375,   -1,   -1,   -1, 
           -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267,  268, 
          269,   -1,  271,   -1,   -1,   -1,  395,   -1,   91,   -1, 
           -1,   -1,  281,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
            0,  290,  291,   -1,  293,  294,  295,  296,  297,   -1, 
           10,   -1,  328,   -1,   -1,   -1,   -1,   -1,   -1,  335, 
          336,   -1,   -1,   -1,   -1,   -1,  342,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1, 
           -1,  267,  268,  269,   44,  271,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  342,   -1,  281,  282,   -1,   -1,   59, 
           -1,   -1,   -1,   63,  290,  291,   -1,  293,  294,  295, 
          296,  297,   -1,   -1,  363,   -1,   -1,   -1,   -1,  305, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   91,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,   -1,  395,  333,  334,  335, 
          336,   -1,    0,  339,   -1,   -1,  342,   -1,   -1,  345, 
           -1,  347,   10,  349,   -1,  351,   -1,  353,  354,  355, 
          356,  357,  358,  359,   -1,  361,   -1,  363,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  375, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,   -1,   -1,  262, 
          263,  264,   -1,   -1,  267,  268,  269,   -1,  271,  395, 
           58,   59,   -1,   -1,   -1,   63,   -1,   -1,  281,  282, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1, 
          293,  294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  317,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,  328,  329,  330,   -1,   -1, 
          333,  334,  335,  336,   -1,   -1,  339,   -1,   -1,  342, 
           -1,   -1,  345,   -1,  347,   -1,  349,   -1,  351,   -1, 
          353,  354,  355,  356,  357,  358,  359,   -1,  361,   -1, 
          363,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  281,  282,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          290,  291,  395,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,    0,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   10,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   -1,   -1,  333,  334,  335,  336,   -1,   -1,  339, 
           -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1,  349, 
           44,  351,   -1,  353,  354,  355,  356,  357,  358,  359, 
           -1,  361,   -1,  363,   58,   59,   -1,   -1,   -1,   63, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  281,  282,  395,   -1,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,   -1,   -1, 
           -1,   10,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
            0,  329,  330,   -1,   -1,  333,  334,   -1,   -1,   -1, 
           10,  339,   -1,   -1,  342,   44,   -1,  345,   10,  347, 
           -1,  349,   -1,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,   61,  361,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,   -1,   44,   -1,   -1,  375,   -1,   -1, 
           -1,   -1,   44,   -1,   -1,   -1,   -1,   -1,   58,   59, 
           -1,   -1,   91,   63,   -1,   -1,   -1,  395,   -1,   61, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   61,   -1,   -1,   -1,   -1,   -1,   91, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
           -1,   -1,   -1,   91,   -1,   -1,   -1,   -1,  262,  263, 
          264,   -1,   -1,  267,  268,  269,   -1,  271,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  281,  282,   -1, 
           -1,   -1,   -1,   44,   -1,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   58,   59,   -1, 
           -1,  305,   63,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  317,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,   -1,  329,  330,   -1,   -1,  333, 
          334,   -1,   -1,   -1,   -1,  339,   -1,   -1,  342,   -1, 
           -1,  345,   -1,  347,   -1,  349,   -1,  351,   -1,  353, 
          354,  355,  356,  357,  358,  359,    0,  361,   -1,  363, 
           -1,   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1,   -1, 
           -1,  375,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          279,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  395,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           44,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  281,  282,   -1,   58,   59,   -1,  279,   -1,   63, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,  328, 
           -1,   -1,   -1,   -1,   -1,  305,  335,  336,   -1,   -1, 
           -1,  279,   -1,  342,   -1,   -1,   -1,   -1,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,   -1,  329, 
          330,   -1,   -1,  333,  334,   -1,  328,   -1,   -1,  339, 
           -1,   -1,  342,  335,  336,  345,   -1,  347,   -1,  349, 
          342,  351,   -1,  353,  354,  355,  356,  357,  358,  359, 
          328,  361,   -1,  363,   -1,   -1,   -1,  335,  336,   -1, 
           -1,  262,  263,  264,  342,  375,  267,  268,  269,   -1, 
          271,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          281,  282,   10,   -1,   -1,  395,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  305,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,   -1,  329,  330, 
           58,   59,  333,  334,   -1,   63,   -1,   -1,  339,   -1, 
           -1,  342,   -1,   -1,  345,   -1,  347,   -1,  349,   -1, 
          351,   -1,  353,  354,  355,  356,  357,  358,  359,   -1, 
          361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  375,   -1,   -1,   -1,  262,  263, 
          264,   -1,   -1,  267,  268,  269,   -1,  271,   -1,    0, 
           -1,   -1,   -1,   -1,  395,   -1,   -1,  281,  282,   10, 
           -1,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  305,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   44,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,   -1,  329,  330,   58,   59,  333, 
          334,   -1,   63,   -1,   -1,  339,   -1,   -1,  342,   -1, 
           -1,  345,   -1,  347,   -1,  349,   -1,  351,   -1,  353, 
          354,  355,  356,  357,  358,  359,   -1,  361,   -1,  363, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  375,   -1,   -1,   -1,   -1,   -1,   -1,    0,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,   -1, 
           -1,  395,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,  267, 
          268,  269,   44,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  281,  282,   -1,   58,   59,   -1,   -1, 
           -1,   63,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
           -1,  329,  330,   -1,   -1,  333,  334,   -1,   -1,   -1, 
           -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1,  347, 
           -1,  349,   -1,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,   -1,  361,    0,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   10,   -1,   -1,  375,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,   -1,   -1,   -1,   -1,   -1,   -1,  395,   -1,   -1, 
          281,  282,   -1,   -1,   -1,   -1,   -1,   -1,   44,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
           -1,   -1,   -1,   59,  305,   61,   -1,   63,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,   -1,  329,  330, 
           -1,   -1,  333,  334,   -1,   91,   -1,   -1,  339,   -1, 
           -1,  342,   -1,   -1,  345,   -1,  347,   -1,  349,   -1, 
          351,   -1,  353,  354,  355,  356,  357,  358,  359,    0, 
          361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
          262,  263,  264,   -1,  375,  267,  268,  269,   -1,  271, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  281, 
          282,   -1,   -1,   -1,  395,   -1,   -1,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  305,    0,   -1,   -1,   -1,   59,   -1, 
           -1,   -1,   -1,   -1,   10,   -1,  318,  319,  320,  321, 
          322,  323,  324,  325,  326,  327,   -1,  329,  330,   -1, 
           -1,  333,  334,   -1,   -1,   -1,   -1,  339,   -1,   -1, 
          342,   -1,   -1,  345,   -1,  347,   -1,  349,   44,  351, 
           -1,  353,  354,  355,  356,  357,  358,  359,   -1,  361, 
           -1,  363,   -1,   59,   -1,   61,   -1,   63,   -1,   -1, 
           -1,   -1,   -1,  375,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  395,   -1,   91,  262,  263,  264,   -1, 
           -1,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  280,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
            0,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           10,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,   -1,   -1,  333,  334,  335, 
          336,   -1,   -1,   -1,   -1,   -1,  342,   -1,   -1,   -1, 
           -1,   -1,   -1,  349,   44,  351,   -1,  353,  354,  355, 
      };
   }

   private static final short[] yyCheck3() {
      return new short[] {

          356,  357,  358,  359,   -1,  361,  362,  363,   -1,   59, 
           -1,   61,   -1,   63,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  395, 
          281,   91,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
           -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,   -1,   -1,   -1,  262,  263,  264,   -1, 
           -1,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  280,   -1,   -1,   -1,   -1,   -1, 
           -1,  342,   -1,   -1,  290,  291,   44,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   59,  363,   61,   -1,   63,   -1,   -1,   -1,   -1, 
           -1,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,   -1,   -1,  333,  334,  335, 
          336,   -1,   -1,   91,  395,   -1,  342,   -1,   -1,   -1, 
           -1,   -1,   -1,  349,   -1,  351,   -1,  353,  354,  355, 
          356,  357,  358,  359,   -1,  361,  362,  363,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
            0,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           10,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  395, 
           -1,   -1,  262,  263,  264,   -1,   -1,   -1,  268,  269, 
           -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   59, 
           -1,   61,   -1,   63,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   91,   -1,  333,  334,  335,  336,   -1,  338,   -1, 
           -1,   -1,  342,   -1,   -1,   -1,   -1,   -1,   -1,  349, 
           -1,  351,    0,  353,  354,  355,  356,  357,  358,  359, 
           -1,  361,   10,  363,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,   -1, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  395,   44,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   59,   -1,   61,   -1,   63,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   91,   -1,  333,  334,  335,  336,   -1, 
          338,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1,   -1, 
           -1,  349,    0,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,   10,  361,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  262,  263,  264,   -1,   44,  395,  268,  269, 
           -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   59,   -1,   61,   -1,   63,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   91,   -1,   -1,   -1,  317,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,  328,  329, 
          330,   -1,   -1,  333,  334,  335,  336,   -1,  338,   -1, 
           -1,   -1,  342,   -1,   -1,   -1,   -1,   -1,   -1,  349, 
           -1,  351,    0,  353,  354,  355,  356,  357,  358,  359, 
           -1,  361,   10,  363,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,   -1, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  395,   44,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   59,   -1,   61,   -1,   63,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   91,   -1,  333,  334,  335,  336,   -1, 
          338,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1,   -1, 
           -1,  349,    0,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,   10,  361,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,   -1, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   44,  395,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   59,   -1,   61,   -1,   63,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   91,   -1,  333,  334,  335,  336,   -1, 
          338,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1,   -1, 
           -1,  349,   -1,  351,    0,  353,  354,  355,  356,  357, 
          358,  359,   -1,  361,   10,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,  395,   44,   -1, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   59,   -1,   61,   -1,   63,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   91,   -1,   -1,   -1,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,    0,   -1,  333,  334,  335,  336,   -1, 
          338,   -1,   -1,   10,  342,   -1,   -1,   -1,   -1,   -1, 
           -1,  349,   -1,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,   -1,  361,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   44,   -1,   -1, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   59,   -1,   61,   -1,   63,  395,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   91,   -1,   -1,   -1,   -1,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   -1,   -1,  333,  334,  335,  336,   -1, 
          338,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1,   -1, 
           -1,  349,   -1,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,    0,  361,   -1,  363,   -1,   -1,   -1,   -1, 
           -1,   -1,   10,   -1,   -1,   -1,  262,  263,  264,   -1, 
           -1,   -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  395,   -1,   -1, 
           -1,   -1,   -1,   -1,  290,  291,   44,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   59,   -1,   61,   -1,   63,   -1,   -1,   -1,   -1, 
           -1,  317,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,  328,  329,  330,   -1,   -1,  333,  334,  335, 
          336,   -1,  338,   91,   -1,   -1,  342,   -1,   -1,   -1, 
           -1,   -1,   -1,  349,   -1,  351,    0,  353,  354,  355, 
          356,  357,  358,  359,   -1,  361,   10,  363,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1, 
           -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  395, 
           44,   -1,   -1,  290,  291,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   59,   -1,   -1,   -1,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   -1,   -1,  333,  334,  335,  336, 
           -1,  338,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1, 
           -1,   -1,  349,   44,  351,   -1,  353,  354,  355,  356, 
          357,  358,  359,   -1,  361,   -1,  363,   58,   59,   -1, 
           -1,   -1,   63,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  395,   -1, 
           -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1,   -1, 
          268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  290,  291,   -1,  293,  294,  295,  296,  297, 
           -1,   -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   10,   -1,   -1,   -1,   -1,   -1,  317, 
          318,  319,  320,  321,  322,  323,  324,  325,  326,  327, 
          328,  329,  330,   -1,   -1,  333,  334,  335,  336,   -1, 
          338,   -1,   -1,   -1,  342,   -1,   -1,   44,   -1,   -1, 
           -1,  349,   -1,  351,   -1,  353,  354,  355,  356,  357, 
          358,  359,   59,  361,   61,  363,   63,   -1,  262,  263, 
          264,   -1,   -1,  267,  268,  269,   -1,  271,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   91,   -1,   -1,  395,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,   -1,   -1,   -1,   -1,  339,   -1,   -1,  342,   -1, 
          281,  282,   -1,   44,   -1,   -1,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   59,  363, 
           61,   -1,   63,   -1,  305,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  317,  318,  319,  320, 
          321,   -1,   -1,  324,  325,  326,  327,   -1,  329,  330, 
           91,  395,  333,  334,   -1,   -1,   -1,   -1,  339,   -1, 
           -1,  342,   -1,   -1,  345,   -1,  347,    0,  349,   -1, 
          351,   -1,  353,  354,  355,  356,   -1,   10,  359,   -1, 
          361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  375,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   44,   -1,   -1,  395,  262,  263,  264,   -1,   -1, 
           -1,  268,  269,   -1,  271,   58,   59,   -1,   -1,   -1, 
           63,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  290,  291,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   -1,   -1,  333,  334,  335,  336, 
           -1,   -1,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1, 
           -1,   -1,  349,    0,  351,   -1,  353,  354,  355,  356, 
          357,  358,  359,   10,  361,   -1,  363,   -1,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,   -1,  268,  269,   -1, 
          271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   44,  395,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
           -1,   58,   59,   -1,   -1,   -1,   63,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  317,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,  328,  329,  330, 
           -1,   -1,  333,  334,  335,  336,   -1,   -1,   -1,   -1, 
           -1,  342,   -1,   -1,   -1,   -1,   -1,   -1,  349,   -1, 
          351,   -1,  353,  354,  355,  356,  357,  358,  359,   -1, 
          361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,  262, 
          263,  264,   -1,   -1,  267,  268,  269,   -1,  271,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  281,  282, 
           -1,   -1,   -1,   -1,  395,   -1,   -1,  290,  291,   -1, 
          293,  294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  305,    0,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   10,   -1,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,   -1,  329,  330,   -1,   -1, 
          333,  334,   -1,   -1,   -1,   -1,  339,   -1,   -1,  342, 
           -1,   -1,  345,   -1,  347,   -1,   -1,   44,  351,   -1, 
           -1,   -1,  355,  356,  357,  358,  359,   -1,  361,   -1, 
          363,   -1,   59,   -1,   61,   -1,   63,   -1,   -1,   -1, 
           -1,   -1,  375,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  395,   -1,   91,  262,  263,  264,   -1,   -1, 
          267,  268,  269,   -1,  271,   -1,    0,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  281,  282,   10,   -1,   -1,   -1, 
           -1,   -1,   -1,  290,  291,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           44,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,   -1,  329,  330,   58,   59,  333,  334,   -1,   63, 
           -1,   -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1, 
          347,   -1,   -1,   -1,  351,   -1,   -1,   -1,  355,  356, 
          357,  358,  359,   -1,  361,   -1,  363,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  306,  307,   -1,  375,  310, 
           -1,   -1,   -1,  314,  315,   -1,  317,  318,  319,  320, 
          321,  322,  323,   -1,    0,  326,  327,   -1,  395,   -1, 
          331,  332,  333,  334,   10,   -1,   -1,   -1,   -1,   -1, 
          341,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1, 
          351,  352,  353,  354,  355,  356,  357,  358,  359,  360, 
          361,   -1,   -1,  364,   -1,  262,  263,  264,   44,   -1, 
           -1,  268,  269,   -1,  271,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   58,   59,   -1,   -1,   -1,   63,   -1,   -1, 
           -1,   -1,   -1,  290,  291,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          317,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,  328,  329,  330,   -1,   -1,  333,  334,  335,  336, 
           -1,   -1,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1, 
           -1,   -1,  349,   -1,  351,   -1,  353,  354,  355,  356, 
          357,  358,  359,   -1,  361,   -1,  363,   -1,  262,  263, 
          264,   -1,   -1,  267,  268,  269,   -1,  271,   -1,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  281,  282,   10, 
           -1,   -1,   -1,   -1,   -1,   -1,  290,  291,  395,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  305,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   44,  318,  319,  320,  321,  322,  323, 
          324,  325,  326,  327,   -1,  329,  330,   58,   59,  333, 
          334,   -1,   63,   -1,   -1,  339,   -1,   -1,  342,   -1, 
           -1,  345,   -1,  347,   -1,   -1,   -1,  351,   -1,   -1, 
           -1,   -1,   -1,  357,  358,  359,   -1,  361,   -1,  363, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  375,   -1,   -1,   -1,   -1,  262,  263,  264,   -1, 
           -1,  267,  268,  269,   -1,  271,   -1,    0,   -1,   -1, 
           -1,  395,   -1,   -1,   -1,  281,  282,   10,   -1,   -1, 
           -1,   -1,   -1,   -1,  290,  291,   -1,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  305, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   44,  318,  319,  320,  321,  322,  323,  324,  325, 
          326,  327,   -1,  329,  330,   58,   59,  333,  334,   -1, 
           63,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1,  345, 
           -1,  347,   -1,   -1,   -1,  351,   -1,   -1,   -1,   -1, 
           -1,  357,  358,  359,   -1,  361,   -1,  363,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  375, 
            0,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           10,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  395, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  262,  263,  264,   44,   -1,  267,  268,  269,   -1, 
          271,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59, 
          281,  282,   -1,   63,   -1,   -1,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  305,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,  319,  320, 
          321,  322,  323,  324,  325,  326,  327,   -1,  329,  330, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  339,   -1, 
           -1,  342,   -1,    0,  345,   -1,  347,   -1,   -1,   -1, 
          351,   -1,   -1,   10,   -1,   -1,  357,  358,  359,   -1, 
          361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  375,   -1,   -1,   -1,   -1,  262, 
          263,  264,   -1,   -1,  267,  268,  269,   44,  271,   -1, 
           -1,   -1,   -1,   -1,  395,   -1,   -1,   -1,  281,  282, 
           -1,   58,   59,   -1,   -1,   -1,   63,  290,  291,   -1, 
          293,  294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  305,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  318,  319,  320,  321,  322, 
          323,  324,  325,  326,  327,   -1,  329,  330,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,  342, 
            0,   -1,  345,   -1,  347,   -1,   -1,   -1,   -1,   -1, 
           10,   -1,   -1,   -1,  357,  358,  359,   -1,  361,   -1, 
          363,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           -1,  271,  375,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  281,  282,   -1,   44,   -1,   -1,   -1,   -1,   -1, 
          290,  291,  395,  293,  294,  295,  296,  297,   58,   59, 
           -1,   -1,   -1,   63,   -1,  305,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,   -1,  329, 
          330,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  339, 
           -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  357,  358,  359, 
           -1,  361,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  375,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1,   -1, 
          267,  268,  269,   -1,  271,  395,    0,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  281,  282,   10,   -1,   -1,   -1, 
           -1,   -1,   -1,  290,  291,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           44,  318,  319,  320,  321,  322,  323,  324,  325,  326, 
          327,   -1,  329,  330,   58,   59,   -1,   -1,   -1,   63, 
           -1,   -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1, 
          347,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          357,  358,   -1,    0,   -1,   -1,  363,   -1,   -1,   -1, 
           -1,   -1,   -1,   10,   -1,   -1,   -1,   -1,  375,   -1, 
           -1,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,  395,   -1, 
           -1,  281,  282,   -1,   -1,   -1,   -1,   44,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,   58,   59,   -1,   -1,  305,   63,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  318,  319, 
          320,  321,  322,  323,  324,  325,  326,  327,   -1,  329, 
          330,   -1,   -1,   -1,    0,   -1,   -1,   -1,   -1,  339, 
           -1,   -1,  342,   -1,   10,  345,   -1,  347,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  357,  358,   -1, 
           -1,   -1,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  375,   -1,   -1,   44,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   58,   59,   -1,  395,   -1,   63,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  262,  263, 
          264,   -1,   -1,  267,  268,  269,   -1,  271,   -1,   -1, 
           -1,    0,   -1,   -1,   -1,   -1,   -1,  281,  282,   -1, 
           -1,   10,   -1,   -1,   -1,   -1,  290,  291,   -1,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  305,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   44,   -1,   -1,   -1,   -1, 
          324,  325,   -1,   -1,   -1,  329,  330,   -1,   -1,   58, 
           59,   -1,   -1,   -1,   63,  339,   -1,   -1,  342,   -1, 
           -1,  345,   -1,  347,   -1,  262,  263,  264,   -1,   -1, 
          267,  268,  269,   -1,  271,   -1,   -1,    0,   -1,  363, 
           -1,   -1,   -1,   -1,  281,  282,   -1,   10,   -1,   -1, 
           -1,  375,   -1,  290,  291,   -1,  293,  294,  295,  296, 
          297,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  305,   -1, 
           -1,  395,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   44,   -1,   -1,   -1,   -1,   -1,  324,  325,   -1, 
           -1,   -1,  329,  330,   -1,   58,   59,   -1,   -1,   -1, 
           63,   -1,  339,   -1,   -1,  342,   -1,   -1,  345,   -1, 
          347,   -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1, 
           -1,  267,  268,  269,   -1,  271,  363,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  281,  282,   -1,  375,   -1, 
           -1,   -1,   -1,   -1,  290,  291,    0,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,   10,   -1,  395,  305, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  324,  325, 
           -1,   -1,   -1,  329,  330,   -1,   -1,   -1,   -1,   -1, 
           44,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1,  345, 
           -1,  347,   -1,   -1,   58,   59,   -1,   -1,   -1,   63, 
           -1,   -1,   -1,  262,  263,  264,   -1,  363,  267,  268, 
          269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,  375, 
           -1,   -1,  281,  282,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,  395, 
           -1,   -1,   -1,   -1,   -1,   -1,  305,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  324,  325,   -1,   -1,   -1, 
          329,  330,   -1,   -1,   -1,    0,   -1,   -1,   -1,   -1, 
          339,   -1,   -1,  342,   -1,   10,  345,   -1,  347,  262, 
          263,  264,   -1,   -1,  267,  268,  269,   -1,  271,   -1, 
           -1,   -1,   -1,   -1,  363,   -1,   -1,   -1,  281,  282, 
           -1,   -1,   -1,   -1,   -1,   -1,  375,  290,  291,   44, 
          293,  294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  305,   58,   59,   -1,  395,   -1,   63,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  324,  325,   -1,   -1,   -1,  329,  330,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,  342, 
           -1,   -1,  345,   -1,  347,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,    0,   -1,   -1,   -1,   -1,   -1, 
          363,   -1,   -1,   -1,   10,   -1,   -1,   -1,  262,  263, 
          264,   -1,  375,  267,  268,  269,   -1,  271,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  281,  282,   -1, 
           -1,   -1,  395,   -1,   -1,   -1,  290,  291,   44,  293, 
          294,  295,  296,  297,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  305,   58,   59,   -1,   -1,   -1,   63,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          324,  325,   -1,   -1,   -1,  329,  330,   -1,   -1,   -1, 
            0,   -1,   -1,   -1,   -1,  339,   -1,   -1,  342,   -1, 
           10,  345,   -1,  347,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  363, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  375,   -1,   -1,   44,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   58,   59, 
           -1,  395,   -1,   63,   -1,   -1,   -1,  262,  263,  264, 
           -1,   -1,  267,  268,  269,   -1,  271,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  281,  282,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  290,  291,    0,  293,  294, 
          295,  296,  297,   -1,   -1,   -1,   -1,   10,   -1,   -1, 
          305,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  324, 
          325,   -1,   -1,   -1,  329,  330,   -1,   -1,   -1,   -1, 
           -1,   44,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1, 
          345,   -1,  347,   -1,   -1,   58,   59,   -1,   -1,    0, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  363,   10, 
           -1,   -1,   -1,   -1,   -1,   -1,  262,  263,  264,   -1, 
          375,  267,  268,  269,   -1,  271,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  281,  282,   -1,   -1,   -1, 
          395,   -1,   -1,   44,  290,  291,   -1,  293,  294,  295, 
          296,  297,   -1,   -1,   -1,   -1,   -1,   58,   59,  305, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  325, 
           -1,   -1,   -1,  329,  330,   -1,   -1,   -1,    0,   -1, 
           -1,   -1,   -1,  339,   -1,   -1,  342,   -1,   10,  345, 
           -1,  347,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           -1,  271,   -1,   -1,   -1,   -1,   -1,  363,   -1,   -1, 
           -1,  281,  282,   -1,   -1,   -1,   -1,   -1,   -1,  375, 
          290,  291,   44,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  305,   58,   59,   -1,  395, 
            0,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           10,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  339, 
           -1,   -1,  342,   -1,   -1,  345,   -1,  347,   -1,   -1, 
           -1,   -1,   -1,   -1,   44,   -1,   -1,   -1,   -1,  262, 
          263,  264,   -1,  363,  267,  268,  269,   -1,  271,   59, 
           -1,   -1,   -1,   -1,   -1,  375,   -1,   -1,  281,  282, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1, 
          293,  294,  295,  296,  297,  395,   -1,   -1,   -1,    0, 
           -1,   -1,  305,   -1,   -1,   -1,   -1,   -1,   -1,   10, 
           -1,   -1,   -1,   -1,   -1,    0,   -1,   -1,   -1,   -1, 
           -1,  262,  263,  264,   -1,   10,  267,  268,  269,   -1, 
          271,   -1,   -1,   -1,   -1,    0,  339,   -1,   -1,  342, 
          281,  282,  345,   44,  347,   10,   -1,   -1,   -1,  290, 
          291,   -1,  293,  294,  295,  296,  297,   -1,   59,   44, 
          363,   -1,   -1,   -1,  305,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  375,   -1,   59,   -1,   -1,   -1,   -1,   44, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  395,   -1,   59,   -1,   -1,   -1,  339,   -1, 
           -1,  342,   -1,   -1,  345,   -1,  347,   -1,   -1,   -1, 
          262,  263,  264,   -1,   -1,  267,  268,  269,   -1,  271, 
           -1,    0,  363,   -1,   -1,   -1,   -1,   -1,   -1,  281, 
          282,   10,   -1,   -1,  375,   -1,   -1,   -1,  290,  291, 
           -1,  293,  294,  295,  296,  297,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  305,  395,   -1,   -1,    0,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   10,   -1,   -1, 
           -1,   -1,  262,  263,  264,   -1,   -1,  267,  268,  269, 
           59,  271,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1, 
          342,  281,  282,  345,   -1,  347,   -1,   -1,   -1,   -1, 
          290,  291,   -1,  293,  294,  295,  296,  297,   -1,   -1, 
           -1,  363,   -1,   -1,   -1,  305,   59,   -1,   -1,   -1, 
           -1,   -1,   -1,  375,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,    0,   -1,  395,   -1,   -1,   -1,   -1,   -1,  339, 
           -1,   10,  342,   -1,   -1,  345,   -1,  347,   -1,   -1, 
           -1,  262,  263,  264,   -1,   -1,  267,  268,  269,   -1, 
          271,   -1,   -1,  363,   -1,   -1,   -1,  262,  263,  264, 
          281,  282,  267,  268,  269,  375,  271,   -1,   -1,  290, 
          291,    0,  293,  294,  295,  296,  297,  262,  263,  264, 
           59,   10,  267,  268,  269,  395,  271,   -1,  293,  294, 
          295,  296,  297,   -1,   -1,   -1,  281,  282,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  290,  291,    0,  293,  294, 
          295,  296,  297,   -1,   -1,   44,   -1,   10,   -1,   -1, 
           -1,  342,   -1,   -1,  345,   -1,  347,   -1,   -1,   -1, 
           59,   -1,   -1,   -1,  339,   -1,   -1,  342,   -1,   -1, 
           -1,   -1,  363,    0,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   10,   -1,   -1,   -1,  342,  363,   -1, 
          345,    0,  347,  262,  263,  264,   59,   -1,  267,  268, 
          269,   10,  271,   -1,  395,   -1,   -1,   -1,  363,    0, 
           -1,   -1,  281,  282,   -1,   -1,   -1,   44,   -1,   10, 
          395,  290,  291,   -1,  293,  294,  295,  296,  297,  262, 
          263,  264,   59,   -1,  267,  268,  269,   -1,  271,   -1, 
          395,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  281,  282, 
           59,   -1,   -1,   -1,   -1,   -1,   -1,  290,  291,   -1, 
          293,  294,  295,  296,  297,   -1,   -1,   -1,   59,   -1, 
           -1,   -1,   -1,  342,   -1,   -1,  345,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  363,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  342, 
           -1,   -1,  345,  262,  263,  264,   -1,   -1,  267,  268, 
          269,   -1,  271,   -1,   -1,   -1,  395,   -1,   -1,   -1, 
          363,   -1,  281,   10,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  290,  291,   -1,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  395,  262,  263,  264,   -1,   -1,  267,  268, 
          269,   -1,  271,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  342,  293,  294,  295,  296,  297,  262, 
          263,  264,   -1,   -1,  267,  268,  269,   -1,  271,   -1, 
           -1,   -1,   -1,   -1,  363,   -1,   -1,   -1,  281,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          293,  294,  295,  296,  297,  262,  263,  264,   -1,   -1, 
          339,  268,  269,  342,  271,   -1,  395,   -1,   -1,   -1, 
           -1,   -1,   -1,  262,  263,  264,   -1,   -1,   -1,  268, 
          269,   -1,  271,   -1,  363,   -1,  293,  294,  295,  296, 
          297,  262,  263,  264,   10,   -1,   -1,  268,  269,  342, 
          271,   -1,   -1,   -1,  293,  294,  295,  296,  297,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  395,   -1,   -1,   -1, 
          363,   -1,  293,  294,  295,  296,  297,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   59,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  395,  342,   -1,   -1,  363,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  342,   -1,   -1,  363,   91,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  395,   -1, 
           -1,   -1,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,   -1,  395,   -1,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,  395,   -1,  283,  284,  285,  286, 
          287,  288,  289,   10,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   59,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   91,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  392,  393,  394,   -1,   -1, 
          256,  257,  258,  259,  260,  261,  262,  263,  264,  265, 
          266,   -1,   -1,  269,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,  280,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   10,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,  298,  299,   -1,  301,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  328,   -1,   -1,   -1,   -1,   -1,   -1,  335, 
          336,  337,   -1,   59,  340,  341,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,  357,   -1,   -1,  360,   -1,  362,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   91,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,  256, 
          257,  258,  259,  260,  261,  262,  263,  264,  265,  266, 
           -1,   -1,  269,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,  280,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   10,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,  298,  299,   -1,  301,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,   -1,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  328,   -1,   -1,   -1,   -1,   -1,   -1,  335,  336, 
          337,   -1,   59,  340,  341,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
          357,   -1,   -1,  360,   -1,  362,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   91,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  392,  393,   -1,   -1,   -1, 
          256,  257,  258,  259,  260,  261,  262,  263,  264,  265, 
          266,   -1,   -1,  269,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,  280,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   10,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,  298,  299,   -1,  301,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  328,   -1,   -1,   -1,   -1,   -1,   -1,  335, 
          336,  337,   -1,   59,  340,  341,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,  357,   -1,   -1,  360,   -1,  362,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   91,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,  256, 
          257,  258,  259,  260,  261,  262,  263,  264,  265,  266, 
           -1,   -1,  269,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   10,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,  298,  299,   -1,  301,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,   -1,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  328,   -1,   -1,   -1,   -1,   -1,   -1,  335,  336, 
          337,   -1,   59,  340,  341,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
          357,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  392,  393,   -1,   -1,   -1, 
          256,  257,  258,  259,  260,  261,  262,  263,  264,  265, 
          266,   -1,   -1,  269,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   10,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,  298,  299,   -1,  301,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  328,   -1,   -1,   -1,   -1,   -1,   -1,  335, 
          336,  337,   -1,   59,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,  357,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,  256, 
          257,  258,  259,  260,  261,  262,  263,  264,  265,  266, 
           -1,  268,  269,  270,  271,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   10,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,  298,  299,   -1,  301,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,   -1,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   59,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  392,  393,   -1,   -1,   -1, 
          256,  257,  258,  259,  260,  261,  262,  263,  264,  265, 
          266,   -1,   -1,  269,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   10,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,  298,  299,   -1,  301,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   59,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,  362,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,  256, 
          257,  258,  259,  260,  261,  262,  263,  264,  265,  266, 
           -1,   -1,  269,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   10,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,  298,  299,   -1,  301,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,   -1,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   59,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  392,  393,   -1,   -1,   -1, 
          256,  257,  258,  259,  260,  261,  262,  263,  264,  265, 
          266,   -1,   -1,  269,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   10,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,  298,  299,   -1,  301,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   59,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
      };
   }

   private static final short[] yyCheck4() {
      return new short[] {

          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
          256,  257,  258,  259,  260,  261,  392,  393,  264,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   10,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,  298,  299,   -1,  301,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,  256, 
          257,  258,  259,  260,  261,   -1,   -1,   -1,  265,  266, 
           -1,   -1,   10,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,  298,  299,   -1,  301,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,   -1,  313,  314,  315,  316, 
           -1,   59,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,  363,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
          257,  258,  259,  260,  261,  392,  393,   -1,  265,  266, 
           -1,   10,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,  298,  299,   -1,  301,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,   -1,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,  342,  343,  344,   -1,  346, 
           -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,  257, 
          258,  259,   -1,  261,   -1,  392,  393,  265,  266,   -1, 
           -1,   -1,  270,   -1,  272,  273,  274,  275,  276,  277, 
          278,   -1,   -1,   -1,   -1,  283,  284,  285,  286,  287, 
          288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1, 
           -1,  299,   -1,   -1,  302,  303,  304,   -1,  306,  307, 
          308,  309,  310,  311,   44,  313,  314,  315,  316,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  337, 
           -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346,   -1, 
           -1,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366,  367, 
          368,  369,  370,   -1,   -1,   -1,  374,   -1,  376,  377, 
           -1,  379,  380,  381,  382,   -1,   -1,   -1,  257,  258, 
          259,   -1,  261,   -1,  392,  393,  265,  266,   -1,   -1, 
           -1,  270,   -1,  272,  273,  274,  275,  276,  277,  278, 
           -1,   -1,   -1,   -1,  283,  284,  285,  286,  287,  288, 
          289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1, 
          299,   -1,   -1,  302,  303,  304,   -1,  306,  307,  308, 
          309,  310,  311,   44,  313,  314,  315,  316,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  337,   -1, 
           -1,  340,   -1,   -1,  343,  344,   -1,  346,   -1,   -1, 
           -1,   -1,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  360,   -1,   -1,   -1,   -1,  365,  366,  367,  368, 
          369,  370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1, 
          379,  380,  381,  382,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  392,  393,   -1,   -1,  257,  258,  259, 
           -1,  261,   -1,   -1,   -1,  265,  266,   -1,   -1,   -1, 
          270,   -1,  272,  273,  274,  275,  276,  277,  278,   -1, 
           -1,   -1,   -1,  283,  284,  285,  286,  287,  288,  289, 
           -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1,  299, 
           -1,   -1,  302,  303,  304,   -1,  306,  307,  308,  309, 
          310,  311,   -1,  313,   -1,   -1,  316,   59,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  337,   -1,   -1, 
          340,   -1,   -1,  343,  344,   -1,  346,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  365,  366,  367,  368,  369, 
          370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379, 
          380,  381,  382,   -1,   -1,   -1,  257,  258,  259,   -1, 
          261,   -1,  392,  393,  265,  266,   -1,   -1,   -1,  270, 
           -1,  272,  273,  274,  275,  276,  277,  278,   -1,   -1, 
           -1,   -1,  283,  284,  285,  286,  287,  288,  289,   -1, 
           -1,  292,   -1,   -1,   -1,   -1,   -1,   -1,  299,   -1, 
           -1,  302,  303,  304,   -1,  306,  307,  308,  309,  310, 
          311,   -1,  313,   -1,   -1,  316,   59,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340, 
           -1,   -1,  343,  344,   -1,  346,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  365,  366,  367,  368,  369,  370, 
           -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379,  380, 
          381,  382,   -1,   -1,   -1,  257,  258,  259,  260,  261, 
           -1,  392,  393,  265,  266,   -1,   -1,   -1,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,   -1,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   -1,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,  298,  299,  300,  301, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1, 
           -1,   -1,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,  257,  258,  259,  260,  261,   -1, 
          392,  393,  265,  266,   -1,   -1,   -1,  270,   -1,  272, 
          273,  274,  275,  276,  277,  278,   -1,   -1,   -1,   -1, 
          283,  284,  285,  286,  287,  288,  289,   -1,   -1,  292, 
           -1,   -1,   -1,   -1,   -1,  298,  299,   -1,  301,  302, 
          303,  304,   -1,  306,  307,  308,  309,  310,  311,   -1, 
          313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1,   -1, 
          343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1,  352, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1,   -1, 
           -1,   -1,  365,  366,  367,  368,  369,  370,   -1,   -1, 
           -1,  374,   -1,  376,  377,   -1,  379,  380,  381,  382, 
           -1,   -1,  256,  257,  258,  259,  260,  261,   -1,  392, 
          393,  265,  266,   -1,   -1,   -1,  270,   -1,  272,  273, 
          274,  275,  276,  277,  278,   -1,   -1,   -1,   -1,  283, 
          284,  285,  286,  287,  288,  289,   -1,   -1,  292,   -1, 
           -1,   -1,   -1,   -1,  298,  299,  300,  301,  302,  303, 
          304,   -1,  306,  307,  308,  309,  310,  311,   -1,  313, 
          314,  315,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343, 
          344,   -1,  346,   -1,  348,   -1,   -1,   -1,  352,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  360,   -1,   -1,   -1, 
           -1,  365,  366,  367,  368,  369,  370,   -1,   -1,   -1, 
          374,   -1,  376,  377,   -1,  379,  380,  381,  382,   -1, 
           -1,  256,  257,  258,  259,  260,  261,   -1,  392,  393, 
          265,  266,   -1,   -1,   -1,  270,   -1,  272,  273,  274, 
          275,  276,  277,  278,   -1,   -1,   -1,   -1,  283,  284, 
          285,  286,  287,  288,  289,   -1,   -1,  292,   -1,   -1, 
           -1,   -1,   -1,  298,  299,   -1,  301,  302,  303,  304, 
           -1,  306,  307,  308,  309,  310,  311,   -1,  313,  314, 
          315,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344, 
           -1,  346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,   -1,  374, 
           -1,  376,  377,   -1,  379,  380,  381,  382,   -1,   -1, 
          256,  257,  258,  259,  260,  261,   -1,  392,  393,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,  298,  299,   -1,  301,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,   -1,  392,  393,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,  392,  393,  394,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,  392,  393,  394,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,  392,  393,  394,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,  392,  393,  394,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,   -1,   -1,  350,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,  392,  393,  394,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,  392,  393,  394,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,  312,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,   -1,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,   -1, 
          257,  258,  259,  260,  261,  392,  393,  394,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,  298,  299,  300,  301,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,   -1,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,  348,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,  257, 
          258,  259,  260,  261,   -1,  392,  393,  265,  266,   -1, 
           -1,   -1,  270,   -1,  272,  273,  274,  275,  276,  277, 
          278,   -1,   -1,   -1,   -1,  283,  284,  285,  286,  287, 
          288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1, 
          298,  299,   -1,  301,  302,  303,  304,   -1,  306,  307, 
          308,  309,  310,  311,   -1,  313,  314,  315,  316,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  337, 
           -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346,   -1, 
          348,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366,  367, 
          368,  369,  370,   -1,   -1,   -1,  374,   -1,  376,  377, 
           -1,  379,  380,  381,  382,   -1,   -1,   -1,  257,  258, 
          259,   -1,  261,   -1,  392,  393,  265,  266,   -1,   -1, 
           -1,  270,   -1,  272,  273,  274,  275,  276,  277,  278, 
           -1,   -1,   -1,   -1,  283,  284,  285,  286,  287,  288, 
          289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1, 
          299,   -1,   -1,  302,  303,  304,   -1,  306,  307,  308, 
          309,  310,  311,   -1,  313,  314,  315,  316,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  337,   -1, 
           -1,  340,  341,   -1,  343,  344,   -1,  346,   -1,   -1, 
           -1,   -1,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  360,   -1,   -1,   -1,   -1,  365,  366,  367,  368, 
          369,  370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1, 
          379,  380,  381,  382,   -1,   -1,   -1,  257,  258,  259, 
           -1,  261,   -1,  392,  393,  265,  266,   -1,   -1,   -1, 
          270,   -1,  272,  273,  274,  275,  276,  277,  278,   -1, 
           -1,   -1,   -1,  283,  284,  285,  286,  287,  288,  289, 
           -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1,  299, 
           -1,   -1,  302,  303,  304,   -1,  306,  307,  308,  309, 
          310,  311,   -1,  313,  314,  315,  316,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  337,   -1,   -1, 
          340,   -1,   -1,  343,  344,   -1,  346,   -1,  348,   -1, 
           -1,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          360,   -1,   -1,   -1,   -1,  365,  366,  367,  368,  369, 
          370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379, 
          380,  381,  382,   -1,   -1,   -1,  257,  258,  259,   -1, 
          261,   -1,  392,  393,  265,  266,   -1,   -1,   -1,  270, 
           -1,  272,  273,  274,  275,  276,  277,  278,   -1,   -1, 
           -1,   -1,  283,  284,  285,  286,  287,  288,  289,   -1, 
           -1,  292,   -1,   -1,   -1,   -1,   -1,   -1,  299,   -1, 
           -1,  302,  303,  304,   -1,  306,  307,  308,  309,  310, 
          311,   -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340, 
           -1,   -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1, 
           -1,  352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360, 
           -1,   -1,   -1,   -1,  365,  366,  367,  368,  369,  370, 
           -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379,  380, 
          381,  382,   -1,   -1,   -1,  257,  258,  259,   -1,  261, 
           -1,  392,  393,  265,  266,   -1,   -1,   -1,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,   -1,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   -1,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,   -1,  299,   -1,   -1, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
          352,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1, 
           -1,   -1,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,  257,  258,  259,   -1,  261,   -1, 
          392,  393,  265,  266,   -1,   -1,   -1,  270,   -1,  272, 
          273,  274,  275,  276,  277,  278,   -1,   -1,   -1,   -1, 
          283,  284,  285,  286,  287,  288,  289,   -1,   -1,  292, 
           -1,   -1,   -1,   -1,   -1,   -1,  299,   -1,   -1,  302, 
          303,  304,   -1,  306,  307,  308,  309,  310,  311,   -1, 
          313,  314,  315,  316,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1,   -1, 
          343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1,  352, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,  360,   -1,   -1, 
           -1,   -1,  365,  366,  367,  368,  369,  370,   -1,   -1, 
           -1,  374,   -1,  376,  377,   -1,  379,  380,  381,  382, 
           -1,   -1,   -1,  257,  258,  259,   -1,  261,   -1,  392, 
          393,  265,  266,   -1,   -1,   -1,  270,   -1,  272,  273, 
          274,  275,  276,  277,  278,   -1,   -1,   -1,   -1,  283, 
          284,  285,  286,  287,  288,  289,   -1,   -1,  292,   -1, 
           -1,   -1,   -1,   -1,   -1,  299,   -1,   -1,  302,  303, 
          304,   -1,  306,  307,  308,  309,  310,  311,   -1,  313, 
          314,  315,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343, 
          344,   -1,  346,   -1,  348,   -1,   -1,   -1,  352,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  360,   -1,   -1,   -1, 
           -1,  365,  366,  367,  368,  369,  370,   -1,   -1,   -1, 
          374,   -1,  376,  377,   -1,  379,  380,  381,  382,   -1, 
           -1,   -1,  257,  258,  259,   -1,  261,   -1,  392,  393, 
          265,  266,   -1,   -1,   -1,  270,   -1,  272,  273,  274, 
          275,  276,  277,  278,   -1,   -1,   -1,   -1,  283,  284, 
          285,  286,  287,  288,  289,   -1,   -1,  292,   -1,   -1, 
           -1,   -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304, 
           -1,  306,  307,  308,  309,  310,  311,   -1,  313,  314, 
          315,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  337,   -1,   -1,  340,  341,   -1,  343,  344, 
           -1,  346,   -1,   -1,   -1,   -1,   -1,  352,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,   -1,  374, 
           -1,  376,  377,   -1,  379,  380,  381,  382,   -1,   -1, 
           -1,  257,  258,  259,   -1,  261,   -1,  392,  393,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,  314,  315, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,   -1,   -1,   -1,   -1,  352,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
          257,  258,  259,   -1,  261,   -1,  392,  393,  265,  266, 
           -1,   -1,   -1,  270,   -1,  272,  273,  274,  275,  276, 
          277,  278,   -1,   -1,   -1,   -1,  283,  284,  285,  286, 
          287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1, 
           -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1,  306, 
          307,  308,  309,  310,  311,   -1,  313,  314,  315,  316, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346, 
           -1,   -1,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366, 
          367,  368,  369,  370,   -1,   -1,   -1,  374,   -1,  376, 
          377,   -1,  379,  380,  381,  382,   -1,   -1,   -1,  257, 
          258,  259,   -1,  261,   -1,  392,  393,  265,  266,   -1, 
           -1,   -1,  270,   -1,  272,  273,  274,  275,  276,  277, 
          278,   -1,   -1,   -1,   -1,  283,  284,  285,  286,  287, 
          288,  289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1, 
           -1,  299,   -1,   -1,  302,  303,  304,   -1,  306,  307, 
          308,  309,  310,  311,   -1,  313,  314,  315,  316,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  337, 
           -1,   -1,  340,   -1,   -1,  343,  344,   -1,  346,   -1, 
           -1,   -1,   -1,   -1,  352,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  360,   -1,   -1,   -1,   -1,  365,  366,  367, 
          368,  369,  370,   -1,   -1,   -1,  374,   -1,  376,  377, 
           -1,  379,  380,  381,  382,   -1,   -1,   -1,  257,  258, 
          259,   -1,  261,   -1,  392,  393,  265,  266,   -1,   -1, 
           -1,  270,   -1,  272,  273,  274,  275,  276,  277,  278, 
           -1,   -1,   -1,   -1,  283,  284,  285,  286,  287,  288, 
          289,   -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1, 
          299,   -1,   -1,  302,  303,  304,   -1,  306,  307,  308, 
          309,  310,  311,   -1,  313,  314,  315,  316,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  337,   -1, 
           -1,  340,   -1,   -1,  343,  344,   -1,  346,   -1,   -1, 
           -1,   -1,   -1,  352,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  360,   -1,   -1,   -1,   -1,  365,  366,  367,  368, 
          369,  370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1, 
          379,  380,  381,  382,   -1,   -1,   -1,  257,  258,  259, 
           -1,  261,   -1,  392,  393,  265,  266,   -1,   -1,   -1, 
          270,   -1,  272,  273,  274,  275,  276,  277,  278,   -1, 
           -1,   -1,   -1,  283,  284,  285,  286,  287,  288,  289, 
           -1,   -1,  292,   -1,   -1,   -1,   -1,   -1,   -1,  299, 
           -1,   -1,  302,  303,  304,   -1,  306,  307,  308,  309, 
          310,  311,   -1,  313,   -1,   -1,  316,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  333,   -1,   -1,   -1,  337,   -1,   -1, 
          340,   -1,   -1,  343,  344,   -1,  346,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  365,  366,  367,  368,  369, 
          370,   -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379, 
          380,  381,  382,   -1,   -1,   -1,  257,  258,  259,   -1, 
          261,   -1,  392,  393,  265,  266,   -1,   -1,   -1,  270, 
           -1,  272,  273,  274,  275,  276,  277,  278,   -1,   -1, 
           -1,   -1,  283,  284,  285,  286,  287,  288,  289,   -1, 
           -1,  292,   -1,   -1,   -1,   -1,   -1,   -1,  299,   -1, 
           -1,  302,  303,  304,   -1,  306,  307,  308,  309,  310, 
          311,   -1,  313,   -1,   -1,  316,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340, 
           -1,   -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  365,  366,  367,  368,  369,  370, 
           -1,   -1,   -1,  374,   -1,  376,  377,   -1,  379,  380, 
          381,  382,   -1,   -1,   -1,  257,  258,  259,   -1,  261, 
           -1,  392,  393,  265,  266,   -1,   -1,   -1,  270,   -1, 
          272,  273,  274,  275,  276,  277,  278,   -1,   -1,   -1, 
           -1,  283,  284,  285,  286,  287,  288,  289,   -1,   -1, 
          292,   -1,   -1,   -1,   -1,   -1,   -1,  299,   -1,   -1, 
          302,  303,  304,   -1,  306,  307,  308,  309,  310,  311, 
           -1,  313,   -1,   -1,  316,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1, 
           -1,  343,  344,   -1,  346,   -1,  348,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  365,  366,  367,  368,  369,  370,   -1, 
           -1,   -1,  374,   -1,  376,  377,   -1,  379,  380,  381, 
          382,   -1,   -1,   -1,  257,  258,  259,   -1,  261,   -1, 
          392,  393,  265,  266,   -1,   -1,   -1,  270,   -1,  272, 
          273,  274,  275,  276,  277,  278,   -1,   -1,   -1,   -1, 
          283,  284,  285,  286,  287,  288,  289,   -1,   -1,  292, 
           -1,   -1,   -1,   -1,   -1,   -1,  299,   -1,   -1,  302, 
          303,  304,   -1,  306,  307,  308,  309,  310,  311,   -1, 
          313,   -1,   -1,  316,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,  337,   -1,   -1,  340,   -1,   -1, 
          343,  344,   -1,  346,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  365,  366,  367,  368,  369,  370,   -1,   -1, 
           -1,  374,   -1,  376,  377,   -1,  379,  380,  381,  382, 
           -1,   -1,   -1,  257,  258,  259,   -1,  261,   -1,  392, 
          393,  265,  266,   -1,   -1,   -1,  270,   -1,  272,  273, 
          274,  275,  276,  277,  278,   -1,   -1,   -1,   -1,  283, 
          284,  285,  286,  287,  288,  289,   -1,   -1,  292,   -1, 
           -1,   -1,   -1,   -1,   -1,  299,   -1,   -1,  302,  303, 
          304,   -1,  306,  307,  308,  309,  310,  311,   -1,  313, 
           -1,   -1,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343, 
          344,   -1,  346,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  365,  366,  367,  368,  369,  370,   -1,   -1,   -1, 
          374,   -1,  376,  377,   -1,  379,  380,  381,  382,   -1, 
           -1,   -1,  257,  258,  259,   -1,  261,   -1,  392,  393, 
          265,  266,   -1,   -1,   -1,  270,   -1,  272,  273,  274, 
          275,  276,  277,  278,   -1,   -1,   -1,   -1,  283,  284, 
          285,  286,  287,  288,  289,   -1,   -1,  292,   -1,   -1, 
           -1,   -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304, 
           -1,  306,  307,  308,  309,  310,  311,   -1,  313,   -1, 
           -1,  316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344, 
           -1,  346,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
          365,  366,  367,  368,  369,  370,   -1,   -1,   -1,  374, 
           -1,  376,  377,   -1,  379,  380,  381,  382,   -1,   -1, 
           -1,  257,  258,  259,   -1,  261,   -1,  392,  393,  265, 
          266,   -1,   -1,   -1,  270,   -1,  272,  273,  274,  275, 
          276,  277,  278,   -1,   -1,   -1,   -1,  283,  284,  285, 
          286,  287,  288,  289,   -1,   -1,  292,   -1,   -1,   -1, 
           -1,   -1,   -1,  299,   -1,   -1,  302,  303,  304,   -1, 
          306,  307,  308,  309,  310,  311,   -1,  313,   -1,   -1, 
          316,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,  337,   -1,   -1,  340,   -1,   -1,  343,  344,   -1, 
          346,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  365, 
          366,  367,  368,  369,  370,   -1,   -1,   -1,  374,   -1, 
          376,  377,   -1,  379,  380,  381,  382,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,  392,  393,  257,  258, 
          259,  260,  261,  262,  263,  264,  265,  266,  267,  268, 
          269,  270,  271,  272,  273,  274,  275,  276,  277,  278, 
          279,  280,   -1,   -1,  283,  284,  285,  286,  287,  288, 
          289,  290,  291,  292,  293,  294,  295,  296,  297,  298, 
          299,  300,  301,  302,  303,  304,   -1,  306,  307,  308, 
           -1,  310,   -1,   -1,   -1,  314,  315,   -1,  317,  318, 
          319,  320,  321,  322,  323,   -1,   -1,  326,  327,   -1, 
           -1,   -1,  331,  332,  333,  334,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348, 
          349,   -1,  351,  352,  353,  354,  355,  356,  357,  358, 
          359,  360,  361,   -1,   -1,  364,  365,  257,  258,  259, 
          260,  261,  262,  263,  264,  265,  266,  267,  268,  269, 
          270,  271,  272,  273,  274,  275,  276,  277,  278,  279, 
          280,   -1,   -1,  283,  284,  285,  286,  287,  288,  289, 
          290,  291,  292,  293,  294,  295,  296,  297,  298,  299, 
          300,  301,  302,  303,  304,   -1,  306,  307,   -1,   -1, 
          310,   -1,   -1,   -1,  314,  315,   -1,  317,  318,  319, 
          320,  321,  322,  323,   -1,   -1,  326,  327,   -1,   -1, 
           -1,  331,  332,  333,  334,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349, 
           -1,  351,  352,  353,  354,  355,  356,  357,  358,  359, 
          360,  361,   -1,   -1,  364,  365,  257,  258,  259,  260, 
          261,  262,  263,  264,  265,  266,  267,  268,  269,  270, 
          271,  272,  273,  274,  275,  276,  277,  278,  279,  280, 
           -1,   -1,  283,  284,  285,  286,  287,  288,  289,  290, 
          291,  292,  293,  294,  295,  296,  297,  298,  299,  300, 
          301,  302,  303,  304,   -1,  306,  307,  308,  309,  310, 
          311,   -1,   -1,  314,  315,   -1,  317,  318,  319,  320, 
          321,  322,  323,   -1,   -1,  326,  327,   -1,   -1,   -1, 
          331,  332,  333,  334,   -1,   -1,   -1,   -1,   -1,   -1, 
          341,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1, 
          351,  352,  353,  354,  355,  356,  357,  358,  359,  360, 
          361,   -1,   -1,  364,  257,  258,  259,  260,  261,  262, 
          263,  264,  265,  266,  267,  268,  269,  270,  271,  272, 
          273,  274,  275,  276,  277,  278,  279,  280,   -1,   -1, 
          283,  284,  285,  286,  287,  288,  289,  290,  291,  292, 
          293,  294,  295,  296,  297,  298,  299,  300,  301,  302, 
          303,  304,   -1,  306,  307,  308,  309,  310,  311,   -1, 
           -1,  314,  315,   -1,  317,  318,  319,  320,  321,  322, 
          323,   -1,   -1,  326,  327,   -1,   -1,   -1,  331,  332, 
          333,  334,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,   -1,   -1,  348,  349,   -1,  351,  352, 
          353,  354,  355,  356,  357,  358,  359,  360,  361,   -1, 
           -1,  364,  257,  258,  259,  260,  261,  262,  263,  264, 
          265,  266,  267,  268,  269,  270,  271,  272,  273,  274, 
          275,  276,  277,  278,  279,  280,   -1,   -1,  283,  284, 
          285,  286,  287,  288,  289,  290,  291,  292,  293,  294, 
          295,  296,  297,  298,  299,  300,  301,  302,  303,  304, 
           -1,  306,  307,   -1,   -1,  310,   -1,   -1,   -1,  314, 
          315,   -1,  317,  318,  319,  320,  321,  322,  323,   -1, 
           -1,  326,  327,   -1,   -1,   -1,  331,  332,  333,  334, 
           -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 
           -1,   -1,   -1,  348,  349,   -1,  351,  352,  353,  354, 
          355,  356,  357,  358,  359,  360,  361,  306,  307,  364, 
           -1,  310,   -1,   -1,   -1,  314,  315,   -1,  317,  318, 
          319,  320,  321,  322,  323,   -1,   -1,  326,  327,   -1, 
           -1,   -1,  331,  332,  333,  334,   -1,   -1,   -1,   -1, 
           -1,   -1,  341,   -1,   -1,   -1,   -1,   -1,   -1,  348, 
          349,   -1,  351,  352,  353,  354,  355,  356,  357,  358, 
          359,  360,  361,  306,  307,  364,   -1,  310,   -1,   -1, 
           -1,  314,  315,   -1,  317,  318,  319,  320,  321,  322, 
          323,   -1,   -1,  326,  327,   -1,   -1,   -1,  331,  332, 
          333,  334,   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1, 
           -1,   -1,   -1,   -1,   -1,  348,  349,   -1,  351,  352, 
          353,  354,  355,  356,  357,  358,  359,  360,  361,  306, 
          307,  364,   -1,  310,   -1,   -1,   -1,  314,  315,   -1, 
          317,  318,  319,  320,  321,  322,  323,   -1,   -1,  326, 
          327,   -1,   -1,   -1,  331,  332,  333,  334,   -1,   -1, 
           -1,   -1,   -1,   -1,  341,   -1,   -1,   -1,   -1,   -1, 
           -1,  348,  349,   -1,  351,  352,  353,  354,  355,  356, 
          357,  358,  359,  360,  361,  306,  307,  364,   -1,  310, 
           -1,   -1,   -1,  314,  315,   -1,  317,  318,  319,  320, 
          321,  322,  323,   -1,   -1,  326,  327,   -1,   -1,   -1, 
          331,  332,  333,  334,   -1,   -1,   -1,   -1,   -1,   -1, 
          341,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1, 
          351,  352,  353,  354,  355,  356,  357,  358,  359,  360, 
          361,  306,  307,  364,   -1,  310,   -1,   -1,   -1,  314, 
          315,   -1,  317,  318,  319,  320,  321,  322,  323,   -1, 
           -1,  326,  327,   -1,   -1,   -1,  331,  332,  333,  334, 
           -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,   -1,   -1, 
           -1,   -1,   -1,  348,  349,   -1,  351,  352,  353,  354, 
          355,  356,  357,  358,  359,  360,  361,  306,  307,  364, 
           -1,  310,   -1,   -1,   -1,  314,  315,   -1,  317,  318, 
          319,  320,  321,  322,  323,   -1,   -1,  326,  327,   -1, 
           -1,   -1,  331,  332,  333,  334,   -1,   -1,   -1,   -1, 
           -1,   -1,  341,   -1,   -1,   -1,   -1,   -1,   -1,  348, 
          349,   -1,  351,  352,  353,  354,  355,  356,  357,  358, 
          359,  360,  361,  306,  307,  364,   -1,  310,   -1,   -1, 
           -1,  314,  315,   -1,  317,  318,  319,  320,  321,  322, 
          323,   -1,   -1,  326,  327,   -1,   -1,   -1,  331,  332, 
          333,  334,   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1, 
           -1,   -1,   -1,   -1,   -1,  348,  349,   -1,  351,  352, 
          353,  354,  355,  356,  357,  358,  359,  360,  361,  306, 
          307,  364,   -1,  310,   -1,   -1,   -1,  314,  315,   -1, 
          317,  318,  319,  320,  321,  322,  323,   -1,   -1,  326, 
          327,   -1,   -1,   -1,  331,  332,  333,  334,   -1,   -1, 
           -1,   -1,   -1,   -1,  341,   -1,   -1,   -1,   -1,   -1, 
           -1,  348,  349,   -1,  351,  352,  353,  354,  355,  356, 
          357,  358,  359,  360,  361,  306,  307,  364,   -1,  310, 
           -1,   -1,   -1,  314,  315,   -1,  317,  318,  319,  320, 
          321,  322,  323,   -1,   -1,  326,  327,   -1,   -1,   -1, 
          331,  332,  333,  334,   -1,   -1,   -1,   -1,   -1,   -1, 
          341,   -1,   -1,   -1,   -1,   -1,   -1,  348,  349,   -1, 
          351,  352,  353,  354,  355,  356,  357,  358,  359,  360, 
          361,  306,  307,  364,   -1,  310,   -1,   -1,   -1,  314, 
          315,   -1,  317,  318,  319,  320,  321,  322,  323,   -1, 
           -1,  326,  327,   -1,   -1,   -1,  331,  332,  333,  334, 
           -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,   -1,   -1, 
           -1,   -1,   -1,  348,  349,   -1,  351,  352,  353,  354, 
          355,  356,  357,  358,  359,  360,  361,   -1,   -1,  364, 
      };
   }

}
